package be.iminds.ilabt.jfed.lowlevel.api.test;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnection;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.rspec.rspec_source.RspecSource;
import be.iminds.ilabt.jfed.util.*;
import ch.ethz.ssh2.Connection;
import ch.ethz.ssh2.Session;

import javax.inject.Inject;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.InetSocketAddress;
import java.net.SocketImplFactory;
import java.security.NoSuchAlgorithmException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * TestMultiNodeLogin
 *
 * TestMultiNodeLogin will test node login for AMv2 or AMv3 (automatic and configurable).
 * It can test a specific nodes, and it can test any nodes.
 * More then 1 node will be tested, and links between the nodes will be created.
 * The nodelogin test will try to ping between the nodes over these links.
 *
 * NOTE: This does NOT support stitching: all nodes and links have to be on the same AM
 */
public class TestMultiNodeLogin extends TestNodeLogin {

    @Inject
    public TestMultiNodeLogin(Logger logger, TargetAuthority testedAuthority,
                              GeniUser user, JFedConnectionProvider connectionProvider,
                              AuthorityListModel authorityListModel, JFedPreferences jFedPreferences,
                              AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory automaticAggregateManagerWrapperFactory) {
        super(logger, testedAuthority, user, connectionProvider, authorityListModel, jFedPreferences, automaticAggregateManagerWrapperFactory);

    }

    public String getTestDescription() {
        return "This tests login to a multiple node on the same AM. AMv2 or AMv3 can be used. " +
                "Links between the nodes will be tested using ping." +
                "The sliver will be deleted afterwards.";
    }

    @Override
    public List<String> getOptConfKeys() {
        List<String> res = new ArrayList<String>(super.getOptConfKeys());

        res.add("node_count"); //number of nodes to test. Must be >=2

        res.remove("fixed_node_urn");
//        res.add("fixed_node_urns");  //currently not yet supported: modify getRequestRspec() to support this.

        //res.add("fixed_rspec"); //rspec to use
        //res.add("fixed_rspec_file"); //file containing the rspec to use

        return res;
    }

    @Override
    protected String getRequestRspec() throws IOException {
        final String nodeCountArg = getTestConfig().getProperty("node_count");
        int nodeCount = 2;
        if (nodeCountArg != null)
            nodeCount = Integer.parseInt(nodeCountArg);
        if (nodeCount < 2) {
            warn("Node count is configured as "+nodeCountArg+" which is too small. Changing to minimum value: 2");
            nodeCount = 2;
        }

        String rspecArg = getTestConfig().getProperty("fixed_rspec");
        final String rspecFileArg = getTestConfig().getProperty("fixed_rspec_file");
        if (rspecFileArg != null && rspecArg != null)
            throw new RuntimeException("Either specify fixed_rspec directly, or specify fixed_rpsec_file, but not both.");
        if (rspecFileArg != null) {
            rspecArg = IOUtils.fileToString(rspecFileArg);
        }
        if (rspecArg == null) {
            String sliverType = CommonAMTest.getSliverTypeForSite(testedAuthority.getSfaAuthorityForRspec());
            boolean exclusive = CommonAMTest.getExclusiveForSite(testedAuthority.getSfaAuthorityForRspec());
            String cmUrn = testedAuthority.getSfaAuthorityForRspec().getUrnString();

            RspecFactory rspecFactory = RspecFactoryFactory.getRspecFactoryInstance(ModelRspecType.BASIC);
            ModelRspec newRspec = rspecFactory.createModelRspec("request");

            RspecLink link = rspecFactory.createLinkWithClientId(newRspec, "link0", "lan");
            newRspec.addLink(link);
            for (int i = 0; i < nodeCount; i++) {
                RspecNode node = rspecFactory.createNodeWithClientId(newRspec, "n"+i);
                node.setComponentManagerId(cmUrn);
                node.setSliverTypeName(sliverType);
                node.setExclusive(exclusive);

                newRspec.addNode(node);

                RspecInterface iface = rspecFactory.createInterface(node, link, "n" + i + ":if0");
                iface.getIpAddresses().add(new RspecInterface.IpAddress("192.168.1."+(i+1), "255.255.255.0", "ipv4"));
            }

            assert newRspec.getNodes().size() == nodeCount;
            assert newRspec.getLinks().size() == 1;
            return newRspec.toGeni3Rspec();
        } else
            return rspecArg;
    }

    private List<String> pingTargets(ModelRspec modelRspec, RspecNode node) {
        Set<String> res = new HashSet<String>();
        //add all directly connected IPs
        for (RspecLink link : node.getLinks())
            for (RspecInterface iface : link.getInterfaces())
                for (RspecInterface.IpAddress ip :iface.getIpAddresses())
                    res.add(ip.getAddress());

        //remove own IPs
        for (RspecInterface iface : node.getInterfaces())
            for (RspecInterface.IpAddress ip :iface.getIpAddresses())
                res.remove(ip.getAddress());

        return new ArrayList(res);
    }

    @Override
    @Test(hardDepends = {"createSliver"}, softDepends = {"waitForSliverReady",
            "describeReadySliver"}, groups = {
            "nodelogin"})
    public void testNodeLogin() throws JFedException, IOException {
        assertNotNull(notEnoughResourcesDetected);
//        if (notEnoughResourcesDetected) skip("Test skipped because not enough free resources detected while tying to create the sliver(s)");
        if (notEnoughResourcesDetected) {
            warn("Test skipped because not enough free resources detected while tying to create the sliver(s)");
            return;
        }

        assert nodeLoginTester != null;
        commonAMTest.setupProxyForNodeLogin(getTestConfig(), nodeLoginTester, slice.urn);
        NodeLoginTester basicNodeLoginTester = nodeLoginTester;

        RspecSource rspecSource = new RequestRspecSource(requestRspec, ModelRspecType.BASIC);
        ModelRspec modelRspec = rspecSource.getModelRspec();

        long now = System.currentTimeMillis();
        long start = now;
        long deadline = now + (20 * 60 * 1000);

        boolean allLoginOk = true;
        for (RspecNode node : modelRspec.getNodes()) {
            NodeLoginTester nodeLoginTester = basicNodeLoginTester.copy();
            boolean foundNodeLogin = false;
            if (!foundNodeLogin && manifestRspec2 != null)
                foundNodeLogin = nodeLoginTester.parseSshInfoFromGeni3ManifestRspec(manifestRspec2, null/*auth*/, node.getClientId());
            if (!foundNodeLogin && manifestRspec != null)
                foundNodeLogin = nodeLoginTester.parseSshInfoFromGeni3ManifestRspec(manifestRspec, null/*auth*/, node.getClientId());

            if (!foundNodeLogin || nodeLoginTester.getSshHostname() == null) {
                errorNonFatal("No node / service / login in manifest RSpec for node \"" + node.getClientId() + "\", so SSH login cannot be tested.");
                allLoginOk = false;
                continue;
            }
            boolean isAuthenticated;
            String remainingMethods = "";
            Connection conn = null;
            JFedConnection.SshProxyInfo sshProxy = nodeLoginTester.getSshProxy();

            do { //try to setup connection until deadline
                SocketImplFactory socketImplFactory = null;
                if (sshProxy != null) {
                    SshServerProxyHelper.SshProxyInfo sshProxyInfo = new SshServerProxyHelper.SshProxyInfo(
                            new InetSocketAddress(sshProxy.getHostname(), sshProxy.getPort()),
                            sshProxy.getUsername(),
                            sshProxy.getHostKey(),
                            new String(KeyUtil.privateKeyToAnyPem(sshProxy.getSshKeyInfo().getPrivateKey()))
                    );
                    note("Using SSH proxy for SSH node login test. info: " + sshProxyInfo);
                    socketImplFactory = GanymedSshSocketImplFactory.getFactory(sshProxyInfo);
                }

                conn = new Connection(socketImplFactory, nodeLoginTester.getSshHostname(), nodeLoginTester.getSshPort());
                try {
                    conn.connect();

                    String privateKeyToPrint = new String(nodeLoginTester.getSshKeyHelper().getPEMAnyPrivateKey());
                    if (privateKeyToPrint.length() > 75) privateKeyToPrint = privateKeyToPrint.substring(0, 75) + " ...";
                    note("Connected to target node. Authenticating as " + nodeLoginTester.getSshUsername() + " with PEM private key:\n" + privateKeyToPrint);

                    isAuthenticated =
                            conn.authenticateWithPublicKey(nodeLoginTester.getSshUsername(), nodeLoginTester.getSshKeyHelper().getPEMRsaPrivateKey(), "nopass"); //needs RSA PRIVATE KEY   does not work with PRIVATE KEY
                } catch (IOException e) {
                    note("Exception while setting up SSH connection: " + e.getMessage());
                    isAuthenticated = false;
                }
                if (!isAuthenticated) {
                    String[] remainingMethodsArray;
                    try {
                        remainingMethodsArray = conn == null ? new String[]{} : conn.getRemainingAuthMethods(nodeLoginTester.getSshUsername());
                    } catch (Throwable e) {
                        warn("Exception while calling getRemainingAuthMethods: " + e.getMessage());
                        remainingMethodsArray = new String[]{};
                    }
                    remainingMethods = "";
                    for (String remainingMethod : remainingMethodsArray)
                        remainingMethods += " " + remainingMethod;
                    if (remainingMethodsArray.length == 0)
                        remainingMethods = "   no other authentication methods returned by server.";
                    else
                        remainingMethods =
                                "   Remaining authentication methods:" + remainingMethods + "  (these will not be tried by this test)";

                    note("SSH connection not successful (should be available from the moment that status is 'ready').  Trying again in 30 seconds...");
                    try {
                        Thread.sleep(30000 /*ms*/);
                    } catch (InterruptedException e) { /* Ignore*/ }
                    now = System.currentTimeMillis();
                }
            } while (!isAuthenticated && now < deadline);

            if (!isAuthenticated) {
                errorNonFatal("Could not connect to node \"" + nodeLoginTester.getSshHostname() + "\" at port " + nodeLoginTester.getSshPort()+" "+remainingMethods);
                allLoginOk = false;
                continue;
            }

            note("SSH connection authenticated successfully, will open session.");

            List<String> pingTargets = pingTargets(modelRspec, node);
            for (String targetIP : pingTargets) {
                note("Trying to ping " + targetIP);
                     /*
                     Ping options on linux:
                            -c count
                                   Stop after sending count ECHO_REQUEST packets. With deadline option, ping waits for count ECHO_REPLY packets, until the timeout expires.
                            -n     Numeric output only.  No attempt will be made to lookup symbolic names for host addresses.
                            -w deadline
                                   Specify a timeout, in seconds, before ping exits regardless of how many packets have been sent or received. In this case ping does not stop after count packet are sent, it
                                   waits either for deadline expire or until count probes are answered or for some error notification from network.
                            -W timeout
                                   Time to wait for a response, in seconds. The option affects only timeout in absense of any responses, otherwise ping waits for two RTTs.
                     * */

                Session session = conn.openSession();
                BufferedReader sout = new BufferedReader(new InputStreamReader(session.getStdout()));
                BufferedReader serr = new BufferedReader(new InputStreamReader(session.getStderr()));

                String command = "ping -c 5 -n -w 30 " + targetIP;
                session.execCommand(command);
                String result = "";
                String err = "";
                String line = sout.readLine();
                while (line != null) {
                    result += line + "\n";
                    line = sout.readLine();
                }
                line = serr.readLine();
                while (line != null) {
                    err += line + "\n";
                    line = serr.readLine();
                }
                session.close();
                note("\"" + command + "\" command on " + nodeLoginTester.getSshUsername() + "@" + nodeLoginTester.getSshHostname() + ":" + nodeLoginTester.getSshPort() +
                        " result: \"" + result.trim() + "\". (stderr is: \"" + err + "\")");
                setErrorsNotFatal();
                assertTrue(result.length() > 5, "I executed \"ping\" on the remote host, and expected some reply. Instead I got: \"" + result.trim() + "\". (stderr is: \"" + err + "\")");

                if (result.length() > 5) {
                    //example reply:
                    //   5 packets transmitted, 3 received
                    Pattern patternTrans = Pattern.compile("([0-9]*) packets transmitted");
                    Matcher matcherTrans = patternTrans.matcher(result);
                    assertTrue(matcherTrans.find(), "Did not find \"[0-9]* packets transmitted\" in result");
                    String sent = matcherTrans.group(1);

                    Pattern patternRecv = Pattern.compile("([0-9]*) received");
                    Matcher matcherRecv = patternRecv.matcher(result);
                    assertTrue(matcherRecv.find(), "Did not find \"[0-9]* received\" in result");
                    String recv = matcherRecv.group(1);

                    if (!sent.equals("5") || !recv.equals("5")) {
                        session = conn.openSession();
                        sout = new BufferedReader(new InputStreamReader(session.getStdout()));
                        serr = new BufferedReader(new InputStreamReader(session.getStderr()));
                        command = "echo; route - n; echo; echo; ifconfig -a";
                        session.execCommand(command);
                        result = "";
                        err = "";
                        line = sout.readLine();
                        while (line != null) {
                            result += line + "\n";
                            line = sout.readLine();
                        }
                        line = serr.readLine();
                        while (line != null) {
                            err += line + "\n";
                            line = serr.readLine();
                        }
                        session.close();
                        note("Ping failed, gathering informational with: \"" + command + "\" command. result: \"" + result.trim() + "\". (stderr is: \"" + err + "\")");
                    }

                    assertEquals("5", sent, "Packets sent count is not 5: sent=\"" + sent + "\"  (recv=\"" + recv + "\")");
                    assertEquals("5", recv, "Packets received count is not 5: \"" + recv + "\"");
                }
                setErrorsFatal();
            }

            conn.close();
        }

        assertTrue(allLoginOk, "one or more login tests failed");
    }
}
