package be.iminds.ilabt.jfed.testing.base;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.AutomaticAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.TargetAuthority;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Arrays;

/**
 * ApiTestFactory
 */
public class ApiTestFactory {
    private final static org.slf4j.Logger LOG = LoggerFactory.getLogger(ApiTestFactory.class);

    protected final GeniUser user;

    protected final JFedConnectionProvider connectionProvider;
    protected final AuthorityListModel authorityListModel;
    protected final JFedPreferences jFedPreferences;
    protected final AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory automaticAggregateManagerWrapperFactory;

    @Inject
    public ApiTestFactory(GeniUser user, JFedConnectionProvider connectionProvider, AuthorityListModel authorityListModel, JFedPreferences jFedPreferences,
                          AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory automaticAggregateManagerWrapperFactory) {
        this.user = user;
        this.connectionProvider = connectionProvider;
        this.authorityListModel = authorityListModel;
        this.jFedPreferences = jFedPreferences;
        this.automaticAggregateManagerWrapperFactory = automaticAggregateManagerWrapperFactory;
    }

    public ApiTest createApiTest(Logger logger, TargetAuthority testedAuthority, String testClassName) throws InvalidApiTestException {
        try {
            Class<?> foundClass = Class.forName(testClassName);
            if (!ApiTest.class.isAssignableFrom(foundClass)) {
                throw new InvalidApiTestException("The test class \""+testClassName+"\" is not an ApiTest");
            }
            Class<? extends ApiTest> testClass = (Class<? extends ApiTest>) foundClass;
            return createApiTest(logger, testedAuthority, testClass);
        } catch (ClassNotFoundException e) {
            throw new InvalidApiTestException("The test class \""+testClassName+"\" cannot be found", e);
        }
    }
    public ApiTest createApiTest(Logger logger, TargetAuthority testedAuthority, Class<? extends ApiTest> testClass) throws InvalidApiTestException {

        //all constructors must take these arguments:
        //Logger logger, TargetAuthority testedAuthority, GeniUser user, JFedConnectionProvider connectionProvider, AuthorityListModel authorityListModel

        //and optionally: AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory automaticAggregateManagerWrapperFactory

        try {
            ApiTest test = null;

            Constructor[] constructors = testClass.getConstructors();
            for (Constructor c : constructors) {
                if (c.getParameterCount() == 6) {
                    assert test == null;
                    Constructor<? extends ApiTest> constructor = testClass.getConstructor(
                            Logger.class,
                            TargetAuthority.class,
                            GeniUser.class,
                            JFedConnectionProvider.class,
                            AuthorityListModel.class,
                            JFedPreferences.class);

                    test = constructor.newInstance(
                            logger,
                            testedAuthority,
                            user,
                            connectionProvider,
                            authorityListModel,
                            jFedPreferences);
                }
                if (c.getParameterCount() == 7) {
                    assert test == null;
                    Constructor<? extends ApiTest> constructor = testClass.getConstructor(
                            Logger.class,
                            TargetAuthority.class,
                            GeniUser.class,
                            JFedConnectionProvider.class,
                            AuthorityListModel.class,
                            JFedPreferences.class,
                                                AutomaticAggregateManagerWrapper.AutomaticAggregateManagerWrapperFactory.class);

                    test = constructor.newInstance(
                            logger,
                            testedAuthority,
                            user,
                            connectionProvider,
                            authorityListModel,
                            jFedPreferences,
                            automaticAggregateManagerWrapperFactory);
                }
            }

            if (test == null) {
                throw new RuntimeException("Test class \""+testClass.getName()+"\" has no suitable constructor");
            }

            return test;
        } catch (NoSuchMethodException e) {
            LOG.error("ApiTest subclass constructor not found. Available constructors: " + Arrays.asList(testClass.getConstructors()));
            throw new InvalidApiTestException("The class \""+testClass.getName()+"\" is not a valid ApiTest", e);
        } catch (InstantiationException e) {
            throw new InvalidApiTestException("The class \""+testClass.getName()+"\" is not a valid ApiTest", e);
        } catch (IllegalAccessException e) {
            throw new InvalidApiTestException("The class \""+testClass.getName()+"\" is not a valid ApiTest", e);
        } catch (InvocationTargetException e) {
            throw new InvalidApiTestException("The class \""+testClass.getName()+"\" is not a valid ApiTest", e);
        }
    }

    public class InvalidApiTestException extends Exception {
        public InvalidApiTestException() {
        }

        public InvalidApiTestException(String message) {
            super(message);
        }

        public InvalidApiTestException(String message, Throwable cause) {
            super(message, cause);
        }

        public InvalidApiTestException(Throwable cause) {
            super(cause);
        }

        public InvalidApiTestException(String message, Throwable cause, boolean enableSuppression, boolean writableStackTrace) {
            super(message, cause, enableSuppression, writableStackTrace);
        }
    }
}
