package be.iminds.ilabt.jfed.ssh_terminal_tool.putty;

import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfo;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.tmp_file_helpers.TmpFile;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Singleton;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.*;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

@Singleton
public class PageantHelper {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(PageantHelper.class);

    private final JFedPreferences jFedPreferences;
    private Set<SshKeyInfo> keyHistory = new HashSet<>();

    @Inject
    public PageantHelper(JFedPreferences jFedPreferences) {
        this.jFedPreferences = jFedPreferences;
    }

    /**
     * Non-blocking call. Registers key with pageant.
     *
     * @param sshKeyInfo null is also allowed (does nothing then)
     */
    public synchronized void registerKey(final SshKeyInfo sshKeyInfo) {
        if (sshKeyInfo == null) return;

        final List<SshKeyInfo> sshKeyInfos = new ArrayList<>();
        sshKeyInfos.add(sshKeyInfo);

        registerKeys(sshKeyInfos);
    }

    public synchronized void registerKeys(SshKeyInfo... keys) {
        List<SshKeyInfo> sshKeyInfos = new ArrayList<>();
        for (SshKeyInfo key : keys)
            sshKeyInfos.add(key);
        registerKeys(sshKeyInfos);
    }

    public synchronized void registerKeys(final List<SshKeyInfo> sshKeyInfos) {
        if (sshKeyInfos == null || sshKeyInfos.isEmpty()) return;

        //return;
        sshKeyInfos.stream().filter(sshKeyInfo -> keyHistory.contains(sshKeyInfo)).forEach(sshKeyInfo -> {
            LOG.info("Key was already registered with pageant. Not trying again.");
            //return;
            sshKeyInfos.remove(sshKeyInfo);
        });

        if (sshKeyInfos.isEmpty()) return;

//        try {
//            registeredKeyQueue.clear();

        registerKeyQueue.addAll(sshKeyInfos);

//            List<SshKeyInfo> registeredKeys = registeredKeyQueue.poll(1000l, TimeUnit.MILLISECONDS);
//            //ignore if null
//
//            if (registeredKeys != null)
//                keyHistory.addAll(registeredKeys);
//        }
//        catch (InterruptedException e) {
//            //ignore (whether from put or from poll)
//        }
        if (pageantThread == null)
            startPageantThread();
    }

    private LinkedBlockingQueue<SshKeyInfo> registerKeyQueue = new LinkedBlockingQueue<>();


//    /**
//     * This is only used for sloppy waiting until key is registered.
//     *
//     * Basically, we wait up to 1 second, and then give up. If pageant.exe returns within one second, registerKey returns faster.
//     * This sloppy approach is ok because:
//     *   - it's ok to wait to long
//     *   - it's not that bad to not wait long enough. Things are usually still ok.
//     * */
//    private static BlockingQueue<List<SshKeyInfo>> registeredKeyQueue = new LinkedBlockingQueue<List<SshKeyInfo>>();


    private boolean startedPageantBackgroundProcess = false;
    /**
     * pageant blocks when not yet running.
     */
//    private static void launchBackgroundPageant() {
//        try {
//            ArrayList<String> command = new ArrayList<>();
//            //start pageant in the background
//            command.add("cmd");
//            command.add("/c");
//            command.add("start");
//            command.add("\"pageant\"");
//
//            command.add(PreferencesUtil.getFile(PreferencesUtil.Preference.PREF_PUTTY_DIRECTORY).getPath() + File.separator + "pageant.exe");
//
//            LOG.debug("Starting background Pageant and waiting for it to stop");
//            LOG.info("background Pageant command: "+command);
//            Process p = Runtime.getRuntime().exec(command.toArray(new String[command.size()]));
//            p.getOutputStream().close();
//            BufferedReader input = new BufferedReader(new InputStreamReader(p.getInputStream()));
//            String line = null;
//            while ((line = input.readLine()) != null) {
//                LOG.debug("background pageant.exe output: " + line);
//            }
//            input.close();
//            LOG.debug("background Pageant has no more output");
//            try {
//                p.waitFor();
//            } catch (InterruptedException e) {
//                //ignore
//            }
//            LOG.debug("background Pageant returned");
//        } catch (Throwable t) {
//            LOG.warn("Error staring Pageant in the background");
//        }
//        startedPageantBackgroundProcess = true;
//    }

    private Thread pageantThread;

    private synchronized void startPageantThread() {
        if (pageantThread != null)
            throw new RuntimeException("Pageant Thread has already started! Cannot run more than 1 instance...");

        LOG.debug("Starting Pageant thread.");
        Runnable pageantRunnable = () -> {
            while (!Thread.interrupted()) {
                try {
                    final List<SshKeyInfo> sshKeyInfos = new ArrayList<>();
                    final SshKeyInfo firstSshKeyInfo =
                            registerKeyQueue.take(); //blocking call, waits until item in queue
                    sshKeyInfos.add(firstSshKeyInfo);

                    //wait at least 100ms for additional keys.
                    boolean othersAvailable = true;
                    while (othersAvailable) {
                        final SshKeyInfo additionalSshKeyInfo = registerKeyQueue.poll(100, TimeUnit.MILLISECONDS);
                        if (additionalSshKeyInfo != null) {
                            sshKeyInfos.add(additionalSshKeyInfo);
                        } else {
                            othersAvailable = false;
                        }
                    }

                    final List<TmpFile> tmpFiles = new ArrayList<>();
                    if (!sshKeyInfos.isEmpty())
                        try {
                            final LinkedList<String> command = new LinkedList<>();

                            //add the files, starting with unprotected, and ending with the ones that ask passphrase.
                            //this is a good order, because the unprotected ones will lickely be removed soon,
                            // while the protected ones are typically not removed.
                            //  but it still leaves some chance that a file gets deleted before pageant.exe deals with it.
                            //creates the file if needed. (release later deletes it again)
//jFed doesn't have key to PPK file. User will have to provide pageant with it.
//creates the file if needed. (release later deletes it again)
                            sshKeyInfos.stream().filter(sshKeyInfo -> sshKeyInfo != null && sshKeyInfo.getPuttyKeyFile() != null).forEach(sshKeyInfo -> {
                                TmpFile tmpUnlockedKeyFile = sshKeyInfo.getUnlockedPuttyKeyFile();
                                if (tmpUnlockedKeyFile != null) {
                                    tmpFiles.add(tmpUnlockedKeyFile);
                                    tmpUnlockedKeyFile.store();
                                    command.addFirst(tmpUnlockedKeyFile.getFilename()); //creates the file if needed. (release later deletes it again)
                                } else {
                                    //jFed doesn't have key to PPK file. User will have to provide pageant with it.
                                    TmpFile tmpKeyFile = sshKeyInfo.getPuttyKeyFile();
                                    if (tmpKeyFile != null) {
                                        tmpFiles.add(tmpKeyFile);
                                        tmpKeyFile.store();
                                        command.addLast(tmpKeyFile.getFilename()); //creates the file if needed. (release later deletes it again)
                                    }
                                }
                            });

                            command.addFirst(jFedPreferences.getFile(GuiPreferenceKey.PREF_PUTTY_DIRECTORY).getPath() + File.separator + "pageant.exe");

                            if (!startedPageantBackgroundProcess) {
                                LOG.debug("Starting Pageant first time, so starting in background thread. Not waiting for it to stop, as it might never.");
                                Thread backgroundPageantThread = new Thread(() -> {
                                    //launchBackgroundPageant();

                                    //just launch the first command to add a key
                                    try {
                                        LOG.debug("Starting first pageant.exe in background thread");
                                        LOG.info("First pageant.exe command: " + command);
                                        Process p = null;
                                        p =
                                                Runtime.getRuntime().exec(command.toArray(new String[command.size()]));
                                        p.getOutputStream().close();
                                        BufferedReader input =
                                                new BufferedReader(new InputStreamReader(p.getInputStream()));
                                        String line = null;
                                        while ((line = input.readLine()) != null) {
                                            LOG.debug("First pageant.exe output: " + line);
                                        }
                                        input.close();
                                        LOG.debug("First pageant.exe has no more output");
                                        try {
                                            p.waitFor();
                                        } catch (InterruptedException e) {
                                            //ignore
                                        }
                                        LOG.debug("First pageant.exe returned");

                                        //key is registered, let it be known
                                        keyHistory.addAll(sshKeyInfos);
//                                                try {
//                                                    registeredKeyQueue.put(sshKeyInfos);
//                                                } catch (InterruptedException e) {
                                        //ignore
//                                                }
                                    } catch (IOException e) {
                                        LOG.error("First pageant.exe threw IOException: " + e.getMessage(), e);
                                    }
                                });
                                backgroundPageantThread.setDaemon(true);
                                backgroundPageantThread.start();

                                startedPageantBackgroundProcess = true;

//                                    while (!startedPageantBackgroundProcess) { //don't do this, pageant might never stop when started the first time!
                                try {
                                    Thread.sleep(5000); //give pageant some time to start, otherwise, we might delete key files to early
                                } catch (InterruptedException e) {
                                    //ignore
                                }
//                                    }
                            } else {
                                LOG.debug("Starting Pageant and waiting for it to stop");
                                LOG.info("Pageant command: " + command);
                                Process p = Runtime.getRuntime().exec(command.toArray(new String[command.size()]));
                                p.getOutputStream().close();
                                BufferedReader input =
                                        new BufferedReader(new InputStreamReader(p.getInputStream()));
                                String line = null;
                                while ((line = input.readLine()) != null) {
                                    LOG.debug("pageant.exe output: " + line);
                                }
                                input.close();
                                LOG.debug("Pageant has no more output");
                                try {
                                    p.waitFor();
                                } catch (InterruptedException e) {
                                    //ignore
                                }
                                LOG.debug("Pageant returned");

                                //key is registered, let it be known
//                                    registeredKeyQueue.put(sshKeyInfos);
                                keyHistory.addAll(sshKeyInfos);

                                try {
                                    Thread.sleep(1000); //wait a while before allowing next pageant to be started
                                } catch (InterruptedException e) {
                                    //ignore
                                }
                            }

                        } catch (IOException e) {
                            LOG.error("IOException while trying to launch Pageant: " + e.getMessage(), e);
                        } finally {
                            LOG.debug("removing unencrypted Pageant key files");
                            //remove private key files (if needed)
                            for (TmpFile tmpFile : tmpFiles)
                                tmpFile.delete();
                        }
                } catch (InterruptedException e) {
                    //occurs only when take() was interrupted
                    //ignore
                }
            }
            LOG.debug("Pageant thread was interrupted and returned.");
            pageantThread = null;
        };

        pageantThread = new Thread(pageantRunnable);
        pageantThread.setName("Pageant-Thread");
        pageantThread.setDaemon(true); //=> this application may exit before putty stops
        pageantThread.start();
        LOG.debug("Started Pageant thread.");
    }
}
