package be.iminds.ilabt.jfed.experimenter_gui;

import be.iminds.ilabt.jfed.BaseGuiModule;
import be.iminds.ilabt.jfed.experimenter_gui.login.UserLoginController;
import be.iminds.ilabt.jfed.experimenter_gui.preferences.PreferencesDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.slice.Experiment;
import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentViewController;
import be.iminds.ilabt.jfed.experimenter_gui.ui.AcceleratorService;
import be.iminds.ilabt.jfed.highlevel.HighLevelModule;
import be.iminds.ilabt.jfed.highlevel.LowLevelModule;
import be.iminds.ilabt.jfed.highlevel.UserLoginModelModule;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.logging.JFedLoggerManager;
import be.iminds.ilabt.jfed.lowlevel.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfo;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.InvalidLoginException;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.PlanetlabUserLoginModel;
import be.iminds.ilabt.jfed.lowlevel.userloginmodel.UserLoginModelManager;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.preferences.JFedPreferencesApplier;
import be.iminds.ilabt.jfed.ssh_terminal_tool.putty.PageantHelper;
import be.iminds.ilabt.jfed.ssh_terminal_tool.ssh_key_info.UserSshKeyInfo;
import be.iminds.ilabt.jfed.ssh_terminal_tool.unix.SshAgentHelper;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.ui.javafx.util.ExtraInfoCallBackGUIs;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.IOUtils;
import be.iminds.ilabt.jfed.util.JFedUtils;
import be.iminds.ilabt.jfed.util.OSDetector;
import com.cathive.fx.guice.GuiceApplication;
import com.cathive.fx.guice.GuiceFXMLLoader;
import com.google.inject.Module;
import com.google.inject.grapher.graphviz.GraphvizModule;
import javafx.application.Platform;
import javafx.collections.ObservableMap;
import javafx.scene.Scene;
import javafx.scene.control.ButtonType;
import javafx.scene.image.Image;
import javafx.stage.Stage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Provider;
import javax.xml.bind.DatatypeConverter;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.*;
import java.util.ArrayList;
import java.util.List;

/**
 * User: twalcari
 * Date: 10/31/13
 * Time: 6:17 PM
 */
public class Launch extends GuiceApplication implements Provider<Stage> {
    /* solution for MAC OS HeadlessException bug. See https://ibcn-jira.intec.ugent.be/browse/FEDIBBTDEV-239 */
    static {
        System.setProperty("java.awt.headless", "false");
    }

    private static final Logger LOG = LoggerFactory.getLogger(Launch.class);

    private static final String SCHEME_FILE = "file";


    @Inject
    private JFedLoggerManager jFedLoggerManager;
    @Inject
    private JFedGuiPreferences jFedPreferences;
    @Inject
    private JFedPreferencesApplier jFedPreferencesApplier;
    @Inject
    private UserLoginModelManager userLoginModelManager;
    @Inject
    private ObservableMap<GeniUrn, ExperimentViewController> experimentViewControllers;
    @Inject
    private TaskThread tt;
    @Inject
    private GuiceFXMLLoader loader;
    @Inject
    private PageantHelper pageantHelper;

    @Inject
    private PreferencesDialogFactory preferencesDialogFactory;

    private Stage primaryStage;
    @Inject
    private AcceleratorService acceleratorService;

    public static void main(String[] args) {
        launch(args);
    }

    @Override
    public void start(final Stage stage) throws Exception {
        this.primaryStage = stage;

        try {
            if (getSliceUrnParameter() != null) {
                LOG.info("Received slice-urn '{}' as parameter", getSliceUrnParameter());
            } else {
                LOG.info("Did not receive a slice-urn as parameter");
            }


            //check for filesystem access
            if (JFedUtils.getUserDataDirectory() == null) {

                JFDialogs.create().owner(stage).message("Please verify your security settings and try again.\n\n" +
                        "If you are trying to run jFed through Safari on OS X, you need to change your Safari security settings.\n" +
                        "(cfr. http://doc.fed4fire.eu/firstexperiment.html#start-up-jfed)")
                        .masthead("jFed was unable to access the filesystem.")
                        .title("Fatal Error").showError();

                exit();
                return;
            }

            //This should already be done (when the injected object was created), but just to be sure, we again update the Loggers here
            jFedLoggerManager.updateLoggers();

            ExtraInfoCallBackGUIs.prepareExtraInfoCallback(stage.getOwner());

            boolean loggedIn = loginWithParametersInformation();

            if (!loggedIn) {
                //does not wait! UserLoginControllerListener can be used to see if login window is closed
                showUserLogin(stage);
            } else {
                onSuccessfulLogin(stage);
            }
        } catch (Throwable ex) {
            LOG.error("Fatal error in jFed Experimenter Main Thread: " + ex.getMessage(), ex);
            JFDialogs.create()
                    .owner(stage)
                    .message("jFed experimenter will terminate.")
                    .masthead("A fatal error occurred!")
                    .title("Fatal Error")
                    .showException(ex);

            throw ex;
        }
    }

    private void onSuccessfulLogin(Stage stage) {
        try {
            //show information dialog about public/private keys on first run of jFed

            if (jFedPreferences.getInt(GuiPreferenceKey.PREF_FIRSTRUN_VERSION) < JFedPreferences.CURRENT_FIRSTRUN_VERSION) {

                if (OSDetector.findOS() == OSDetector.OS.WIN) {


                    JFDialogs.create()
                            .owner(stage)
                            .message("The default preferences should be okay. If you have trouble running experiments or connecting to nodes, consider to turn on 'Proxy' in the jFed Proxy preferences.\n\n" +
                                    "Next you will see the preferences window, you can always revisit this by clicking the preferences button.\n\n" +
                                    "Please verify that jFed has found the PuTTY SSH software (you should see 4 green checkmarks). If not install first PuTTY (full suite) or point jFed to the right PuTTY install directory.")
                            .masthead("Initial configuration")
                            .title("Welcome to the jFed Experimenter Tool!")
                            .showInformation();


                } else {

                    JFDialogs.create()
                            .owner(stage)
                            .message("The default preferences should be okay. If you have trouble running experiments or connecting to nodes, consider to turn on 'Proxy' in the jFed Proxy preferences.\n\n" +
                                    "Next you will see the preferences window, you can always revisit this by clicking the preferences button.")
                            .masthead("Initial configuration")
                            .title("Welcome to the jFed Experimenter Tool!")
                            .showInformation();

                }

                preferencesDialogFactory.showDialog();

                jFedPreferences.setInt(GuiPreferenceKey.PREF_FIRSTRUN_VERSION, JFedPreferences.CURRENT_FIRSTRUN_VERSION);

//                if (PreferencesUtil.getKeyLocation() == null) {
//                    Dialogs2.showWarningDialog(null, "Logging into nodes will be impossible until you correct this.",
//                            "You did not specify a valid private key");
//                }
            }

            jFedPreferencesApplier.applyPreferences();

            showExperimenterGui(stage);
        } catch (Throwable ex) {
            LOG.error("Fatal error in jFed Experimenter Main Thread: " + ex.getMessage(), ex);
            JFDialogs.create()
                    .owner(stage)
                    .message("jFed experimenter will terminate.")
                    .masthead("A fatal error occurred!")
                    .title("Fatal Error")
                    .showException(ex);

            throw ex;
        }
    }

    private void registerKeysWithPageant() {
        if (jFedPreferences.getBoolean(GuiPreferenceKey.PREF_PUTTY_PAGEANT, false)) {
            List<SshKeyInfo> keys = new ArrayList<>();

            SshKeyInfo sshKeyInfo = jFedPreferences.getOverriddenSshKeyInfo();
            if (sshKeyInfo == null && userLoginModelManager.isUserLoggedIn()) {
                GeniUser geniUser = userLoginModelManager.getLoggedInGeniUser();
                sshKeyInfo = new UserSshKeyInfo(geniUser);
            }
            //PageantHelper.registerKey(sshKeyInfo);
            keys.add(sshKeyInfo);

            if (!jFedPreferences.getString(GuiPreferenceKey.PREF_SSHPROXY_USE_FOR_SSH).equalsIgnoreCase("NEVER")) {
                SshKeyInfo proxySshKeyInfo = jFedPreferences.getOverriddenProxySshKeyInfo();
                if (proxySshKeyInfo == null && userLoginModelManager.isUserLoggedIn()) {
                    GeniUser geniUser = userLoginModelManager.getLoggedInGeniUser();
                    proxySshKeyInfo = new UserSshKeyInfo(geniUser);
                }
                //PageantHelper.registerKey(proxySshKeyInfo);
                keys.add(proxySshKeyInfo);
            }

            pageantHelper.registerKeys(keys);
        }
    }

    public StartupParameters getStartupParameters() {
        return new StartupParameters(getParameters());
    }

    /**
     * Checks if login-information was passed via the parameters, and if so: tries to load this information into jFed.
     *
     * @return success
     */
    private boolean loginWithParametersInformation() {
        StartupParameters startupParameters = getStartupParameters();

        boolean isPlanetLab = false;

        //check if we got login-information via the parameters
        boolean loginDefined = false;
        if (startupParameters.contains(StartupParameters.PARAM_LOGIN_CERTIFICATE)) {
            String loginCertificateLocation = startupParameters.getValue(StartupParameters.PARAM_LOGIN_CERTIFICATE);

            LOG.info("Found login-certificate in parameters. URI={}", loginCertificateLocation);

            URI uri;
            try {
                uri = new URI(loginCertificateLocation);
            } catch (URISyntaxException ex) {
                LOG.error("Could not parse provided login certificate location", ex);
                return false;
            }

            //try to update KeyCertUserLoginModel with the new information
            if (uri.getScheme().equals(SCHEME_FILE)) {
                File loginCertificateFile = new File(uri);

                //try to detect type of certificate
                try {
                    String fileContents = IOUtils.fileToString(loginCertificateFile);
                    isPlanetLab = fileContents.startsWith("fed4fire");
                } catch (IOException ex) {
                    LOG.warn("Could not detect the type of login file: error reading provided file", ex);
                }

                if (!isPlanetLab) {
                    try {
                        userLoginModelManager.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.KEY_CERT_INTERNAL_INFO);
                        userLoginModelManager.getKeyCertUserLoginModel().setKeyCertPemFile(loginCertificateFile);
                        userLoginModelManager.save();
                        loginDefined = true;
                    } catch (InvalidLoginException e) {
                        LOG.error("Could not properly load the provided key-certificate-pair into jFed", e);
                        return false;
                    }
                } else {
                    try {
                        userLoginModelManager.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.PLANETLAB);
                        userLoginModelManager.getPlanetlabUserLoginModel().setSshPrivateKeyFile(loginCertificateFile);
                        userLoginModelManager.save();
                        loginDefined = true;
                    } catch (InvalidLoginException e) {
                        LOG.error("Could not properly load the provided planetlab login file into jFed", e);
                        return false;
                    }
                }

            } else {
                //try to fetch from the internet
                URL url;
                try {
                    url = uri.toURL();
                } catch (MalformedURLException e) {
                    LOG.error("Got an illegal URL, cannot continue", e);
                    return false;
                }

                //try to detect the type of certificate
                try {
                    URLConnection conn = url.openConnection();
                    BufferedReader br = new BufferedReader(new InputStreamReader(conn.getInputStream()));

                    String firstLine = br.readLine();
                    br.close();

                    isPlanetLab = firstLine.startsWith("fed4fire");

                } catch (IOException ex) {
                    LOG.warn("Could not detect the type of login file: error reading provided url", ex);
                }

                if (!isPlanetLab) {
                    try {
                        userLoginModelManager.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.KEY_CERT_INTERNAL_INFO);
                        userLoginModelManager.getKeyCertUserLoginModel().setKeyCertPemUrl(url);
                        userLoginModelManager.save();
                        loginDefined = true;
                    } catch (InvalidLoginException e) {
                        LOG.error("Could not properly load the provided key-certificate-pair into jFed", e);
                        userLoginModelManager.load();
                        return false;
                    }
                } else {
                    try {
                        userLoginModelManager.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.PLANETLAB);
                        userLoginModelManager.getPlanetlabUserLoginModel().setSshPrivateKeyUrl(url);
                        userLoginModelManager.save();
                        loginDefined = true;
                    } catch (InvalidLoginException e) {
                        LOG.error("Could not properly load the provided planetlab-string into jFed", e);
                        userLoginModelManager.load();
                        return false;
                    }
                }
            }

        } else if (startupParameters.contains(StartupParameters.PARAM_LOGIN_CERTIFICATE_STRING)) {
            String loginCertificateString =
                    startupParameters.getValue(StartupParameters.PARAM_LOGIN_CERTIFICATE_STRING);

            LOG.info("Found login-certificate string in parameters. size={}", loginCertificateString.length());

            byte[] decodedLoginCertificate;
            try {
                decodedLoginCertificate = DatatypeConverter.parseBase64Binary(loginCertificateString);
            } catch (IllegalArgumentException ex) {
                LOG.warn("Could not Base64-decode login-certificate-string", ex);
                userLoginModelManager.load();
                return false;
            }


            String decodedLoginCertificateString = new String(decodedLoginCertificate);
            LOG.debug("Base64-decoded login certificate (size={})", decodedLoginCertificateString.length());

            isPlanetLab =  decodedLoginCertificateString.startsWith("fed4fire");

            if (isPlanetLab) {
                //this is a planetlab file!
                userLoginModelManager.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.PLANETLAB);

                try {
                    userLoginModelManager.getPlanetlabUserLoginModel().setSshPrivateKeyString(decodedLoginCertificateString);
                    loginDefined = true;
                } catch (InvalidLoginException e) {
                    LOG.error("Could not properly load the provided planetlab-string into jFed", e);
                    userLoginModelManager.load();
                    return false;
                }

            } else {
                //assume this is a key-cert-combo

                userLoginModelManager.setUserLoginModelType(UserLoginModelManager.UserLoginModelType.KEY_CERT_INTERNAL_INFO);
                try {
                    userLoginModelManager.getKeyCertUserLoginModel().setKeyCertPemString(decodedLoginCertificateString);
                    loginDefined = true;
                } catch (InvalidLoginException e) {
                    LOG.error("Could not properly load the provided key-certificate-pair-string into jFed", e);
                    userLoginModelManager.load();
                    return false;
                }
            }
        }

        if (loginDefined) {
            if (!isPlanetLab) {
                if (!userLoginModelManager.getKeyCertUserLoginModel().isPasswordRequired()) {
                    userLoginModelManager.login();
                    return true;
                } else if (startupParameters.contains(StartupParameters.PARAM_LOGIN_PASSWORD)) {
                    if (userLoginModelManager.getKeyCertUserLoginModel().unlock(startupParameters.getValue(StartupParameters.PARAM_LOGIN_PASSWORD))) {
                        userLoginModelManager.login();
                        return true;
                    } else {
                        LOG.debug("Could not login with provided password");

                        return false;
                    }
                } else {
                    //we need to show the login screen
                    return false;
                }
            } else {
                if (userLoginModelManager.getPlanetlabUserLoginModel().isReadyToFetchCertificate()) {
                    //first try to fetch the login certificate before continuing
                    try {
                        userLoginModelManager.getPlanetlabUserLoginModel().fetchCertificate();
                    } catch (PlanetlabUserLoginModel.CertificateFetchException ex) {
                        JFDialogs.create()
                                .owner(null)
                                .message("jFed could not fetch the user certificate from the authority. Contact support for more information.")
                                .masthead("Login Failed")
                                .title("Login Failed")
                                .showException(ex);
                        return false;
                    }

                    if (!userLoginModelManager.getPlanetlabUserLoginModel().isPasswordRequired()) {
                        userLoginModelManager.login();
                        return true;
                    } else if (startupParameters.contains(StartupParameters.PARAM_LOGIN_PASSWORD)) {
                        if (userLoginModelManager.getPlanetlabUserLoginModel().unlock(startupParameters.getValue(StartupParameters.PARAM_LOGIN_PASSWORD))) {
                            userLoginModelManager.login();
                            return true;
                        } else {
                            LOG.debug("Could not login with provided password");

                            return false;
                        }
                    } else {
                        //we need to show the login screen
                        return false;
                    }

                } else {
                    return false;
                }
            }
        } else {
            return false;
        }

    }

    public String getSliceUrnParameter() {
        return getStartupParameters().getValue(StartupParameters.PARAM_SLICE_URN);
    }


    /**
     * doesn't wait
     */
    public void showUserLogin(final Stage parentStage) {

        UserLoginController.UserLoginControllerListener userLoginControllerListener = successfulLogin -> {
            if (successfulLogin)
                onSuccessfulLogin(parentStage);
            else {
                exit();
            }
        };

        try {
            URL location = UserLoginController.class.getResource("UserLogin.fxml");

            GuiceFXMLLoader.Result result = loader.load(location);

            final UserLoginController userLoginController = result.getController();
            userLoginController.addListener(userLoginControllerListener);

            Stage userLoginStage = new Stage();
            userLoginController.setupStage(userLoginStage);

            userLoginStage.setTitle("jFed login");
            userLoginStage.setResizable(false);
            userLoginStage.setScene(new Scene(result.getRoot()));
            userLoginStage.getIcons().addAll(new Image(UserLoginController.class.getResource("/images/jfed-logo_16.png").toExternalForm())
                    , new Image(UserLoginController.class.getResource("/images/jfed-logo_32.png").toExternalForm())
                    , new Image(UserLoginController.class.getResource("/images/jfed-logo_64.png").toExternalForm()));


            //don't wait!
            userLoginStage.show();
        } catch (Exception e) {
            throw new RuntimeException("Something went wrong showing the User Login Screen: " + e.getMessage(), e);
        }
    }


    private boolean isExperimentActive(Experiment experiment) {
        if (experiment.getSlice() == null)
            return false;

        for (Sliver sliver : experiment.getSlice().getSlivers()) {
            switch (sliver.getStatus().getGlobalStatus()) {
                case READY:
                case UNKNOWN:
                case CHANGING:
                    return true;
            }
        }
        return false;
    }


    public void showExperimenterGui(Stage stage) {
        try {
            GuiceFXMLLoader.Result result = loader.load(getClass().getResource("ExperimenterGUI.fxml"));
            ExperimenterGUI experimenterGUI = result.getController();


            Scene scene = new Scene(result.getRoot());

            stage.setScene(scene);
            stage.setMinHeight(600);
            stage.setMinWidth(800);

            stage.setTitle("jFed Experimenter Toolkit");
            stage.getIcons().addAll(new Image(Launch.class.getResource("/images/jfed-logo_16.png").toExternalForm())
                    , new Image(Launch.class.getResource("/images/jfed-logo_32.png").toExternalForm())
                    , new Image(Launch.class.getResource("/images/jfed-logo_64.png").toExternalForm()));

            stage.setOnCloseRequest(windowEvent -> {
                boolean hasActiveSlices = experimentViewControllers.values().stream()
                        .map(ExperimentViewController::getExperiment)
                        .anyMatch(this::isExperimentActive);

                if (hasActiveSlices) {
                    //show a warning
                    JFDialogs.create()
                            .message("There are currently resources in use.\nAre you sure you want to exit?")
                            .title("Exit while resources in use").buttonTypes(ButtonType.YES, ButtonType.NO, ButtonType.CANCEL)
                            .showWarning()
                            .filter(response -> response != ButtonType.YES)
                            .ifPresent(response -> windowEvent.consume());

                    if (windowEvent.isConsumed()) {
                        return;
                    }

                }

                exit();
            });

            //Called just after the Window is shown.
            stage.setOnShown(windowEvent -> registerKeysWithPageant());


            acceleratorService.registerOnScene(scene);

            stage.show();
        } catch (IOException ex) {
            LOG.error("Error while loading ExperimenterGUI", ex);
            throw new RuntimeException("Error while loading ExperimenterGUI", ex);
        }
    }

    public void exit() {
        LOG.info("Close was requested. Trying to terminate");
        tt.requestStop();

        //stop any running ssh-agent
        SshAgentHelper.stopSshAgent();

        Platform.exit();
        System.exit(0);
    }

    @Override
    public void init(List<Module> modules) throws Exception {
        modules.add(new LowLevelModule());
        modules.add(new HighLevelModule());
        modules.add(new UserLoginModelModule());
        modules.add(new ExperimenterModule());
        modules.add(new BaseGuiModule());
        modules.add(new GraphvizModule());
    }

    @Override
    public Stage get() {
        return primaryStage;
    }
}
