package be.iminds.ilabt.jfed.experimenter_gui.canvas;

import be.iminds.ilabt.jfed.experimenter_gui.JFedConfiguration;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.impl.CanvasNodeFactory;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.impl.RspecCanvasLink;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.impl.RspecCanvasNode;
import be.iminds.ilabt.jfed.experimenter_gui.canvas.impl.RspecChannelCanvasNode;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ExperimentEditor;
import be.iminds.ilabt.jfed.experimenter_gui.editor.ModelRspecEditor;
import be.iminds.ilabt.jfed.experimenter_gui.editor.NewCanvasLink;
import be.iminds.ilabt.jfed.experimenter_gui.editor.channel_properties.ChannelPropertiesDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.editor.impl.ChannelToolboxItem;
import be.iminds.ilabt.jfed.experimenter_gui.editor.impl.NodeDescriptionToolboxItem;
import be.iminds.ilabt.jfed.experimenter_gui.editor.link_properties.LinkPropertiesDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.AdaptivePropertiesDialogFactory;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.DelayNodePropertiesDialog;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityList;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.*;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import javafx.beans.InvalidationListener;
import javafx.beans.Observable;
import javafx.beans.property.ReadOnlyListProperty;
import javafx.beans.property.ReadOnlyListWrapper;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.geometry.Bounds;
import javafx.geometry.Point2D;
import javafx.scene.control.ButtonType;
import javafx.scene.control.ContextMenu;
import javafx.scene.control.MenuItem;
import javafx.scene.input.*;
import org.controlsfx.glyphfont.FontAwesome;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

/**
 * User: twalcari
 * Date: 11/18/13
 * Time: 5:11 PM
 */
public class EditableExperimentCanvas extends ExperimentCanvas {
    private static final Logger LOG = LoggerFactory.getLogger(EditableExperimentCanvas.class);
    private static final double EDITOR_LOCATION_SHIFT = 15;

    private final ModelRspecEditor modelRspecEditor;

    private final ObservableList<CanvasNode> clipboard = FXCollections.observableArrayList();

    private final AuthorityList authorityList;
    private final AdaptivePropertiesDialogFactory adaptivePropertiesDialogFactory;

    private final NewCanvasLink newCanvasLink;
    private final ContextMenu nodeContextMenu;
    private final ContextMenu linkContextMenu;
    private final ContextMenu linkInterfaceContextMenu;

    private final ContextMenu channelContextMenu;

    private final ExperimentEditor experimentEditor;

    private final FXRspecFactory rspecFactory = FXRspecFactory.getInstance();
    private final LinkPropertiesDialogFactory linkPropertiesDialogFactory;
    private final ChannelPropertiesDialogFactory channelPropertiesDialogFactory;

    public EditableExperimentCanvas(ExperimentEditor experimentEditor, final ModelRspecEditor modelRspecEditor,
                                    JFedConfiguration config, CanvasNodeFactory canvasNodeFactory, AuthorityList authorityList,
                                    AdaptivePropertiesDialogFactory adaptivePropertiesDialogFactory,
                                    LinkPropertiesDialogFactory linkPropertiesDialogFactory,
                                    ChannelPropertiesDialogFactory channelPropertiesDialogFactory) {
        super(modelRspecEditor.getModelRspec(), config, canvasNodeFactory);
        this.experimentEditor = experimentEditor;
        this.modelRspecEditor = modelRspecEditor;
        this.authorityList = authorityList;
        this.adaptivePropertiesDialogFactory = adaptivePropertiesDialogFactory;
        this.linkPropertiesDialogFactory = linkPropertiesDialogFactory;
        this.channelPropertiesDialogFactory = channelPropertiesDialogFactory;

        this.newCanvasLink = new NewCanvasLink(this, authorityList);

        this.setOnMouseDragReleased(this::onMouseDragReleased);


        // create context menus
        this.nodeContextMenu = new ContextMenu();

        MenuItem configureItem = new MenuItem("Configure node", GlyphUtils.createSmallGlyph(FontAwesome.Glyph.PENCIL));
        configureItem.setOnAction(actionEvent -> {
            assert (getSelectionProvider().getSelectedCanvasNode() instanceof RspecCanvasNode);
            RspecCanvasNode rspecCanvasNode = (RspecCanvasNode) getSelectionProvider().getSelectedCanvasNode();

            onConfigureNodePressed(rspecCanvasNode);
        });
        nodeContextMenu.getItems().add(configureItem);

        MenuItem deleteItem = new MenuItem("Delete", GlyphUtils.createSmallGlyph(FontAwesome.Glyph.TIMES));
        deleteItem.setOnAction(actionEvent -> {
            assert (getSelectionProvider().getSelectedCanvasNode() instanceof RspecCanvasNode);
            RspecCanvasNode rspecCanvasNode = (RspecCanvasNode) getSelectionProvider().getSelectedCanvasNode();

            onDeleteNodePressed(rspecCanvasNode);
        });
        nodeContextMenu.getItems().add(deleteItem);

        //contextMenu for links
        this.linkContextMenu = new ContextMenu();

        //contextMenu for interfaces
        this.linkInterfaceContextMenu = new ContextMenu();

        // was disabled as FXRspecLink.LinkSetting is currently not supported in practice, but it seems to work for VM's now
        MenuItem configureLinkItem =
                new MenuItem("Configure link", GlyphUtils.createSmallGlyph(FontAwesome.Glyph.PENCIL));
        configureLinkItem.setOnAction(actionEvent -> {
            RspecCanvasLink rspecCanvasLink = getSelectionProvider().getSelectedCanvasLink();

            onConfigureLinkPressed(rspecCanvasLink.getRspecLink());
        });
        linkContextMenu.getItems().add(configureLinkItem);

        MenuItem impairLinkItem =
                new MenuItem("Add impairment bridge to link", GlyphUtils.createSmallGlyph(FontAwesome.Glyph.EXPAND));
        impairLinkItem.setOnAction(actionEvent -> {
            RspecCanvasLink rspecCanvasLink = getSelectionProvider().getSelectedCanvasLink();

            convertLinkToImpairmentBridge(rspecCanvasLink.getRspecLink());
        });
        linkContextMenu.getItems().add(impairLinkItem);

        MenuItem linkDeleteItem = new MenuItem("Delete", GlyphUtils.createSmallGlyph(FontAwesome.Glyph.TIMES));
        linkDeleteItem.setOnAction(actionEvent -> {
            RspecCanvasLink rspecCanvasLink = getSelectionProvider().getSelectedCanvasLink();

            onDeleteLinkPressed(rspecCanvasLink);
        });
        linkContextMenu.getItems().add(linkDeleteItem);


        MenuItem linkInterfaceDeleteItem = new MenuItem("Delete", GlyphUtils.createSmallGlyph(FontAwesome.Glyph.TIMES));
        linkInterfaceDeleteItem.setOnAction(actionEvent -> {
            RspecCanvasLink.InterfaceLink ifaceLink = getSelectionProvider().getSelectedInterfaceLink();
            assert (ifaceLink != null);

            onDeleteLinkInterfacePressed(ifaceLink);
        });
        linkInterfaceContextMenu.getItems().add(linkInterfaceDeleteItem);


        channelContextMenu = new ContextMenu();
        MenuItem configureChannelMenuItem =
                new MenuItem("Configure channel", GlyphUtils.createSmallGlyph(FontAwesome.Glyph.PENCIL));
        configureChannelMenuItem.setOnAction(actionEvent -> {
            RspecChannelCanvasNode rspecChannelCanvasNode = getSelectionProvider().getSelectedChannel();

            onConfigureChannelPressed(rspecChannelCanvasNode.getChannel());
        });
        channelContextMenu.getItems().add(configureChannelMenuItem);

        MenuItem channelDeleteItem = new MenuItem("Delete", GlyphUtils.createSmallGlyph(FontAwesome.Glyph.TIMES));
        channelDeleteItem.setOnAction(actionEvent -> {
            RspecChannelCanvasNode rspecChannelCanvasNode = getSelectionProvider().getSelectedChannel();

            onDeleteChannelPressed(rspecChannelCanvasNode);
        });
        channelContextMenu.getItems().add(channelDeleteItem);


        //make sure that the initialize-function is called when the scene-property is set
        sceneProperty().addListener(new InvalidationListener() {
            @Override
            public void invalidated(Observable observable) {
                if (getScene() != null) {
                    initialize();
                    sceneProperty().removeListener(this);
                }
            }
        });


    }

    private void onDeleteChannelPressed(RspecChannelCanvasNode rspecChannelCanvasNode) {
        model.deleteChannel(rspecChannelCanvasNode.getChannel());
        if (getSelectionProvider().getSelectedChannel() == rspecChannelCanvasNode) {
            getSelectionProvider().clearAllSelections();
        }
    }

    private void onConfigureChannelPressed(FXRspecChannel channel) {
        channelPropertiesDialogFactory.showChannelPropertiesDialog(getScene().getWindow(), channel, getModelRspec());

    }

    @FXML
    private void initialize() {
        //add mnemorics
        getScene().getAccelerators().put(new KeyCodeCombination(KeyCode.DELETE), () -> {
            if (getSelectionProvider().getSelectedCanvasNode() != null) {
                CanvasNode canvasNode = getSelectionProvider().getSelectedCanvasNode();
                if (canvasNode instanceof RspecCanvasNode) {
                    onDeleteNodePressed((RspecCanvasNode) canvasNode);
                }
            } else if (getSelectionProvider().getSelectedCanvasLink() != null) {
                RspecCanvasLink rspecCanvasLink = getSelectionProvider().getSelectedCanvasLink();
                onDeleteLinkPressed(rspecCanvasLink);
            } else if (getSelectionProvider().getSelectedInterfaceLink() != null) {
                onDeleteLinkInterfacePressed(getSelectionProvider().getSelectedInterfaceLink());
            } else if (getSelectionProvider().getSelectedChannel() != null) {
                onDeleteChannelPressed(getSelectionProvider().getSelectedChannel());
            }
        });
    }

    private void convertLinkToImpairmentBridge(FXRspecLink rspecLink) {
        //check if convertion is possible
        if (rspecLink.getInterfaces().size() > 2) {
            JFDialogs.create()
                    .owner(this)
                    .message("An impairment bridge can only be added on a link between two nodes!")
                    .masthead("Unsupported operation")
                    .showWarning();
            return;
        }


        DelayRspecNode delayRspecNode = modelRspecEditor.convertLinkToImpairmentBridge(rspecLink);

        //show configuration-window for impairmentBridgeNode now
        showNodeConfigurationWindow(delayRspecNode);
    }

    private void showNodeConfigurationWindow(FXRspecNode node) {
        if (node instanceof DelayRspecNode) {
            DelayNodePropertiesDialog.showNodePropertiesDialog(this.getScene().getWindow(),
                    this, (DelayRspecNode) node);
        } else
            adaptivePropertiesDialogFactory.showAdaptivePropertiesDialog(this.getScene().getWindow(),
                    modelRspecEditor, node);
    }


    private void onDeleteLinkInterfacePressed(RspecCanvasLink.InterfaceLink ifaceLink) {

        if (ifaceLink.getRspecInterface().getLink().getInterfaces().size() <= 2) {
            model.deleteLink(ifaceLink.getRspecInterface().getLink());
        } else
            ifaceLink.getRspecInterface().delete();

        if (getSelectionProvider().getSelectedInterfaceLink() == ifaceLink) {
            getSelectionProvider().clearAllSelections();
        }
    }

    private void onConfigureNodePressed(RspecCanvasNode rspecCanvasNode) {
        showNodeConfigurationWindow(rspecCanvasNode.getRspecNode());
    }

    private void onConfigureLinkPressed(FXRspecLink rspecLink) {
        linkPropertiesDialogFactory.showLinkPropertiesDialog(this.getScene().getWindow(),
                modelRspecEditor, rspecLink);
    }


    private void onDeleteNodePressed(RspecCanvasNode rspecCanvasNode) {
        model.deleteNode(rspecCanvasNode.getRspecNode());
        if (getSelectionProvider().getSelectedCanvasNode() == rspecCanvasNode) {
            getSelectionProvider().clearAllSelections();
        }

    }

    private void onDeleteLinkPressed(RspecCanvasLink rspecCanvasLink) {
        model.deleteLink(rspecCanvasLink.getRspecLink());
        if (getSelectionProvider().getSelectedCanvasLink() == rspecCanvasLink) {
            getSelectionProvider().clearAllSelections();
        }
    }

    @Override
    protected void canvasItemOnMouseClicked(MouseEvent mouseEvent) {
        super.canvasItemOnMouseClicked(mouseEvent);

        if (mouseEvent.getSource() instanceof RspecCanvasNode) {
            RspecCanvasNode node = (RspecCanvasNode) mouseEvent.getSource();
            getSelectionProvider().setSelectedCanvasNode(node);

            if (mouseEvent.getButton() == MouseButton.PRIMARY && mouseEvent.getClickCount() == 2) {
                onConfigureNodePressed(node);

            } else if (mouseEvent.getButton() != MouseButton.PRIMARY || mouseEvent.isControlDown()) {
                //show context menu
                nodeContextMenu.show(node, mouseEvent.getScreenX(), mouseEvent.getScreenY());

            }
        }

        if (mouseEvent.getSource() instanceof RspecCanvasLink.LinkCenter) {
            RspecCanvasLink link =
                    ((RspecCanvasLink.LinkCenter) mouseEvent.getSource()).getRspecCanvasLink();


            getSelectionProvider().setSelectedCanvasLink(link);
            assert (getSelectionProvider().getSelectedCanvasLink() != null);

            if (mouseEvent.getButton() == MouseButton.PRIMARY && mouseEvent.getClickCount() == 2) {
                onConfigureLinkPressed(link.getRspecLink());
            } else if (mouseEvent.getButton() != MouseButton.PRIMARY || mouseEvent.isControlDown()) {
                //show context menu
                linkContextMenu.show(getSelectionProvider().getSelectedCanvasLink().getLinkCenter(),
                        mouseEvent.getScreenX(), mouseEvent.getScreenY());
            }
        }

        if (mouseEvent.getSource() instanceof RspecCanvasLink.InterfaceLink) {
            if (mouseEvent.getButton() != MouseButton.PRIMARY || mouseEvent.isControlDown()) {

                getSelectionProvider().setSelectedInterfaceLink((RspecCanvasLink.InterfaceLink) mouseEvent.getSource());

                //show context menu
                linkInterfaceContextMenu.show(getSelectionProvider().getSelectedInterfaceLink(),
                        mouseEvent.getScreenX(), mouseEvent.getScreenY());
            }
        }

        if (mouseEvent.getSource() instanceof RspecChannelCanvasNode) {
            final RspecChannelCanvasNode channelCanvasNode = (RspecChannelCanvasNode) mouseEvent.getSource();
            getSelectionProvider().setSelectedChannel(channelCanvasNode);

            if (mouseEvent.getButton() == MouseButton.PRIMARY && mouseEvent.getClickCount() == 2) {
                onConfigureChannelPressed(channelCanvasNode.getChannel());
            } else if (mouseEvent.getButton() != MouseButton.PRIMARY || mouseEvent.isControlDown()) {
                //show context menu
                channelContextMenu.show(getSelectionProvider().getSelectedChannel(),
                        mouseEvent.getScreenX(), mouseEvent.getScreenY());
            }
        }

    }

    private void setStatus(String status) {
        experimentEditor.setStatus(status);
    }

    @Override
    protected void canvasNodeOnDragDetected(MouseEvent mouseEvent) {
        super.canvasNodeOnDragDetected(mouseEvent);

        //make canvas link invisible when starting to drag
        newCanvasLink.setVisible(false);
    }

    @Override
    protected void canvasNodeOnMouseDragged(MouseEvent mouseEvent) {
        CanvasNode node = (CanvasNode) mouseEvent.getSource();
        Bounds canvasOrigin = getCanvas().
                localToScene(getCanvas().getBoundsInLocal());
        double newX = mouseEvent.getSceneX()
                - canvasOrigin.getMinX()
                - node.getBoundsInParent().getWidth() / 2;
        double newY = mouseEvent.getSceneY()
                - canvasOrigin.getMinY()
                - node.getBoundsInParent().getHeight() / 2;

        //setSecondaryStatus("Moving to " + newX + ", " + newY);

        node.relocate(newX, newY);

        mouseEvent.consume();
    }

    private boolean testForTestbedLinkSupport(FXRspecNode node) {
        if (node.getComponentManagerId() == null) {
            LOG.warn("ComponentManagerId is null for node '{}'", node.getUniqueId());
            return true;
        }

        JFedConfiguration.TestbedDescription testbedDescription =
                config.getTestbedDescription(node.getComponentManagerId());

        if (testbedDescription != null && !testbedDescription.hasLinkSupport()) {
            JFDialogs.create()
                    .owner(this)
                    .message(String.format("The testbed '%s' does not support defining dedicated links.",
                            authorityList.finder().findByUrn(node.getComponentManagerId(), AuthorityFinder.Purpose.CREATE_SLIVER).getHrn()
                    ))
                    .masthead("Requested link cannot be added")
                    .showError();
            return false;
        } else
            return true;
    }


    private boolean processNewLinkRequest() {
        LOG.trace("Received new link request between {} or {} AND {} {}", newCanvasLink.getOriginNode(), newCanvasLink.getOriginLink(),
                newCanvasLink.getTargetNode(), newCanvasLink.getTargetLink());


        //check if new link may be created

        if (!newCanvasLink.isLegalLink()) {
            setStatus("The requested link is illegal and was ignored.");
            LOG.trace("Link is not legal, and is thus rejected.");
            return false;
        }

        if (newCanvasLink.getOriginNode() != null) {
            if (!testForTestbedLinkSupport(newCanvasLink.getOriginNode().getRspecNode())) {
                LOG.trace("Testbed of origin node does not support links. Ignoring!");
                return false;
            }
        }
        if (newCanvasLink.getTargetNode() != null) {
            if (!testForTestbedLinkSupport(newCanvasLink.getTargetNode().getRspecNode())) {
                LOG.trace("Testbed of target node does not support links. Ignoring!");
                return false;
            }
        }

        //warn for stitched link if necessary

        if (newCanvasLink.isStitchedLink()) {
            Optional<ButtonType> reply = JFDialogs.create()
                    .title("Experimental functionality detected.")
                    .message("Stitched links between testbeds are currently VERY experimental and not supported. Are you sure you want to continue?")
                    .showConfirm();

            if (!reply.isPresent() && reply.get() != ButtonType.YES)
                return false;
        }

        //delegate the real link-creation to ModelRspecEditor
        LOG.trace("Received link request is OK!");

        if (newCanvasLink.getOriginNode() != null && newCanvasLink.getTargetNode() != null) {
            FXRspecLink newLink = modelRspecEditor.addLink(newCanvasLink.getOriginNode().getRspecNode(),
                    newCanvasLink.getTargetNode().getRspecNode());

            setStatus("Created a new link '" + newLink.getClientId() + "'");

        } else if (newCanvasLink.getOriginNode() != null) {
            assert newCanvasLink.getTargetLink() != null;
            boolean success = modelRspecEditor.addLink(newCanvasLink.getTargetLink().getRspecLink(),
                    newCanvasLink.getOriginNode().getRspecNode());

            if (success) {
                setStatus(String.format("Added node %s to existing link %s",
                        newCanvasLink.getOriginNode().getRspecNode().getClientId(),
                        newCanvasLink.getTargetLink().getRspecLink().getClientId()));
            }

        } else {
            assert newCanvasLink.getOriginLink() != null;
            assert newCanvasLink.getTargetNode() != null;
            boolean success = modelRspecEditor.addLink(newCanvasLink.getOriginLink().getRspecLink(),
                    newCanvasLink.getTargetNode().getRspecNode());

            if (success) {
                setStatus(String.format("Added node %s to existing link %s",
                        newCanvasLink.getTargetNode().getRspecNode().getClientId(),
                        newCanvasLink.getOriginLink().getRspecLink().getClientId()));
            }
        }

        return true;
    }

    @Override
    protected void onCanvasMouseReleased(MouseEvent mouseEvent) {
        super.onCanvasMouseReleased(mouseEvent);

        //have we made a new connection?
        if (newCanvasLink.isLinkComplete()) {
            processNewLinkRequest();
        }
        newCanvasLink.clear();
        newCanvasLink.setVisible(false);
    }

    @Override
    protected void onCanvasMouseMoved(MouseEvent mouseEvent) {
        if (!mouseEvent.isPrimaryButtonDown()) {

            //if the mouse is close nearby a node, show a "newCanvasLink"

            RspecCanvasNode nearbyNode = findNearbyNode(mouseEvent.getX(), mouseEvent.getY());
            RspecCanvasLink nearbyLink = null;
            if (nearbyNode != null) {
                newCanvasLink.setOriginNode(nearbyNode);
            } else {
                nearbyLink = findNearbyLink(mouseEvent.getX(), mouseEvent.getY());
                if (nearbyLink != null) {
                    newCanvasLink.setOriginLink(nearbyLink);
                }
            }

            if (nearbyNode != null || nearbyLink != null) {
                newCanvasLink.setNewEnd(mouseEvent.getX(), mouseEvent.getY());
                newCanvasLink.setVisible(true);

                setStatus("Click and drag the circle to a target node to create a new link.");
            } else {

                newCanvasLink.setVisible(false);
            }
        }
    }

    /**
     * Performs panning of the canvas when the newCanvasLink is not visible, or when the shift-button is down.
     * Otherwise, it tries to connect two nodes
     */
    protected void onCanvasMouseDragged(MouseEvent mouseEvent) {
        if (mouseEvent.isShiftDown() || !newCanvasLink.isVisible()) {
            super.onCanvasMouseDragged(mouseEvent);
        } else {

            //try to connect this node if canvas isn't fixed
            RspecCanvasNode targetNode = findNearbyNode(mouseEvent.getX(), mouseEvent.getY());

            if (targetNode != null && targetNode != newCanvasLink.getOriginNode()) {
                newCanvasLink.setTargetNode(targetNode);
            } else {

                RspecCanvasLink targetLink = findNearbyLink(mouseEvent.getX(), mouseEvent.getY());

                if (targetLink != null) {
                    newCanvasLink.setTargetLink(targetLink);
                } else {
                    newCanvasLink.setNewEnd(mouseEvent.getX(), mouseEvent.getY());
                }
            }
        }

    }

    public void duplicateSelected() {

        if (getSelectionProvider().getSelectedCanvasNode() != null &&
                getSelectionProvider().getSelectedCanvasNode() instanceof RspecCanvasNode) {
            RspecCanvasNode rspecCanvasNode = (RspecCanvasNode) getSelectionProvider().getSelectedCanvasNode();

            FXRspecNode newRspecNode =
                    rspecFactory.copyNode(model, rspecCanvasNode.getRspecNode(), RspecNode.InterfaceCopyMethod.NO_COPY);
            newRspecNode.setClientId(model.nextNodeName());
            newRspecNode.setEditorX(newRspecNode.getEditorX() + EDITOR_LOCATION_SHIFT);
            newRspecNode.setEditorY(newRspecNode.getEditorY() + EDITOR_LOCATION_SHIFT);

            model.getNodes().add(newRspecNode);

            //just add a new interface, never mind about making an exact copy.
            //todo impairment could be copied if useful. (ip probably shouldn't be copied, clientId certainly shouldn't be copied)
            rspecCanvasNode.getRspecNode().getInterfaces().stream()
                    .filter(FXRspecInterface::isLinkBound)
                    .forEach(rspecInterface -> {

                        if (rspecInterface.isLinkBound()) {

                            LOG.info("Duplicating interface to link {}", rspecInterface.getLink().getUniqueId());
                            modelRspecEditor.addLink(rspecInterface.getLink(), newRspecNode);
                        } else {
                            LOG.warn("Cannot copy interface on duplication, as it is not bound!");
                        }
                    });

            RspecCanvasNode newRspecCanvasNode = getRspecCanvasNode(newRspecNode);
            getSelectionProvider().setSelectedCanvasNode(newRspecCanvasNode);
        } else {
            JFDialogs.create()
                    .owner(this)
                    .message("Nothing to duplicate!\r\nPlease select an item first.")
                    .masthead("Duplicate action failed")
                    .showWarning();

        }
    }

    @FXML
    public void onMouseDragReleased(MouseDragEvent mouseDragEvent) {
        //we are adding a new element from the toolbox
        assert (experimentEditor.getDraggedItem() != null);

        Point2D location = new Point2D(
                mouseDragEvent.getSceneX()
                        - getCanvas().getLocalToSceneTransform().getTx()
                        - experimentEditor.getDraggedItem().getPrefWidth() / 2,
                mouseDragEvent.getSceneY()
                        - getCanvas().getLocalToSceneTransform().getTy()
                        - experimentEditor.getDraggedItem().getPrefHeight() / 2
        );

        if (experimentEditor.getDraggedItem() instanceof NodeDescriptionToolboxItem) {
            NodeDescriptionToolboxItem ndti = (NodeDescriptionToolboxItem) experimentEditor.getDraggedItem();
            FXRspecNode rspecNode = modelRspecEditor.addNode(ndti.getNodeDescription(), location);

            setStatus("Added node " + rspecNode.getClientId() + " to the canvas");

            setCursor(null);

            experimentEditor.clearDraggedItem();
            mouseDragEvent.consume();
        } else if (experimentEditor.getDraggedItem() instanceof ChannelToolboxItem) {
            ChannelToolboxItem cti = (ChannelToolboxItem) experimentEditor.getDraggedItem();

            FXRspecChannel channel = modelRspecEditor.addChannel(location);

            boolean success =
                    channelPropertiesDialogFactory.showChannelPropertiesDialog(getScene().getWindow(), channel, getModelRspec());

            if (!success) {
                getModelRspec().deleteChannel(channel);
                setStatus("Channel was not properly configured!");
            } else {
                setStatus("Added channel " + channel.getComponentName() + " to the canvas");
            }

            experimentEditor.clearDraggedItem();
            mouseDragEvent.consume();

        } else {
            throw new RuntimeException("Unexpected dragged item type: " + experimentEditor.getDraggedItem().getClass().getName());
        }

    }

    public void copySelected() {
        clipboard.clear();

        if (getSelectionProvider().getSelectedCanvasNode() == null) {
            JFDialogs.create()
                    .owner(this)
                    .message("Nothing to copy!\r\nPlease select an item first.")
                    .masthead("Copy action failed")
                    .showWarning();
        }

        clipboard.add(getSelectionProvider().getSelectedCanvasNode());
    }

    public void paste() {
        if (clipboard.isEmpty()) {
            JFDialogs.create()
                    .owner(this)
                    .message("Nothing to paste!\r\nPlease copy an item first.")
                    .masthead("Paste action failed")
                    .showWarning();
        }

        /* If the user pastes the same node two times, we want it to appear on two different places.
           Therefore we modify the clipboard to the newly created items    */
        List<CanvasNode> newClipboard = new ArrayList<>();

        //we now know that a RSpecCanvasNode has been created, and we can thus safely create the interfaces
        clipboard.stream()
                .filter(canvasNode -> canvasNode instanceof RspecCanvasNode)
                .forEach(canvasNode -> {
                    RspecCanvasNode rspecCanvasNode = (RspecCanvasNode) canvasNode;

                    FXRspecNode newRspecNode =
                            rspecFactory.copyNode(model, rspecCanvasNode.getRspecNode(), RspecNode.InterfaceCopyMethod.NO_COPY);
                    newRspecNode.setClientId(model.nextNodeName());
                    newRspecNode.setEditorX(newRspecNode.getEditorX() + EDITOR_LOCATION_SHIFT);
                    newRspecNode.setEditorY(newRspecNode.getEditorY() + EDITOR_LOCATION_SHIFT);

                    model.getNodes().add(newRspecNode);

                    //we now know that a RSpecCanvasNode has been created, and we can thus safely create the interfaces
                    newRspecNode.copyInterfaces(rspecCanvasNode.getRspecNode());

                    newClipboard.add(getRspecCanvasNode(newRspecNode));
                });

        clipboard.clear();
        clipboard.addAll(newClipboard);
    }

    public ReadOnlyListProperty<CanvasNode> getClipboard() {
        return new ReadOnlyListWrapper<>(clipboard).getReadOnlyProperty();
    }
}
