package be.iminds.ilabt.jfed.experimenter_gui.dialogs;

import be.iminds.ilabt.jfed.experimenter_gui.slice.ExperimentController;
import be.iminds.ilabt.jfed.experimenter_gui.slice.tasks.ExperimentTasksFactory;
import be.iminds.ilabt.jfed.experimenter_gui.util.DateTimeUtils;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;
import javafx.animation.Animation;
import javafx.animation.KeyFrame;
import javafx.animation.Timeline;
import javafx.beans.binding.DoubleBinding;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.beans.property.SimpleStringProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.control.*;
import javafx.scene.control.cell.CheckBoxTableCell;
import javafx.scene.control.cell.PropertyValueFactory;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.stage.Window;
import javafx.util.Callback;
import javafx.util.StringConverter;
import org.controlsfx.control.ButtonBar;
import org.controlsfx.dialog.Dialog;
import org.controlsfx.dialog.DialogAction;
import org.controlsfx.glyphfont.FontAwesome;

import java.io.IOException;
import java.net.URL;
import java.time.*;
import java.util.Collection;
import java.util.List;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

/**
 * User: twalcari
 * Date: 2/24/14
 * Time: 4:03 PM
 */
public class RenewSliceDialog extends VBox implements Consumer<ActionEvent> {
    private static final String RENEW_SLICE_DIALOG_FXML = "RenewSliceDialog.fxml";

    private final Window parentWindow;

    private final Dialog dialog;
    private final Slice slice;
    private final ExperimentController experimentController;
    private final ExperimentTasksFactory experimentTasksFactory;

    @FXML
    private TableView<RenewableAuthority> authoritiesTableView;

    @FXML
    private TableColumn<RenewableAuthority, Boolean> selectionTableColumn;
    @FXML
    private TableColumn<RenewableAuthority, String> nameTableColumn;
    @FXML
    private TableColumn<RenewableAuthority, Instant> expirationTableColumn;
    @FXML
    private TableColumn<RenewableAuthority, Instant> durationTableColumn;

    @FXML
    protected ComboBox<Integer> hoursComboBox;
    @FXML
    protected ComboBox<Integer> minutesComboBox;
    @FXML
    protected ComboBox<Integer> daysComboBox;

    @FXML
    private Label sliceInfoLabel;

    @FXML
    private Button cancelButton;

    private final ObservableList<RenewableAuthority> renewableAuthorities;

    public static class RenewableAuthority {

        private final SfaAuthority authority;
        private final Collection<Sliver> slivers;
        private final ObjectProperty<Instant> expirationTime;

        private BooleanProperty renewRequested = new SimpleBooleanProperty(true);

        private RenewableAuthority(SfaAuthority authority, Collection<Sliver> slivers, ObjectProperty<Instant> expirationTime) {
            this.authority = authority;
            this.slivers = slivers;
            this.expirationTime = expirationTime;
        }

        public SfaAuthority getAuthority() {
            return authority;
        }

        public Collection<Sliver> getSlivers() {
            return slivers;
        }

        public Instant getExpirationTime() {
            return expirationTime.get();
        }

        public ObjectProperty<Instant> expirationTimeProperty() {
            return expirationTime;
        }

        public void setExpirationTime(Instant expirationTime) {
            this.expirationTime.set(expirationTime);
        }

        public boolean getRenewRequested() {
            return renewRequested.get();
        }

        public BooleanProperty renewRequestedProperty() {
            return renewRequested;
        }

        public void setRenewRequested(boolean renewRequested) {
            this.renewRequested.set(renewRequested);
        }
    }

    RenewSliceDialog(Window parentWindow, ExperimentController experimentController, ExperimentTasksFactory experimentTasksFactory) {
        this.parentWindow = parentWindow;

        this.experimentController = experimentController;
        this.experimentTasksFactory = experimentTasksFactory;
        this.slice = experimentController.getExperiment().getSlice();
        URL location = RenewSliceDialog.class.getResource(RENEW_SLICE_DIALOG_FXML);
        FXMLLoader loader = new FXMLLoader(location);

        loader.setRoot(this);
        loader.setController(this);

        try {
            loader.load();
        } catch (IOException e) {
            e.printStackTrace();
        }


        //populate table
        DoubleBinding dividableWidth =
                authoritiesTableView.widthProperty().subtract(selectionTableColumn.widthProperty()).subtract(2);
        nameTableColumn.prefWidthProperty().bind(dividableWidth.multiply(0.375));
        expirationTableColumn.prefWidthProperty().bind(dividableWidth.multiply(0.25));
        durationTableColumn.prefWidthProperty().bind(dividableWidth.multiply(0.375));


        Multimap<SfaAuthority, Sliver> sliverPerAuthorityMap = HashMultimap.create();
        for (Sliver sliver : slice.getSlivers()) {
            sliverPerAuthorityMap.put(sliver.getAuthority(), sliver);
        }

        renewableAuthorities = FXCollections.observableArrayList();

        for (SfaAuthority authority : sliverPerAuthorityMap.keySet()) {
            Sliver firstSliver = sliverPerAuthorityMap.get(authority).iterator().next();

            renewableAuthorities.add(new RenewableAuthority(authority,
                    sliverPerAuthorityMap.get(authority), firstSliver.expirationDateProperty()));
        }

        authoritiesTableView.setItems(renewableAuthorities);
        authoritiesTableView.setPrefHeight(26/*header*/ + renewableAuthorities.size() * 25);

        selectionTableColumn.setCellValueFactory(new PropertyValueFactory<>("renewRequested"));
        selectionTableColumn.setCellFactory(CheckBoxTableCell.forTableColumn(selectionTableColumn));

        nameTableColumn.setCellValueFactory(t -> new SimpleStringProperty(t.getValue().getAuthority().getHrn()));

        expirationTableColumn.setCellValueFactory(new PropertyValueFactory<>("expirationTime"));

        expirationTableColumn.setCellFactory(t -> new TableCell<RenewableAuthority, Instant>() {
            @Override
            protected void updateItem(Instant date, boolean empty) {
                if (date == null || empty) {
                    setText(null);
                } else {
                    setText(DateTimeUtils.DATE_HOUR_MINUTES_SECONDS_FORMATTER.format(LocalDateTime.ofInstant(date, ZoneId.systemDefault())));
                }
            }
        });
        durationTableColumn.setCellValueFactory(new PropertyValueFactory<>("expirationTime"));

        durationTableColumn.setCellFactory(new Callback<TableColumn<RenewableAuthority, Instant>, TableCell<RenewableAuthority, Instant>>() {
            @Override
            public TableCell<RenewableAuthority, Instant> call(TableColumn<RenewableAuthority, Instant> renewableAuthorityDateTableColumn) {
                TableCell<RenewableAuthority, Instant> tableCell = new TableCell<RenewableAuthority, Instant>() {
                    private Instant dateTime = null;
                    private final Timeline timeline;

                    //instance initializer
                    {
                        this.timeline =
                                new Timeline(new KeyFrame(javafx.util.Duration.ZERO, actionEvent -> updateString()),
                                        new KeyFrame(new javafx.util.Duration(1000)));
                        this.timeline.setCycleCount(Animation.INDEFINITE);
                    }

                    @Override
                    protected void updateItem(Instant date, boolean empty) {
                        if (date == null || empty) {
                            timeline.stop();
                            dateTime = null;
                            setText("");
                        } else {
                            dateTime = date;
                            updateString();
                            timeline.play();

                        }
                    }

                    private void updateString() {
                        setText(DateTimeUtils.prettyPrintDuration(Duration.between(Instant.now(), dateTime).withNanos(0)));
                    }
                };
                return tableCell;
            }
        });


        Duration DEFAULT_DURATION = Duration.ofHours(2);

        Duration periodLeft = DEFAULT_DURATION;
        if (slice.getExpirationDate() != null) {
            periodLeft = Duration.between(Instant.now(), slice.getExpirationDate());

            //update sliceInfoLabel
            sliceInfoLabel.setText(String.format("Your experiment currently expires at %s. When necessary, it will automatically be renewed to match the latest authority expiration time.",
                    DateTimeUtils.DATE_HOUR_MINUTES_FORMATTER.format(LocalDateTime.ofInstant(slice.getExpirationDate(), ZoneId.systemDefault()))));
        }

        //populate combo-boxes
        daysComboBox.getItems().addAll(IntStream.range(0, 100).boxed().collect(Collectors.toList()));
        hoursComboBox.getItems().addAll(IntStream.range(0, 24).boxed().collect(Collectors.toList()));
        minutesComboBox.getItems().addAll(IntStream.range(0, 60).boxed().collect(Collectors.toList()));
        minutesComboBox.setConverter(new StringConverter<Integer>() {
            @Override
            public String toString(Integer integer) {
                return String.format("%02d", integer);
            }

            @Override
            public Integer fromString(String s) {
                return Integer.parseInt(s);
            }
        });

        long time = periodLeft.toMinutes();
        minutesComboBox.setValue((int) time % 60);
        time /= 60;
        hoursComboBox.setValue((int) time % 24);
        time /= 24;
        daysComboBox.setValue((int) time);

        //initialize dialog
        dialog = new Dialog(parentWindow, "Renew Experiment");
        dialog.setMasthead("Enter the new experiment duration:");
        dialog.setGraphic(GlyphUtils.createDialogGlyph(FontAwesome.Glyph.CLOCK_ALT, Color.LAWNGREEN));
        dialog.setContent(this);
        dialog.getActions().setAll(new DialogAction("Renew", ButtonBar.ButtonType.OK_DONE, false, false, true, this),
                Dialog.ACTION_CANCEL);
    }


    public void showDialog() {
        dialog.show();
    }


    @Override
    public void accept(ActionEvent actionEvent) {
        Instant newExpirationTime = ZonedDateTime.now()
                .plusDays(daysComboBox.getSelectionModel().getSelectedItem())
                .plusHours(hoursComboBox.getSelectionModel().getSelectedItem())
                .plusMinutes(minutesComboBox.getSelectionModel().getSelectedItem())
                .plusSeconds(15) //prevent problems with checks that are too strict
                .toInstant();

        if (newExpirationTime.isBefore(slice.getExpirationDate())) {
            int[] timeLeft =
                    DateTimeUtils.getSecondsMinutesHoursDays(Duration.between(Instant.now(), slice.getExpirationDate()));

            dialog.shake();

            JFDialogs.create()
                    .owner(null)
                    .message(String.format("The new expiration duration should be longer than the current one (%d days, %d hours and %d minutes)",
                            timeLeft[DateTimeUtils.DAYS],
                            timeLeft[DateTimeUtils.HOURS],
                            timeLeft[DateTimeUtils.MINUTES]))
                    .masthead("Experiment renewal failed")
                    .showError();
        } else {
            dialog.hide();

            boolean renewAll = true;

            for (RenewableAuthority renewableAuthority : renewableAuthorities) {
                if (!renewableAuthority.getRenewRequested()) {
                    renewAll = false;
                    break;
                }
            }

            if (renewAll) {
                experimentController.submitExperimentTask(experimentTasksFactory.createRenewSliceTask(experimentController.getExperiment(), newExpirationTime));

            } else {
                List<SfaAuthority> selectedAuthorities = renewableAuthorities.stream()
                        .filter(RenewableAuthority::getRenewRequested)
                        .map(RenewableAuthority::getAuthority).collect(Collectors.toList());


                experimentController.submitExperimentTask(
                        experimentTasksFactory.createRenewSliversAtAuthoritiesTask(
                                experimentController.getExperiment(), newExpirationTime, selectedAuthorities));


            }
        }
    }

}
