package be.iminds.ilabt.jfed.experimenter_gui.editor;

import be.iminds.ilabt.jfed.experimenter_gui.canvas.EditableExperimentCanvas;
import be.iminds.ilabt.jfed.experimenter_gui.editor.bo.NodeDescription;
import be.iminds.ilabt.jfed.experimenter_gui.editor.bo.RspecNodeDescription;
import be.iminds.ilabt.jfed.experimenter_gui.util.InetAddressUtil;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityInfo;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityList;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityListModel;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.rspec.model.RspecInterface;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.model.SliverTypeBuilder;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.*;
import com.google.common.base.Preconditions;
import com.google.common.collect.Collections2;
import javafx.geometry.Point2D;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This class is used as a wrapper around an FXModelRspec-object, and contains convenience methods for
 * editing the FXModelRspec
 * <p/>
 * User: twalcari
 * Date: 6/30/2014
 * Time: 14:32
 */
public class ModelRspecEditor {
    private static final Logger LOG = LoggerFactory.getLogger(ModelRspecEditor.class);
    private static final long DEFAULT_LINK_CAPACITY = 10_000l;
    private static final Pattern IP_ADDRESS_PATTERN = Pattern.compile("(\\d{1,3}\\.\\d{1,3}\\.\\d{1,3})\\.(\\d{1,3})");
    public static final String DEFAULT_IPV4_NETMASK = "255.255.255.0";
    public static final String DEFAULT_IPV6_NETMASK = "64";

    private final Map<FXRspecNode, RspecNodeDescription> nodeDescriptionMap = new HashMap<>();
    private final Map<FXRspecLink, String> linkSubnetworks = new HashMap<>();

    private final FXModelRspec modelRspec;
    private final AuthorityList authorityList;

    private static final FXRspecFactory rspecFactory = FXRspecFactory.getInstance();

    private EditableExperimentCanvas editableExperimentCanvas = null;

    public void setEditableExperimentCanvas(EditableExperimentCanvas editableExperimentCanvas) {
        this.editableExperimentCanvas = editableExperimentCanvas;
    }

    public ModelRspecEditor(FXModelRspec modelRspec, AuthorityList authorityList) {
        this.modelRspec = modelRspec;
        this.authorityList = authorityList;

        //try to find the subnetwork used in this link
        for (FXRspecLink link : modelRspec.getLinks()) {
            String subnetwork = getInet4SubnetworkFromLink(link);

            if (subnetwork != null)
                linkSubnetworks.put(link, subnetwork);
        }
    }

    public static String getInet4SubnetworkFromLink(FXRspecLink link) {
        //check if link-interfaces adhere to our restrictions
        for (FXRspecInterface iface : link.getInterfaces())
            for (FXRspecInterface.IpAddress addr : iface.getIpAddresses()) {
                //we only support IPv4-links
                if (!RspecInterface.IpAddress.TYPE_IPV4.equals(addr.getType()))
                    return null;

                //we don't support other netmasks besides DEFAULT_IPV4_NETMASK
                if (!DEFAULT_IPV4_NETMASK.equals(addr.getNetmask()))
                    return null;
            }

        Iterator<FXRspecInterface> it = link.getInterfaces().iterator();

        if (!it.hasNext())
            return null;

        FXRspecInterface iface = it.next();

        if (iface.getIpAddresses().size() != 1)
            return null;

        if (!InetAddressUtil.isValidInet4Address(iface.getIpAddresses().get(0).getAddress()))
            return null;

        String subnetwork =
                InetAddressUtil.getInet4Subnetwork(InetAddressUtil.simplifyInet4Address(iface.getIpAddresses().get(0).getAddress()));

        //check if subnetwork is consistent for all interfaces
        while (it.hasNext()) {
            iface = it.next();
            if (iface.getIpAddresses().size() != 1)
                return null;

            if (!InetAddressUtil.isValidInet4Address(iface.getIpAddresses().get(0).getAddress()) ||
                    !InetAddressUtil.simplifyInet4Address(iface.getIpAddresses().get(0).getAddress()).startsWith(subnetwork))
                return null;
        }

        return subnetwork;
    }

    public static String getInet6SubnetworkFromLink(FXRspecLink link) {
        //check if link-interfaces adhere to our restrictions
        for (FXRspecInterface iface : link.getInterfaces())
            for (FXRspecInterface.IpAddress addr : iface.getIpAddresses()) {
                //we only support IPv4-links
                if (!RspecInterface.IpAddress.TYPE_IPV6.equals(addr.getType()))
                    return null;

                //we don't support other netmasks besides DEFAULT_IPV4_NETMASK
                if (!DEFAULT_IPV6_NETMASK.equals(addr.getNetmask()))
                    return null;
            }

        Iterator<FXRspecInterface> it = link.getInterfaces().iterator();

        if (!it.hasNext())
            return null;

        FXRspecInterface iface = it.next();

        if (iface.getIpAddresses().size() != 1)
            return null;

        if (!InetAddressUtil.isValidInet6Address(iface.getIpAddresses().get(0).getAddress()))
            return null;

        String subnetwork =
                InetAddressUtil.getInet6Subnetwork(iface.getIpAddresses().get(0).getAddress());

        //check if subnetwork is consistent for all interfaces
        while (it.hasNext()) {
            iface = it.next();
            if (iface.getIpAddresses().size() != 1)
                return null;

            if (!iface.getIpAddresses().get(0).getAddress().startsWith(subnetwork))
                return null;
        }

        return subnetwork;
    }

    public FXModelRspec getModelRspec() {
        return modelRspec;
    }

    /**
     * This method changes some settings of the FXRspecLink to make it a correct FXRspecLink
     *
     * @param rspecLink
     */
    public void convertLinkToStitchedLink(FXRspecLink rspecLink) {
        rspecLink.getLinkTypes().remove("lan");

        //add default capacity
        for (FXRspecInterface iface1 : rspecLink.getInterfaces()) {
            for (FXRspecInterface iface2 : rspecLink.getInterfaces()) {
                if (iface1 == iface2)
                    continue;

                if (!rspecLink.getLinkSetting(iface1, iface2).isCapacitySet())
                    rspecLink.getLinkSetting(iface1, iface2).setCapacity_Kbps(DEFAULT_LINK_CAPACITY);
            }
        }
    }


    public void convertLinkToNormalLink(FXRspecLink rspecLink) {
        rspecLink.getLinkTypes().setAll("lan");

        //remove default capacity
        for (FXRspecInterface iface1 : rspecLink.getInterfaces()) {
            for (FXRspecInterface iface2 : rspecLink.getInterfaces()) {
                if (iface1 == iface2)
                    continue;

                if (rspecLink.getLinkSetting(iface1, iface2).getCapacity_Kbps() == DEFAULT_LINK_CAPACITY)
                    rspecLink.getLinkSetting(iface1, iface2).setCapacity_Kbps(0L);
            }
        }
    }

    /**
     * Gives the next available subnetwork in the 10.x.x-range.
     *
     * @return
     */
    public String getNextAvailableInet4Subnetwork() {

        int subnetwork = -1;
        String subnetworkString = null;
        boolean inUse = true;

//        while (inUse) { //for 10.
        while (inUse && subnetwork < 255) { //for 192.168
            subnetwork++;
//            subnetworkString = "10." + (subnetwork / 255) + "." + (subnetwork % 255);
            subnetworkString = "192.168." + (subnetwork % 255);

            inUse = (linkSubnetworks.containsValue(subnetworkString));
        }

//        if ((subnetwork / 255) < 255) { //for 10.
        if (subnetwork < 255) { //for 192.168
            return subnetworkString;
        } else
            return null;
    }


    private int getNextLastIpAddressDigits(String subNetwork) {

        boolean[] usedLastDigits = new boolean[254]; //from 1 to 254

        //multiple links can have the same subnet (ex. with delay bridges)
        for (Map.Entry<FXRspecLink, String> entry : linkSubnetworks.entrySet()) {
            if (!subNetwork.equals(entry.getValue()))
                continue;

            for (FXRspecInterface iface : entry.getKey().getInterfaces()) {
                if (!iface.getIpAddresses().isEmpty()) {
                    Matcher matcher = IP_ADDRESS_PATTERN.matcher(iface.getIpAddresses().get(0).getAddress());

                    if (matcher.matches()) {
                        if (!subNetwork.equals(matcher.group(1)))
                            return -1;

                        int lastDigit = Integer.parseInt(matcher.group(2));
                        if (lastDigit >= 1 && lastDigit <= 254)
                            usedLastDigits[lastDigit - 1] = true;
                        else
                            return -1; //encountered an illegal last digit
                    } else {
                        return -1; //found an interface with an invalid ip-address
                    }
                }
            }
        }

        int lastDigit = 1;
        while (lastDigit < 255 && usedLastDigits[lastDigit - 1]) {
            lastDigit++;
        }

        if (lastDigit == 255) return -1; //could not find usable last digit in subnetwork

        return lastDigit;
    }

    private void configureIfaceForSubnetwork(FXRspecInterface iface, String subnetwork) {
        configureIfaceForSubnetwork(iface, subnetwork, getNextLastIpAddressDigits(subnetwork));
    }

    private void configureIfaceForSubnetwork(FXRspecInterface iface, String subnetwork, int lastDigit) {
        Preconditions.checkNotNull(iface);
        Preconditions.checkNotNull(subnetwork);
        Preconditions.checkArgument(lastDigit > 0 && lastDigit < 255, "Got an illegal last digit");

        iface.getIpAddresses().add(new FXRspecInterface.IpAddress(subnetwork + "." + lastDigit, DEFAULT_IPV4_NETMASK,
                RspecInterface.IpAddress.TYPE_IPV4));
    }


    /**
     * Converts a normal FXRspecLink to a link with an impairment bridge
     *
     * @param rspecLink
     * @return
     */
    public DelayRspecNode convertLinkToImpairmentBridge(FXRspecLink rspecLink) {
        if (rspecLink.getInterfaces().size() > 2) {
            return null;
        }

        String impairmentBridgeSubnetwork = linkSubnetworks.containsKey(rspecLink) ?
                linkSubnetworks.get(rspecLink) : getNextAvailableInet4Subnetwork();

        if (impairmentBridgeSubnetwork == null) {
            return null;  //ran out of available sub-networks!
        }

        FXRspecInterface iface1 = rspecLink.getInterfaces().get(0);
        FXRspecNode node1 = iface1.getNode();
        String ifaceId1 = iface1.getClientId();

        FXRspecInterface iface2 = rspecLink.getInterfaces().get(1);
        FXRspecNode node2 = iface2.getNode();
        String ifaceId2 = iface2.getClientId();

        modelRspec.deleteLink(rspecLink);

        DelayRspecNode impairmentBridgeNode = new DelayRspecNode(modelRspec, getNextImpairmentBridgeName());
        impairmentBridgeNode.setEditorX((node1.getEditorX() + node2.getEditorX()) / 2);
        impairmentBridgeNode.setEditorY((node1.getEditorY() + node2.getEditorY()) / 2);
        impairmentBridgeNode.setExclusive(true);
        impairmentBridgeNode.setComponentManagerId(node1.getComponentManagerId());

        modelRspec.getNodes().add(impairmentBridgeNode);

        FXRspecLink newLink1 = rspecFactory.createLinkWithClientId(modelRspec, modelRspec.nextLinkName(), "lan");

        FXRspecInterface link1Iface1 = rspecFactory.createInterface(node1, newLink1, ifaceId1);
        configureIfaceForSubnetwork(link1Iface1, impairmentBridgeSubnetwork);
        rspecFactory.createInterface(impairmentBridgeNode, newLink1,
                impairmentBridgeNode.getClientId() + ":left");
        modelRspec.getLinks().add(newLink1);

        FXRspecLink newLink2 = rspecFactory.createLinkWithClientId(modelRspec, modelRspec.nextLinkName(), "lan");
        FXRspecInterface link2Iface1 = rspecFactory.createInterface(node2, newLink2, ifaceId2);
        configureIfaceForSubnetwork(link2Iface1, impairmentBridgeSubnetwork);
        rspecFactory.createInterface(impairmentBridgeNode, newLink2,
                impairmentBridgeNode.getClientId() + ":right");
        modelRspec.getLinks().add(newLink2);

        return impairmentBridgeNode;

    }

    private String getNextImpairmentBridgeName() {
        int nr = 0;

        while (modelRspec.getNodeByClientId("bridge" + nr) != null) {
            nr++;
        }

        return "bridge" + nr;
    }


    private boolean isInterfaceConfigurationAllowed(FXRspecNode node) {
        RspecNodeDescription nodeDescription = nodeDescriptionMap.get(node);

        return nodeDescription == null
                || nodeDescription.isConfigureInterfaces();
    }

    public FXRspecNode addNode(NodeDescription nodeDescription, Point2D location) {
        RspecNodeDescription rnd = nodeDescription.getRspecNodeDescriptions().get(0);

        FXRspecNode rspecNode = rspecFactory.createNodeWithClientId(modelRspec, modelRspec.nextNodeName());
        rspecNode.setNodeDescriptionId(nodeDescription.getId());

        rspecNode.setComponentManagerId(rnd.getComponentManagerUrn());
        rspecNode.setExclusive(rnd.isExclusive());
        rspecNode.setComponentId(rnd.getComponentIdUrn());
        SliverTypeBuilder sliverTypeBuilder = new SliverTypeBuilder(rnd.getSliverType());
        if (rnd.getOsImage() != null)
            sliverTypeBuilder.addDiskImage(rnd.getOsImage());
        rspecNode.setSliverType(sliverTypeBuilder.build());
//        rspecNode.setOsImageUrn();


        rspecNode.setEditorX(location.getX());
        rspecNode.setEditorY(location.getY());

        modelRspec.getNodes().add(rspecNode);

        nodeDescriptionMap.put(rspecNode, rnd);

        return rspecNode;
    }


    /**
     * Creates a new link between two nodes
     *
     * @param node1
     * @param node2
     */
    public FXRspecLink addLink(FXRspecNode node1, FXRspecNode node2) {
        LOG.trace("Adding link between {} and {}", node1.getClientId(), node2.getUniqueId());
        //we first need to create a new link
        FXRspecLink rspecLink = rspecFactory.createLinkWithClientId(modelRspec, modelRspec.nextLinkName(), "lan");
        linkSubnetworks.put(rspecLink, getNextAvailableInet4Subnetwork());

        addNodeToLink(node1, rspecLink);
        addNodeToLink(node2, rspecLink);

        modelRspec.getLinks().add(rspecLink);

        if (shouldBeStitchedLink(node1, node2)) {
            convertLinkToStitchedLink(rspecLink);
        }
        return rspecLink;

    }

    /**
     * Creates a new FXRspecInterface to connect an FXRspecLink and an FXRspecNode.
     * Does additional configuration on this interface (or other interfaces on the link) if
     * required by the configuration-file.
     *
     * @param node
     * @param link
     * @return
     */
    private FXRspecInterface addNodeToLink(FXRspecNode node, FXRspecLink link) {
        FXRspecInterface iface = rspecFactory.createInterface(node,
                link,
                modelRspec.nextIfaceName(node));
        LOG.trace("Added node {} to link {} with iface {}", node.getUniqueId(), link.getUniqueId(), iface.getClientId());

        if (isInterfaceConfigurationAllowed(node)) {
            if (linkSubnetworks.containsKey(link)) {
                configureIfaceForSubnetwork(iface, linkSubnetworks.get(link));
            }
        } else {
            //don't add interfaces to fake authorities to prevent confusion. But also don't remove IP-addresses of other nodes on the link
            AuthorityInfo nodeAuthInfo =
                    authorityList.getByUrn(node.getComponentManagerId(), AuthorityListModel.SubAuthMatchAllowed.ALLOW_OTHER_SUBAUTHORITY, AuthorityListModel.SubAuthMatchPreference.PREFER_EXACT_SUBAUTHORITY);
            if (nodeAuthInfo == null || nodeAuthInfo.getSfaAuthority().isReal()) {
                //remove all assigned ip-addresses from this link
                if (linkSubnetworks.containsKey(link)) {
                    linkSubnetworks.remove(link);
                }

                link.getInterfaces().forEach(be.iminds.ilabt.jfed.experimenter_gui.editor.ModelRspecEditor::clearInterface);
            }
        }

        return iface;
    }

    private static void clearInterface(FXRspecInterface iface) {
        iface.getIpAddresses().clear();
    }


    /**
     * Adds a new node to an existing link
     *
     * @param link
     * @param node
     */
    public boolean addLink(FXRspecLink link, FXRspecNode node) {
        return addNodeToLink(node, link) != null;
    }

    private boolean shouldBeStitchedLink(FXRspecNode node1, FXRspecNode node2) {

        return node1.getComponentManagerId() != null &&
                !node1.getComponentManagerId().equals(node2.getComponentManagerId())
                || node2.getComponentManagerId() == null;
    }


    public boolean areAllNodePositionsDefined() {

        for (FXRspecNode node : modelRspec.getNodes()) {
            if (node.getEditorX() == -1 && node.getEditorY() == -1)
                return false;
        }
        return true;

    }

    public void arrangeNodesOnCircle() {
        arrangeNodesOnCircle(600, 400);
    }

    /**
     * This method arranges all the nodes on an ellipse of the given width and height
     *
     * @param width
     * @param height
     */
    public void arrangeNodesOnCircle(double width, double height) {
        if (editableExperimentCanvas != null) {
            editableExperimentCanvas.setZoom(1);
        }
        arrangeNodesOnCircle(modelRspec, width, height);
    }


    /**
     * This method arranges all the nodes on an ellipse of the given width and height
     *
     * @param width
     * @param height
     */
    public static void arrangeNodesOnCircle(FXModelRspec modelRspec, double width, double height) {
        LOG.debug("Arranging all nodes on a circle of {}x{}", width, height);
        if (width < 100.0) width = 300.0;
        if (height < 100.0) height = 300.0;

        int size = modelRspec.getNodes().size();
        double radiusW = width / 3.0; //diameter of 2/3 width
        double radiusH = height / 3.0; //diameter of 2/3 height

        double centerX = width / 2.0;
        double centerY = height / 2.0;
        double angle = 0.0;
        double step = (2 * Math.PI) / size;
        for (FXRspecNode n : modelRspec.getNodes()) {
            double x = (Math.cos(angle) * radiusW) + centerX;
            double y = (Math.sin(angle) * radiusH) + centerY;

            x -= 90;
            y -= 15;

            n.setEditorX(x);
            n.setEditorY(y);

            angle += step;
        }
    }


    /**
     * Returns a list of nodes which are not bound to an component manager. If this list is not empty,
     * the user should be asked which component manager to use.
     *
     * @return A list of unbound Rspecs
     */
    public Collection<FXRspecNode> getUnboundNodes() {
        return Collections2.filter(modelRspec.getNodes(), input -> input.getComponentManagerId() == null);
    }

    public void resolveUnboundRspecNodesToAuthority(SfaAuthority sfaAuthority) {
        for (RspecNode rspecNode : getUnboundNodes()) {
            rspecNode.setComponentManagerId(sfaAuthority.getUrn());
        }
    }

    public FXRspecChannel addChannel(Point2D location) {
        FXRspecChannel channel = (FXRspecChannel) rspecFactory.createChannel(modelRspec);
        channel.setEditorX(location.getX());
        channel.setEditorY(location.getY());

        modelRspec.addChannel(channel);

        return channel;
    }
}
