package be.iminds.ilabt.jfed.experimenter_gui.editor.properties.bonfire;

import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.AdaptivePropertiesDialog;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.AdaptivePropertiesTab;
import be.iminds.ilabt.jfed.highlevel.controller.HighLevelController;
import be.iminds.ilabt.jfed.highlevel.controller.Task;
import be.iminds.ilabt.jfed.highlevel.controller.TaskExecution;
import be.iminds.ilabt.jfed.highlevel.model.AuthorityList;
import be.iminds.ilabt.jfed.highlevel.model.EasyOcciModel;
import be.iminds.ilabt.jfed.lowlevel.JFedConnectivityException;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl.OcciToAggregateManagerWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.rspec.ext.jfedbonfire.BonfireNodeInfo;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.BonfireRspecNode;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.CheckboxList;
import be.iminds.ilabt.jfed.util.GeniUrn;
import com.google.common.collect.Lists;
import javafx.application.Platform;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.Node;
import javafx.scene.control.Label;
import javafx.scene.control.ProgressBar;
import javafx.scene.control.TextField;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.*;
import java.util.stream.Collectors;

public class BonfireNodePropertiesTab extends AdaptivePropertiesTab<BonfireRspecNode> {
    public static final GeniUrn BONFIRE_URN = GeniUrn.parse("urn:publicid:IDN+bonfire-project.eu+authority+cm");
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(BonfireNodePropertiesTab.class);
    private static final double PROGRESS_FETCH_START = 0.0;
    private static final double PROGRESS_FETCH_REQUESTED = -1.0; //0.1
    private static final double PROGRESS_FETCH_COMPLETE = 0.90; //0.60
    private static final String BONFIRE_NODE_PROPERTIES_FXML = "BonfireNodeProperties.fxml";
    private final AdaptivePropertiesDialog adaptivePropertiesDialog;
    @FXML
    private Label noBonfireSiteSelectedLabel;
    @FXML
    private TextField instanceTypeTextField;
    @FXML
    private CheckboxList<EasyOcciModel.Network> networks;
    @FXML
    private CheckboxList<EasyOcciModel.Storage> storages;
    @FXML
    private ProgressBar networksProgressBar;
    @FXML
    private ProgressBar storagesProgressBar;
    private BonfireNodeInfo bonfireNodeInfo = null;
    private String lastInitializedLocation = null;
    private String lastLocationForWhichDefaultNetworkSelected = null;
    private String lastLocationForWhichDefaultStorageSelected = null;

    private final EasyOcciModel easyOcciModel;
    private final AuthorityList authorityList;
    private final HighLevelController hlc;

    public BonfireNodePropertiesTab(AdaptivePropertiesDialog adaptivePropertiesDialog, EasyOcciModel easyOcciModel, AuthorityList authorityList,HighLevelController hlc) {
        super("Bonfire");
        this.adaptivePropertiesDialog = adaptivePropertiesDialog;
        this.easyOcciModel = easyOcciModel;
        this.authorityList = authorityList;
        this.hlc = hlc;
        FXMLLoader fxmlLoader = new FXMLLoader(getClass().getResource(BONFIRE_NODE_PROPERTIES_FXML));
        //fxmlLoader.setRoot(this);
        fxmlLoader.setController(this);
        try {
            Node contents = fxmlLoader.load();
            setContent(contents);
        } catch (IOException e) {
            LOG.error("Error loading FXML: " + e.getMessage(), e);
            throw new RuntimeException("Error loading FXML: " + e.getMessage(), e);
        }
    }

    private static boolean isConnectivityException(Throwable t) {

        while (t != null) {
            if (t.getCause() instanceof JFedConnectivityException)
                return true;
            else
                t = t.getCause();
        }
        return false;
    }

    public void initForLocation(String locationId) {

        if (locationId == null) {
            this.instanceTypeTextField.setText("");
            this.networks.setOptions(Collections.<EasyOcciModel.Network>emptyList());
            this.storages.setOptions(Collections.<EasyOcciModel.Storage>emptyList());
            lastInitializedLocation = null;
            return;
        }

        //only initialize if location differs from previous location
        if (locationId.equals(lastInitializedLocation))
            return;


        if (bonfireNodeInfo != null && bonfireNodeInfo.getInstanceType() != null) {
            instanceTypeTextField.setText(bonfireNodeInfo.getInstanceType());
        } else if (instanceTypeTextField.getText().isEmpty()) {
            instanceTypeTextField.setText("lite");
        }

        initNetworksForLocation(locationId, false);
        initStoragesForLocation(locationId, false);

        lastInitializedLocation = locationId;
    }

    private void initNetworksForLocation(String locationId, boolean calledFromRefresh) {
        if (locationId == null) {
            this.networks.setOptions(Collections.<EasyOcciModel.Network>emptyList());
            return;
        }

        Set<EasyOcciModel.Network> occiNetworks = easyOcciModel.getNetworks(locationId);
        List<EasyOcciModel.Network> newNetworks = new ArrayList<>();
        Set<String> prevSelectedNetworks = new HashSet<>();

        if (bonfireNodeInfo != null) {
            prevSelectedNetworks.addAll(Lists.transform(bonfireNodeInfo.getNetworkHrefs(),
                    BonfireNodeInfo.NetworkHref::getHref
            ));

        } else if (!this.networks.getSelectedItems().isEmpty()) {
            prevSelectedNetworks.addAll(Lists.transform(this.networks.getSelectedItems(),
                    EasyOcciModel.Network::getHref
            ));
        }

        boolean autoSelectNetwork =
                lastLocationForWhichDefaultNetworkSelected != locationId && prevSelectedNetworks.isEmpty();
        LOG.debug("initForLocation locationId={} autoSelectNetwork={} prevSelectedNetworks.size()={}",
                locationId, autoSelectNetwork, prevSelectedNetworks.size());

        if (occiNetworks != null)
            for (EasyOcciModel.Network network : occiNetworks) {
                newNetworks.add(network);
                LOG.debug("initForLocation adding network: " + network.getName());
                if (autoSelectNetwork && network.getName().contains("WAN")) {
                    prevSelectedNetworks.add(network.getHref());
                    autoSelectNetwork = false;
                }
            }
        this.networks.setOptions(newNetworks);

        if (newNetworks.isEmpty()) {
            //automatically fetch
            if (!calledFromRefresh) onRefreshNetworks();
        } else {
            List<EasyOcciModel.Network> selectedItems =
                    newNetworks.stream().filter(network -> prevSelectedNetworks.contains(network.getHref())).collect(Collectors.toList());
            this.networks.setSelectedOptions(selectedItems);
            lastLocationForWhichDefaultNetworkSelected = locationId;

            this.networks.setDisable(false);
        }
    }

    public void initStoragesForLocation(String locationId, boolean calledFromRefresh) {
        if (locationId == null) {
            this.storages.setOptions(Collections.<EasyOcciModel.Storage>emptyList());
            return;
        }

        Set<EasyOcciModel.Storage> occiStorages = easyOcciModel.getStorages(locationId);
        List<EasyOcciModel.Storage> newStorages = new ArrayList<>();
        Set<String> prevSelectedStorages = new HashSet<>();

        if (bonfireNodeInfo != null) {
            prevSelectedStorages.addAll(Lists.transform(bonfireNodeInfo.getStorageHrefs(),
                    BonfireNodeInfo.StorageHref::getHref
            ));

        } else if (!this.networks.getSelectedItems().isEmpty()) {
            prevSelectedStorages.addAll(Lists.transform(this.storages.getSelectedItems(),
                    EasyOcciModel.Storage::getHref
            ));
        }

        boolean autoSelectStorage =
                lastLocationForWhichDefaultStorageSelected != locationId && prevSelectedStorages.isEmpty();
        LOG.debug("initForLocation occiStorages: " + occiStorages);
        if (occiStorages != null)
            for (EasyOcciModel.Storage storage : occiStorages) {
                newStorages.add(storage);
                LOG.debug("initForLocation adding storage: " + storage.getName());

                if (autoSelectStorage) {
                    if (storage.getName().equalsIgnoreCase("BonFIRE Debian Squeeze v6")) {
                        prevSelectedStorages.add(storage.getHref());
                        autoSelectStorage = false;
                    }
                }
            }
        this.storages.setOptions(newStorages);

        if (newStorages.isEmpty()) {
            //automatically fetch
            if (!calledFromRefresh) onRefreshStorages();
        } else {
            List<EasyOcciModel.Storage> selectedItems =
                    newStorages.stream().filter(storage -> prevSelectedStorages.contains(storage.getHref())).collect(Collectors.toList());
            this.storages.setSelectedOptions(selectedItems);
            lastLocationForWhichDefaultStorageSelected = locationId;

            this.storages.setDisable(false);
        }
    }

    @FXML
    public void onRefreshNetworks() {
        //show progressbar
        networks.setDisable(true);
        networks.clear();
        networksProgressBar.setProgress(PROGRESS_FETCH_START);
        networksProgressBar.setVisible(true);

        final FXRspecNode selectedComponent = adaptivePropertiesDialog.getSelectedComponentId();
        if (selectedComponent == null) return;

        final SfaAuthority bonfireAuthority =
                authorityList.getAuthorityListModel().getByUrnExact(OcciToAggregateManagerWrapper.bonfireUrn);
        final String locationId = selectedComponent.getComponentId().getEncodedResourceName();

        Task fetchAdvertisementTask =
                hlc.getOcciNetworksForLocation(bonfireAuthority, locationId);

        fetchAdvertisementTask.addCallback((task, taskExecution, state) -> Platform.runLater(() -> {
            if (state == TaskExecution.TaskState.SUCCESS) {
                networksProgressBar.setProgress(PROGRESS_FETCH_COMPLETE);
                initNetworksForLocation(locationId, true);
            } else if (state == TaskExecution.TaskState.FAILED) {
                networksProgressBar.setVisible(false);
                if (isConnectivityException(taskExecution.getException())) {
                    JFDialogs.create()
                            .owner(null)
                            .message("A connectivity exception occurred while fetching the networks advertisement from " + bonfireAuthority.getName() + ".\n\n" +
                                    "Verify your internet connection, and consider enabling a proxy in the jFed Preferences.")
                            .masthead("Fetching node advertisement failed")
                            .title("Fetching node advertisement failed")
                            .showException(taskExecution.getException());

                } else {
                    JFDialogs.create()
                            .owner(null)
                            .message("An error occurred while fetching the nodes advertisement " +
                                    "from " + bonfireAuthority.getName() + ".")
                            .masthead("Fetching node advertisement failed")
                            .title("Fetching node advertisement failed")
                            .showException(taskExecution.getException());

                }
            }
        }));
        networksProgressBar.setProgress(PROGRESS_FETCH_REQUESTED);

        hlc.submitTask(fetchAdvertisementTask);
    }

    @FXML
    public void onRefreshStorages() {
        //show progressbar
        storages.setDisable(true);
        storages.clear();
        storagesProgressBar.setProgress(PROGRESS_FETCH_START);
        storagesProgressBar.setVisible(true);

        final FXRspecNode selectedComponent = adaptivePropertiesDialog.getSelectedComponentId();
        if (selectedComponent == null) return;

        final SfaAuthority bonfireAuthority =
                authorityList.getAuthorityListModel().getByUrnExact(OcciToAggregateManagerWrapper.bonfireUrn);
        final String locationId = selectedComponent.getComponentId().getEncodedResourceName();

        final Task fetchAdvertisementTask =    hlc.getOcciStoragesForLocation(bonfireAuthority, locationId);

        fetchAdvertisementTask.addCallback((task, taskExecution, state) -> Platform.runLater(() -> {
            if (state == TaskExecution.TaskState.SUCCESS) {
                storagesProgressBar.setProgress(PROGRESS_FETCH_COMPLETE);
                initStoragesForLocation(locationId, true);
                storagesProgressBar.setProgress(1);

            } else if (state == TaskExecution.TaskState.FAILED) {

                storagesProgressBar.setVisible(false);
                if (isConnectivityException(taskExecution.getException())) {

                    JFDialogs.create()
                            .owner(null)
                            .message("A connectivity exception occurred while fetching the nodes advertisement from " + bonfireAuthority.getName() + ".\n\n" +
                                    "Verify your internet connection, and consider enabling a proxy in the jFed Preferences.")
                            .masthead("Fetching node advertisement failed")
                            .title("Fetching node advertisement failed")
                            .showException(taskExecution.getException());

                } else {
                    JFDialogs.create()
                            .owner(null)
                            .message("An error occurred while fetching the nodes advertisement from " + bonfireAuthority.getName() + ".")
                            .masthead("Fetching node advertisement failed")
                            .title("Fetching node advertisement failed")
                            .showException(taskExecution.getException());

                }
            }
        }));
        storagesProgressBar.setProgress(PROGRESS_FETCH_REQUESTED);

        hlc.submitTask(fetchAdvertisementTask);
    }


    @FXML
    private void onSaveAction(ActionEvent actionEvent) {
        adaptivePropertiesDialog.onSaveAction(actionEvent);
    }

    @Override
    public void load(BonfireRspecNode rspecNode) {
        if (rspecNode != null) {
            bonfireNodeInfo = new BonfireNodeInfo(rspecNode.getBonfireNodeInfo());
        }
    }

    @Override
    public void save(BonfireRspecNode rspecNode) {
        BonfireNodeInfo bfni = new BonfireNodeInfo();

        bfni.setInstanceType(instanceTypeTextField.getText());
        for (EasyOcciModel.Network network : networks.getSelectedItems())
            bfni.getNetworkHrefs().add(new BonfireNodeInfo.NetworkHref(network.getHref()));
        for (EasyOcciModel.Storage storage : storages.getSelectedItems())
            bfni.getStorageHrefs().add(new BonfireNodeInfo.StorageHref(storage.getHref()));

        rspecNode.bonfireNodeInfoProperty().set(bfni);
    }

    @Override
    public void onTabShown() {
        //check which bonfire-site is selected, and load the appropriate resources
        FXRspecNode selectedNode =
                adaptivePropertiesDialog.getSelectedComponentId();
        if (selectedNode == null) {
            noBonfireSiteSelectedLabel.setVisible(true);
            return;
        } else {
            noBonfireSiteSelectedLabel.setVisible(false);

            assert selectedNode.getComponentId() != null :
                    "selectedComponentId.getUrn() == null for selectedComponentId=\"" + selectedNode + "\"";

            final String locationId = selectedNode.getComponentId().getEncodedResourceName();
            assert locationId != null :
                    "URN resourceName==null for \"" + selectedNode + "\"";

            initForLocation(locationId);
        }
    }

    @Override
    public boolean shouldBeVisible(FXRspecNode rspecNode) {
        if (adaptivePropertiesDialog.getSelectedComponentManagerInfo() == null) return false;
        if (adaptivePropertiesDialog.getSelectedComponentManagerInfo().getAuthority() == null) return false;
        return adaptivePropertiesDialog.getSelectedComponentManagerInfo().getAuthority().getUrn().equals(BONFIRE_URN);
    }

    @Override
    public String validateConfiguration() {

        if (instanceTypeTextField.getText().isEmpty()) {
            return "The instance type of a Bonfire Node must be specified";
        }

        return null;
    }
}
