package be.iminds.ilabt.jfed.call_log_output;

import be.iminds.ilabt.jfed.log.ApiCallDetails;
import be.iminds.ilabt.jfed.log.ResultListener;
import be.iminds.ilabt.jfed.log.cache.ApiCallDetailsCache;
import be.iminds.ilabt.jfed.log.cache.ApiCallDetailsRef;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import org.rendersnake.HtmlCanvas;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlTransient;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.stream.Collectors;


/**
 * CallReport
 */
public class CallReport extends HtmlLogOutput {
    private final static Logger LOG = LoggerFactory.getLogger(HtmlLogOutput.class);

    private final ApiCallDetailsCache cache;

    @XmlTransient
    protected State curState;

    @XmlElementWrapper(name = "states")
    @XmlElement(name = "state")
    protected ObservableList<State> states = FXCollections.observableArrayList();

    private final String reportObject;


    protected CallReport(String initialStateName, String reportTitle, String reportObject, ApiCallDetailsCache cache) {
        super(reportTitle);
        this.reportObject = reportObject;
        this.cache = cache;
        curState = new State(initialStateName, new Date(), cache);
        states.add(curState);
    }

    protected CallReport(String reportTitle, String reportObject, ApiCallDetailsCache cache) {
        this("Initial State", reportTitle, reportObject, cache);
    }

    protected CallReport(String reportTitle, String reportObject, Collection<State> states, ApiCallDetailsCache cache) {
        super(reportTitle);
        this.reportObject = reportObject;
        this.states.addAll(states);
        this.cache = cache;
    }

    public synchronized State getCurState() {
        return curState;
    }

    public synchronized State changeState(String newStateName) {
        LOG.debug("changeState(" + newStateName + ")");

        //only change if new state is different
        if (newStateName.equals(curState.name))
            return curState;

        Date now = new Date();
        State newState = new State(newStateName, now, cache);
        curState.end.set(now);
        states.add(newState);
        curState = newState;
        return newState;
    }

    public String getReportObject() {
        return reportObject;
    }

    public synchronized ObservableList<State> getStates() {
        return FXCollections.unmodifiableObservableList(states);
        //note: only allowed to do the "unmodifiableObservableList" because JAXB doesn't use this setter
        //if JAXB used this to populate the list, int would cause errors
    }

    public synchronized State getLastState(String stateName) {
        State res = null;
        for (State state : states)
            if (state.name.equals(stateName))
                res = state;
        return res;
    }

    public synchronized List<ApiCallDetailsRef> getAllStateCalls(String stateName) {
        return states.stream()
                .filter(state -> state.name.equals(stateName))
                .flatMap(state -> state.getCalls().stream())
                .collect(Collectors.toList());
    }

    @Override
    public synchronized void htmlContent(HtmlCanvas html) throws IOException {
        if (curState != null && curState.end.get() == null)
            curState.end.set(new Date());

        addClassVisibleToggle(html, "logline-DEBUG", "Debug Log Lines");

        int index = 0;
        for (State state : states) {
            List<SerializableApiCallDetails> callsCopy;
            List<LogLine> logsCopy;
            synchronized (state.calls) {
                callsCopy = state.calls.stream().map(cache::getApiCallDetails).collect(Collectors.toList());
            }
            synchronized (state.logs) {
                logsCopy = new ArrayList<>(state.logs);
            }
            writeGroup(html, index++, state.getState(), state.name + " calls", reportObject + " state \"" + state.name + "\" calls",
                    state.getDurationMs(), state.start, state.end.get(),
                    null/*exception*/,
                    callsCopy,
                    logsCopy);
        }
    }

    public synchronized List<ApiCallDetailsRef> getCalls() {
        List<ApiCallDetailsRef> res = new ArrayList<>();
        for (State state : states) {
            res.addAll(state.getCalls());
        }
        return res;
    }

    public static class State implements ResultListener {
        private final Date start;

        private final ObjectProperty<Date> end = new SimpleObjectProperty<>();

        private final String name;

        private final ObservableList<LogLine> logs =
                FXCollections.synchronizedObservableList(FXCollections.observableArrayList());

        private final ObservableList<ApiCallDetailsRef> calls =
                FXCollections.synchronizedObservableList(FXCollections.observableArrayList());

        private final ObjectProperty<TestResultState> testResultState =
                new SimpleObjectProperty<>(TestResultState.SUCCESS);

        private final ApiCallDetailsCache cache;


        State(String name, Date start, ApiCallDetailsCache cache) {
            this.start = start;
            this.name = name;
            this.cache = cache;
        }

        public long getDurationMs() {
            if (end.get() == null || start == null) return 0;
            return end.get().getTime() - start.getTime();
        }

        public void error(String message, Throwable e) {
            log(message, LogLineType.ERROR, e);
        }

        public void error(String message) {
            error(message, null);
        }

        public void warn(String message, Throwable e) {
            log(message, LogLineType.WARN, e);
        }

        public void warn(String message) {
            warn(message, null);
        }

        public void info(String message, Throwable e) {
            log(message, LogLineType.NOTE, e);
        }

        public void info(String message) {
            info(message, null);
        }

        public void debug(String message, Throwable e) {
            log(message, LogLineType.DEBUG, e);
        }

        public void debug(String message) {
            debug(message, null);
        }

        public void log(String message, LogLineType logLineType) {
            this.log(message, logLineType, null);
        }

        public void log(String message, LogLineType logLineType, Throwable e) {
            assert logLineType != null;
            assert message != null;
            log(new LogLine(logLineType, message, e));
        }

        public void log(LogLine logLine) {
            logs.add(logLine);
        }

        @Override
        public void onResult(ApiCallDetails result) {
            logCall(cache.registerApiCallDetails(result));
        }

        public void logCall(ApiCallDetailsRef callDetails) {
            calls.add(callDetails);
        }

        @XmlElement(name = "statestate")
        public TestResultState getState() {
            return testResultState.get();
        }

        public void setState(TestResultState state) {
            this.testResultState.set(state);
        }

        @Override
        public String toString() {
            return "State{" +
                    "" + name + '\'' +
                    '}';
        }

        public List<ApiCallDetailsRef> getCalls() {
            return calls;
        }

        public ObservableList<LogLine> getLogLines() {
            return FXCollections.unmodifiableObservableList(logs);
        }

        public String getName() {
            return name;
        }

        public Date getStateStart() {
            return start;
        }

        @XmlElement(name = "end")
        public Date getEnd() {
            return end.get();
        }

        void setEnd(Date d) {
            end.setValue(d);
        }

        public ObjectProperty<Date> endProperty() {
            return end;
        }

        public ObjectProperty<TestResultState> stateProperty() {
            return testResultState;
        }
    }


}
