package be.iminds.ilabt.jfed.highlevel.model;

import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.authority.*;
import be.iminds.ilabt.jfed.util.GeniUrn;
import javafx.application.Platform;
import javafx.beans.property.ReadOnlyListProperty;
import javafx.beans.property.ReadOnlyListWrapper;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Singleton;
import java.net.URL;
import java.util.*;

/**
 * AuthorityList acts as a wrapper around AuthorityListModel to provide JavaFX-features
 * <p/>
 * List of EasyAuthorityInfo objects
 * <p/>
 * listeners can be registered to listen for updates
 * <p/>
 * TODO: This needs to be rewritten or at least tested, to make sure load and save work correctly
 */
@Singleton
public class AuthorityList implements AuthorityListModel.AuthorityListModelListener {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(AuthorityList.class);

    private final ReadOnlyListWrapper<AuthorityInfo> infos =
            new ReadOnlyListWrapper<>(FXCollections.<AuthorityInfo>observableArrayList());
    private final Map<String, AuthorityInfo> infosByUrn = new HashMap<>();

    private final AuthorityListModel authorityListModel;
    private final AuthorityFinder authorityFinder;

    /**
     * constructor is limited to package, use EasyModel to get the EasyAuthorityList
     */
    @Inject
    public AuthorityList(AuthorityListModel authorityListModel, AuthorityFinder authorityFinder) {
        this.authorityListModel = authorityListModel;
        this.authorityFinder = authorityFinder;
        authorityListModel.addAuthorityListModelListener(this);
        onAuthorityListChanged();

        synchronized (infos) {
            for (AuthorityInfo authorityInfo : infos) {
                authorityInfo.markSaved();
            }
        }
    }

    public AuthorityFinder finder() {
        return authorityFinder;
    }

    public void add(AuthorityInfo easyAuthorityInfo) {
        synchronized (infos) {
            infos.add(easyAuthorityInfo);
            infosByUrn.put(easyAuthorityInfo.getSfaAuthority().getUrnString(), easyAuthorityInfo);
        }
    }

    public AuthorityInfo add(String urn) {
        SfaAuthority sfaAuthority;
        try {
            sfaAuthority = new SfaAuthority(urn);
        } catch (JFedException e) {
            LOG.warn("Error creating SfaAuthority: " + e.getMessage(), e);
            return null;
        }
        authorityListModel.addAuthority(sfaAuthority);

        return get(sfaAuthority);
    }

    public AuthorityInfo add(SfaAuthority newAuth) {
        assert authorityListModel.getByUrnExact(newAuth.getUrnString()) == null;
        authorityListModel.addAuthority(newAuth);

        return get(newAuth);
    }

    /**
     * get (and add if non existing)
     */
    public AuthorityInfo get(SfaAuthority geniAuthority) {
        if (geniAuthority == null)
            return null;

        AuthorityInfo existing = infosByUrn.get(geniAuthority.getUrn().getValue());

        if (existing == null) {
            existing = new AuthorityInfo(this, geniAuthority, infos.size());
            add(existing);
        }

        return existing;
    }


    public AuthorityListModel getAuthorityListModel() {
        return authorityListModel;
    }

    public AuthorityInfo getByUrn(GeniUrn urn, AuthorityListModel.SubAuthMatchAllowed matchAllowed, AuthorityListModel.SubAuthMatchPreference matchPreferred, boolean ignoreUrnTypeAndName) {
        return get(authorityListModel.getByUrn(urn, matchAllowed, matchPreferred, ignoreUrnTypeAndName));
    }

    public AuthorityInfo getByUrn(String urnString, AuthorityListModel.SubAuthMatchAllowed matchAllowed, AuthorityListModel.SubAuthMatchPreference matchPreferred, boolean ignoreUrnTypeAndName) {
        return get(authorityListModel.getByUrn(urnString, matchAllowed, matchPreferred, ignoreUrnTypeAndName));
    }

    public AuthorityInfo getByUrn(String urnString, AuthorityListModel.SubAuthMatchAllowed matchAllowed, AuthorityListModel.SubAuthMatchPreference matchPreferred) {
        return get(authorityListModel.getByUrn(urnString, matchAllowed, matchPreferred));
    }

    public AuthorityInfo getByUrn(GeniUrn urn, AuthorityListModel.SubAuthMatchAllowed matchAllowed, AuthorityListModel.SubAuthMatchPreference matchPreferred) {
        return get(authorityListModel.getByUrn(urn, matchAllowed, matchPreferred));
    }

    public AuthorityInfo getByUrnExact(String urnString) {
        return get(authorityListModel.getByUrnExact(urnString));
    }

    public AuthorityInfo getByUrnExact(GeniUrn urn) {
        return get(authorityListModel.getByUrnExact(urn));
    }

    public AuthorityInfo getFromAnyUrn(String urnString, AuthorityListModel.SubAuthMatchAllowed matchAllowed, AuthorityListModel.SubAuthMatchPreference matchPreferred) {
        return get(authorityListModel.getFromAnyUrn(urnString, matchAllowed, matchPreferred));
    }

    public AuthorityInfo getFromAnyUrn(GeniUrn urn, AuthorityListModel.SubAuthMatchAllowed matchAllowed, AuthorityListModel.SubAuthMatchPreference matchPreferred) {
        return get(authorityListModel.getFromAnyUrn(urn, matchAllowed, matchPreferred));
    }

    public AuthorityInfo getFromAnyUrnExact(String urnString) {
        return get(authorityListModel.getFromAnyUrnExact(urnString));
    }

    public AuthorityInfo getFromAnyUrnExact(GeniUrn urn) {
        return get(authorityListModel.getFromAnyUrnExact(urn));
    }


    public AuthorityInfo getByUrl(String url) {
        for (SfaAuthority authority : authorityListModel.getAuthorities()) {
            for (URL u : authority.getUrls().values()) {
                if (u.toExternalForm().equals(url)) {
                    return get(authority);
                }
            }
        }
        return null;
    }


    public ReadOnlyListProperty<AuthorityInfo> authorityInfosProperty() {
        return infos.getReadOnlyProperty();
    }

    public ObservableList<AuthorityInfo> getAuthorityInfos() {
        return infos.get();
    }

    @Override
    public void onAuthorityListChanged() {
        Platform.runLater(AuthorityList.this::onAuthorityListChanged_internal);
    }

    private void onAuthorityListChanged_internal() {
        //check if internal mapping is ok (could have been removes, updates and adds)
        //this is not a fast operation. But normally, it should not happen too often or on large sets
        Map<String, AuthorityInfo> removedAuthsByUrn = new HashMap<>(infosByUrn);
        Set<String> origKeys = infosByUrn.keySet();

        for (SfaAuthority sfaAuthority : authorityListModel.getAuthorities()) {
            removedAuthsByUrn.remove(sfaAuthority.getUrnString());

            AuthorityInfo ai = get(sfaAuthority);
            assert ai != null;
        }

        for (AuthorityInfo aiToBeRemoved : removedAuthsByUrn.values()) {
            infosByUrn.remove(aiToBeRemoved.getSfaAuthority().getUrnString());
            synchronized (infos) {
                infos.remove(aiToBeRemoved);
            }
        }

        Set<String> added = new HashSet<>(infosByUrn.keySet());
        added.removeAll(origKeys);

        assert authorityListModel.getAuthorities().size() == infos.size() :
                "authorityListModel.getAuthorities().size()=" + authorityListModel.getAuthorities().size() +
                        " =?= infos.size()=" + infos.size() + " =?= infosByUrn.size()=" + infosByUrn.size() +
                        "   removed: " + removedAuthsByUrn.keySet() + "   added: " + added;
        assert infosByUrn.size() == infos.size() :
                "infosByUrn.size()=" + infosByUrn.size() +
                        " == infos.size()=" + infos.size() +
                        "   removed: " + removedAuthsByUrn.keySet() + "   added: " + added;
    }

    public void save() {
        synchronized (infos) {
            for (AuthorityInfo auth : infos) {
                //assume uncommitted changes need to be saved
                boolean committed = auth.committedProperty().get();
                if (!committed) {
                    auth.commit();
                }
                assert auth.committedProperty().get();
                assert auth.isSame();
            }

            StoredAuthorityList.save(JFedAuthorityList.getDefaultAuthoritiesFile(), authorityListModel);

            //mark all as saved
            infos.forEach(be.iminds.ilabt.jfed.highlevel.model.AuthorityInfo::markSaved);
        }
    }

    public void resetToInternalDefaults() {
        AuthorityListModel tmpAuthorityListModel = new AuthorityListModel();
        BuiltinAuthorityList.load(tmpAuthorityListModel);

        List<String> urnsToRemove = new ArrayList<>();
        urnsToRemove.addAll(infosByUrn.keySet());

        for (SfaAuthority auth : tmpAuthorityListModel.getAuthorities()) {
            AuthorityInfo ai = infosByUrn.get(auth.getUrnString());
            if (ai != null) {
                ai.setTo(auth);
                urnsToRemove.remove(auth.getUrnString());
            }
        }

        for (String urn : urnsToRemove) {
            AuthorityInfo ai = infosByUrn.get(urn);
            assert ai != null;
            ai.delete();
        }
    }
}
