package be.iminds.ilabt.jfed.util;

import be.iminds.ilabt.jfed.log.ApiCallDetails;
import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.log.ResultListener;
import javafx.application.Platform;

import javax.inject.Singleton;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CyclicBarrier;

/**
 * JavaFXLogger exactly the same as Logger, but always handles events in JavaFX thread
 */
@Singleton
public class JavaFXLogger extends Logger {
    private static org.slf4j.Logger LOG = org.slf4j.LoggerFactory.getLogger(JavaFXLogger.class);

    private boolean synchronous;

    /**
     * Creates a synchronous JavaFXLogger
     */
    public JavaFXLogger(){
        this(true);
    }


    /**
     * @param synchronous should fireResult be synchronous, or should it return after all results have been processed
     */
    public JavaFXLogger(boolean synchronous) {
        this.synchronous = synchronous;
    }


    @Override
    public synchronized void fireResult(final ApiCallDetails reply) {
        if (reply.getAuthority() == null) {
            try {
                throw new RuntimeException("reply authority is null");
            } catch (Exception e) {
//                LOG.error("Saw a reply with a null authority. Will not fire event, to prevent trouble!\n\nReply:\n"+reply.toString(), e);
                LOG.error("Saw a reply with a null authority. Will fire event, but this is a bug!\n\nReply:\n"+reply.toString(), e);
            }
//            return;
        }

        if (!synchronous) {
            //asynchronous event processing
            Platform.runLater(new Runnable() {
                @Override
                public void run() {
                    for (ResultListener l : resultListeners)
                        try {
                            l.onResult(reply);
                        } catch (Throwable t) {
                            LOG.error("Error while asynchronously processing Logger ApiCallDetails reply on JavaFX thread. (will be ignored)", t);
                        }
                }
            });
            return;
        }

        if (Platform.isFxApplicationThread())
            for (ResultListener l : resultListeners)
                try {
                    l.onResult(reply);
                } catch (Throwable t) {
                    LOG.error("Error while synchronously processing Logger ApiCallDetails reply on JavaFX thread. (will be ignored)", t);
                }
        else {
            //synchronous event processing
            final CyclicBarrier barrier = new CyclicBarrier(2);

            Platform.runLater(new Runnable() {
                @Override
                public void run() {
                    for (ResultListener l : resultListeners)
                        try {
                            l.onResult(reply);
                        } catch (Throwable e) {
                            //ignore, because we want unlocking below to happen!
                            LOG.error("Error while synchronously processing Logger ApiCallDetails reply on JavaFX thread. (will be ignored)", e);
                        }

                    try {
                        barrier.await();
                    } catch (InterruptedException | BrokenBarrierException ex) {
                        return;
                    }
                }
            });

            try {
                barrier.await();
            } catch (InterruptedException | BrokenBarrierException ex) {
                return;
            }
        }
    }

    private static class MyWrappingJavaFXLogger extends JavaFXLogger {
        private final JavaFXLogger parentLogger;

        public MyWrappingJavaFXLogger(boolean synchronous, JavaFXLogger parentLogger) {
            super(synchronous);
            this.parentLogger = parentLogger;
        }

        @Override
        public synchronized void fireResult(ApiCallDetails reply) {
            //fire for all parentLogger listeners, even if they changed
            parentLogger.fireResult(reply);

            //fire for any listener added to this new wrapping logger
            super.fireResult(reply);
        }
    }

    @Override
    public JavaFXLogger getWrappingLogger() {
        return new MyWrappingJavaFXLogger(synchronous, this);
    }

    public static Logger wrap(final JavaFXLogger logger) {
        return wrap(logger, null);
    }

    public static JavaFXLogger wrap(final JavaFXLogger logger, ResultListener... resultListeners) {
        JavaFXLogger res = new MyWrappingJavaFXLogger(logger.synchronous, logger);
        for (int i = 0; i < resultListeners.length; i++)
            res.addResultListener(resultListeners[i]);
        return res;
    }
}
