package be.iminds.ilabt.jfed.highlevel.stitcher;

import be.iminds.ilabt.jfed.TestAuthoritiesModule;
import be.iminds.ilabt.jfed.call_log_output.SerializableApiCallDetails;
import be.iminds.ilabt.jfed.log.cache.ApiCallDetailsCache;
import be.iminds.ilabt.jfed.log.cache.ApiCallDetailsRef;
import be.iminds.ilabt.jfed.lowlevel.GeniAMResponseCode;
import be.iminds.ilabt.jfed.lowlevel.TestCertHelper;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.lowlevel.stitching.StitchingDirector;
import be.iminds.ilabt.jfed.lowlevel.stitching.VlanRange;
import javafx.embed.swing.JFXPanel;
import org.slf4j.LoggerFactory;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import javax.inject.Inject;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.*;

/*
* This is a complex test that sets up fake servers so that stitching can be tested
* */
public class ParallelStitcherTest {
    private final static org.slf4j.Logger LOG = LoggerFactory.getLogger(ParallelStitcherTest.class);

    @BeforeMethod
    public void setUp() throws Exception {
        //initialize JavaFX
        new JFXPanel();
    }

    @AfterMethod
    public void tearDown() throws Exception {

    }

    @DataProvider(name = "allStitchingSetupConfigs")
    public static Object[][] getAllStitchingSetupConfigs() {
        List<StitchingTestSetupConfig> configs = new ArrayList<>();

        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder();
        for (final boolean fakeAm3 : new boolean[] { false, true} ) {
            confBuilder.setHasFakeAm3(fakeAm3);
            for (final boolean am2DependsOnAm1 : new boolean[] { false, true} ) {
                confBuilder.setAm2DependsOnAm1(am2DependsOnAm1);
                for (final StitchingTestSetupConfigBuilder.AmVersion am1v :
                        new StitchingTestSetupConfigBuilder.AmVersion[]{StitchingTestSetupConfigBuilder.AmVersion.V2, StitchingTestSetupConfigBuilder.AmVersion.V3}) {
                    confBuilder.setAmVersion(1, am1v);
                    for (final StitchingTestSetupConfigBuilder.AmVersion am2v :
                            new StitchingTestSetupConfigBuilder.AmVersion[]{StitchingTestSetupConfigBuilder.AmVersion.V2, StitchingTestSetupConfigBuilder.AmVersion.V3}) {
                        confBuilder.setAmVersion(2, am2v);
                        for (final boolean am1supportsAny : new boolean[]{false, true}) {
                            confBuilder.setSupportsAny(1, am1supportsAny);
                            for (final boolean am2supportsAny : new boolean[]{false, true}) {
                                confBuilder.setSupportsAny(2, am2supportsAny);
                                StitchingTestSetupConfig conf = confBuilder.build();
                                configs.add(conf);
                            }
                        }
                    }
                }
            }
        }

        Object[][] res = new Object[configs.size()][];
        int i = 0;
        for (StitchingTestSetupConfig c : configs)
            res[i++] = new Object[] { c };
        return res;
    }

    @DataProvider(name = "allNoVlanStitchingSetupConfigs")
    public static Object[][] getAllNoVlanStitchingSetupConfigs() {
        List<StitchingTestSetupConfig> configs = new ArrayList<>();

        for (int amIndex : new int[] {1, 2}) {
            StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder();
            confBuilder.addSpecialReply(amIndex, StitchingTestSetupConfig.SpecialReply.ERROR_NO_VLANS);
            confBuilder.setAm2DependsOnAm1(true/*am2DependsOnAm1*/);

            for (final StitchingTestSetupConfigBuilder.AmVersion am1v :
                    new StitchingTestSetupConfigBuilder.AmVersion[] { StitchingTestSetupConfigBuilder.AmVersion.V2, StitchingTestSetupConfigBuilder.AmVersion.V3} ) {
                confBuilder.setAmVersion(1, am1v);
                confBuilder.setAmVersion(2, am1v);
                for (final boolean am1supportsAny : new boolean[]{false, true}) {
                    confBuilder.setSupportsAny(1, am1supportsAny);
                    for (final boolean am2supportsAny : new boolean[]{false, true}) {
                        confBuilder.setSupportsAny(2, am2supportsAny);
                        StitchingTestSetupConfig conf = confBuilder.build();
                        configs.add(conf);
                    }
                }
            }

        }

        Object[][] res = new Object[configs.size()][];
        int i = 0;
        for (StitchingTestSetupConfig c : configs)
            res[i++] = new Object[] { c };
        return res;
    }

    /**
     * All combinations of 2 servers with AMv2 or V3 and any support or not
     * Tests using this will lickely change somerthing in the config before using it.
     * */
    @DataProvider(name = "all32anyCombinations")
    public static Object[][] getAll32anyCombinations() {
        List<StitchingTestSetupConfig> configs = new ArrayList<>();

        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder();
//        confBuilder.setAm2DependsOnAm1(true/*am2DependsOnAm1*/);
//        confBuilder.setSpecialCase(StitchingTestSetupConfig.SpecialCase.am1NoMatchingVlansAm2);

        for (final StitchingTestSetupConfigBuilder.AmVersion am1v :
                new StitchingTestSetupConfigBuilder.AmVersion[] { StitchingTestSetupConfigBuilder.AmVersion.V2, StitchingTestSetupConfigBuilder.AmVersion.V3} ) {
            for (final StitchingTestSetupConfigBuilder.AmVersion am2v :
                    new StitchingTestSetupConfigBuilder.AmVersion[] { StitchingTestSetupConfigBuilder.AmVersion.V2, StitchingTestSetupConfigBuilder.AmVersion.V3} ) {
                confBuilder.setAmVersion(1, am1v);
                confBuilder.setAmVersion(2, am2v);
                for (final boolean am1supportsAny : new boolean[]{false, true}) {
                    confBuilder.setSupportsAny(1, am1supportsAny);
                    for (final boolean am2supportsAny : new boolean[]{false, true}) {
                        confBuilder.setSupportsAny(2, am2supportsAny);
                        StitchingTestSetupConfig conf = confBuilder.build();
                        configs.add(conf);
                    }
                }
            }
        }

        Object[][] res = new Object[configs.size()][];
        int i = 0;
        for (StitchingTestSetupConfig c : configs)
            res[i++] = new Object[] { c };
        return res;
    }

    @DataProvider(name = "33anyany-dep")
    public static Object[][] get33anyany() {
        List<StitchingTestSetupConfig> configs = new ArrayList<>();

        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder();
        confBuilder.setAm2DependsOnAm1(true/*am2DependsOnAm1*/);
//        confBuilder.setSpecialCase(StitchingTestSetupConfig.SpecialCase.am1NoMatchingVlansAm2);

        confBuilder.setAmVersion(1, StitchingTestSetupConfigBuilder.AmVersion.V3);
        confBuilder.setAmVersion(2, StitchingTestSetupConfigBuilder.AmVersion.V3);
        confBuilder.setSupportsAny(1, true);
        confBuilder.setSupportsAny(2, true);
        StitchingTestSetupConfig conf = confBuilder.build();
        configs.add(conf);


        Object[][] res = new Object[configs.size()][];
        int i = 0;
        for (StitchingTestSetupConfig c : configs)
            res[i++] = new Object[] { c };
        return res;
    }

    @DataProvider(name = "22noanynoany-no_dep")
    public static Object[][] get22noanynoany() {
        List<StitchingTestSetupConfig> configs = new ArrayList<>();

        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder();
        confBuilder.setAm2DependsOnAm1(false/*am2DependsOnAm1*/);
//        confBuilder.setSpecialCase(StitchingTestSetupConfig.SpecialCase.am1NoMatchingVlansAm2);

        confBuilder.setAmVersion(1, StitchingTestSetupConfigBuilder.AmVersion.V2);
        confBuilder.setAmVersion(2, StitchingTestSetupConfigBuilder.AmVersion.V2);
        confBuilder.setSupportsAny(1, false);
        confBuilder.setSupportsAny(2, false);
        StitchingTestSetupConfig conf = confBuilder.build();
        configs.add(conf);


        Object[][] res = new Object[configs.size()][];
        int i = 0;
        for (StitchingTestSetupConfig c : configs)
            res[i++] = new Object[] { c };
        return res;
    }

    /**
     * All combinations of 2 servers with AMv2 or V3 and any support or not
     * Tests using this will lickely change somerthing in the config before using it.
     * */
    @DataProvider(name = "all32Combinations")
    public static Object[][] getAll32Combinations() {
        List<StitchingTestSetupConfig> configs = new ArrayList<>();

        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder();
//        confBuilder.setAm2DependsOnAm1(true/*am2DependsOnAm1*/);
//        confBuilder.setSpecialCase(StitchingTestSetupConfig.SpecialCase.am1NoMatchingVlansAm2);

        for (final StitchingTestSetupConfigBuilder.AmVersion am1v :
                new StitchingTestSetupConfigBuilder.AmVersion[] { StitchingTestSetupConfigBuilder.AmVersion.V2, StitchingTestSetupConfigBuilder.AmVersion.V3} ) {
            for (final StitchingTestSetupConfigBuilder.AmVersion am2v :
                    new StitchingTestSetupConfigBuilder.AmVersion[]{StitchingTestSetupConfigBuilder.AmVersion.V2, StitchingTestSetupConfigBuilder.AmVersion.V3}) {
                confBuilder.setAmVersion(1, am1v);
                confBuilder.setAmVersion(2, am2v);
                confBuilder.setSupportsAny(1, true);
                confBuilder.setSupportsAny(2, true);
                StitchingTestSetupConfig conf = confBuilder.build();
                configs.add(conf);
            }
        }

        Object[][] res = new Object[configs.size()][];
        int i = 0;
        for (StitchingTestSetupConfig c : configs)
            res[i++] = new Object[] { c };
        return res;
    }


    @Test
    public void testScsFailure() throws InterruptedException {
        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder();

        confBuilder.setSpecialCase(StitchingTestSetupConfig.SpecialCase.scsFailure);
        confBuilder.setAm2DependsOnAm1(true);
        confBuilder.setAmVersion(1, StitchingTestSetupConfigBuilder.AmVersion.V3);
        confBuilder.setAmVersion(2, StitchingTestSetupConfigBuilder.AmVersion.V3);
        confBuilder.setSupportsAny(1, true);
        confBuilder.setSupportsAny(2, true);
        StitchingTestSetupConfig conf = confBuilder.build();

        //check if builder works correctly
        assertThat(conf.hasSpecialCase(StitchingTestSetupConfig.SpecialCase.scsFailure), equalTo(true));

        StitchingTestSetup setup = new StitchingTestSetup(conf);
        setup.testStitching();
        setup.outputToFile();
        setup.assertAllTestBasics();
        assertThat("Did not give up on stitching. globalPhase is " + setup.getParallelStitcher().getGlobalPhase(), setup.isGaveUpOnStitching(), is(true));
        assertThat(setup.getParallelStitcher().getGlobalPhase(), equalTo(ParallelStitcher.GlobalPhase.FAILED_PERMANENT));
        assertThat("Expected only the failed SCS call", setup.getStitchingReport().getCalls(), hasSize(1));
    }

    @Test(dataProvider = "allStitchingSetupConfigs")
//    @Test(dataProvider = "33anyany-dep") //test with single one
//    @Test(dataProvider = "22noanynoany-no_dep") //test with single one
    public void parallelStitch(StitchingTestSetupConfig stitchingTestSetupConfig) throws InterruptedException {
        StitchingTestSetup setup = new StitchingTestSetup(stitchingTestSetupConfig);
        setup.testStitching();
        setup.outputToFile();
        setup.assertSuccess(stitchingTestSetupConfig.hasSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1)/*allowDelete*/);

        List<ApiCallDetailsRef> allocateStateCallsRef = setup.getStitchingReport().getAllStateCalls("ALLOCATE");

        List<SerializableApiCallDetails> allocateStateCalls = allocateStateCallsRef.stream().map(setup.getApiCallDetailsCache()::getApiCallDetails).collect(Collectors.toList());

        assertThat(allocateStateCalls, hasSize(greaterThan(1)));
        SerializableApiCallDetails firstAllocate = allocateStateCalls.get(0);
        assertThat(firstAllocate.getGeniMethodName(), isOneOf("CreateSliver", "Allocate"));
        assertThat(firstAllocate.getXmlRpcRequestHumanReadableString(), containsString("<suggestedVLANRange>any</suggestedVLANRange>"));
        if (firstAllocate.getAuthorityUrn().equals(TestCertHelper.am1Urn) && !stitchingTestSetupConfig.am(1).supportsAny) {
            if (setup.testAm1Auth.hasSpecialCase(SfaAuthority.SpecialCase.STITCHING_SUPPORTS_ANY))
                assertThat(firstAllocate.getGeniResponseOutput(), equalTo("ERROR: vlan range any is invalid"));
            else
                assertThat(firstAllocate.getGeniResponseOutput(), not(equalTo("ERROR: vlan range any is invalid")));
        }
        if (firstAllocate.getAuthorityUrn().equals(TestCertHelper.am2Urn) && !stitchingTestSetupConfig.am(2).supportsAny) {
            if (setup.testAm2Auth.hasSpecialCase(SfaAuthority.SpecialCase.STITCHING_SUPPORTS_ANY))
                assertThat(firstAllocate.getGeniResponseOutput(), equalTo("ERROR: vlan range any is invalid"));
            else
                assertThat(firstAllocate.getGeniResponseOutput(), not(equalTo("ERROR: vlan range any is invalid")));
        }

        System.out.flush();
        System.err.flush();
        Thread.yield();
        System.out.flush();
        System.err.flush();
        Thread.yield();
        Thread.yield();
    }

    @Test
    public void testParallelStitcherWithFakeAm3() throws InterruptedException {
        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder();

        confBuilder.use3Ams(false);
        confBuilder.setHasFakeAm3(true);

        confBuilder.setAm2DependsOnAm1(true);
        confBuilder.setAmVersion(1, StitchingTestSetupConfigBuilder.AmVersion.V3);
        confBuilder.setAmVersion(2, StitchingTestSetupConfigBuilder.AmVersion.V3);
        confBuilder.setSupportsAny(1, true);
        confBuilder.setSupportsAny(2, true);
        StitchingTestSetupConfig conf = confBuilder.build();

        StitchingTestSetup setup = new StitchingTestSetup(conf);
        setup.testStitching();
        setup.outputToFile();
        setup.assertSuccess(true/*allowDelete*/);

        //there should be at least 3 allocate calls. (1 failed and 2 successfull)
        List<ApiCallDetailsRef> allocateStateCallsRef = setup.getStitchingReport().getAllStateCalls("ALLOCATE");
        List<SerializableApiCallDetails> allocateStateCalls = allocateStateCallsRef.stream()
                .map(setup.getApiCallDetailsCache()::getApiCallDetails).collect(Collectors.toList());
        List<SerializableApiCallDetails> deleteCalls = allocateStateCalls.stream().
                filter(c -> c.getGeniMethodName().startsWith("Delete")).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().equals("Allocate") || c.getGeniMethodName().equals("CreateSliver")).collect(Collectors.toList());

        assertThat(allocateCalls, hasSize(greaterThanOrEqualTo(3)));
    }

    @Test
    public void testParallelStitcherWithNonInvolvedAm3() throws InterruptedException {
        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder();

        confBuilder.use3Ams(true);
        confBuilder.setNoStitching(3, true);

        confBuilder.setAm2DependsOnAm1(true);
        confBuilder.setAmVersion(1, StitchingTestSetupConfigBuilder.AmVersion.V3);
        confBuilder.setAmVersion(2, StitchingTestSetupConfigBuilder.AmVersion.V3);
        confBuilder.setAmVersion(3, StitchingTestSetupConfigBuilder.AmVersion.V3);
        confBuilder.setSupportsAny(1, true);
        confBuilder.setSupportsAny(2, true);
        StitchingTestSetupConfig conf = confBuilder.build();

        //check if builder works correctly
        assertThat(conf.amCount(), is(3));
        assertThat(conf.am(3).noStitching, is(true));

        StitchingTestSetup setup = new StitchingTestSetup(conf);
        setup.testStitching();
        setup.outputToFile();
        setup.assertSuccess(true/*allowDelete*/);

        //there should be at least 3 allocate calls. (1 failed and 2 successfull)
        List<ApiCallDetailsRef> allocateStateCallsRef = setup.getStitchingReport().getAllStateCalls("ALLOCATE");

        List<SerializableApiCallDetails> allocateStateCalls = allocateStateCallsRef.stream().map(setup.getApiCallDetailsCache()::getApiCallDetails).collect(Collectors.toList());


        List<SerializableApiCallDetails> deleteCalls = allocateStateCalls.stream().
                filter(c -> c.getGeniMethodName().startsWith("Delete")).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().equals("Allocate") || c.getGeniMethodName().equals("CreateSliver")).collect(Collectors.toList());

        assertThat(allocateCalls, hasSize(greaterThanOrEqualTo(3)));


        Map<String, Integer> am3CommandCounts = setup.getStitchingReport().getCalls().stream()
                .map(setup.getApiCallDetailsCache()::getApiCallDetails)
                .filter(c -> c.getServerUrl().equals(TestAuthoritiesModule.urlC_am3))
                .collect(Collectors.groupingBy(c -> c.getGeniMethodName(), Collectors.summingInt(c -> 1)));
        System.out.println("am3CommandCounts=" + am3CommandCounts);
        assertThat(am3CommandCounts, hasEntry(is("Allocate"), is(1)));
        assertThat(am3CommandCounts, hasEntry(is("Provision"), is(1)));
        assertThat(am3CommandCounts, hasEntry(is("PerformOperationalAction"), is(1)));
        assertThat(am3CommandCounts, hasEntry(is("Status"), is(greaterThanOrEqualTo(1))));
        assertThat(am3CommandCounts, not(hasKey("Delete")));
    }

    @Test(dataProvider = "all32anyCombinations")
    public void testParallelStitcherWithOneTimeGenericRetryError(StitchingTestSetupConfig stitchingTestSetupConfigBase) throws InterruptedException {
        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder(stitchingTestSetupConfigBase);
        confBuilder.setAm2DependsOnAm1(true);
        confBuilder.addSpecialReply(2, StitchingTestSetupConfig.SpecialReply.ERROR_ONETIME_GENERIC_ALLOCATE);
        StitchingTestSetupConfig stitchingTestSetupConfig = confBuilder.build();
        assert stitchingTestSetupConfig.hasSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1);
        assert stitchingTestSetupConfig.hasSpecialReply(2, StitchingTestSetupConfig.SpecialReply.ERROR_ONETIME_GENERIC_ALLOCATE);

        StitchingTestSetup setup = new StitchingTestSetup(stitchingTestSetupConfig);
        setup.testStitching();
        setup.outputToFile();
        setup.assertSuccess(true/*allowDelete*/);

        //there should be at least 3 allocate calls. (1 failed and 2 successfull)
        List<ApiCallDetailsRef> allocateStateCallsRef = setup.getStitchingReport().getAllStateCalls("ALLOCATE");

        List<SerializableApiCallDetails> allocateStateCalls = allocateStateCallsRef.stream().map(setup.getApiCallDetailsCache()::getApiCallDetails).collect(Collectors.toList());
        List<SerializableApiCallDetails> deleteCalls = allocateStateCalls.stream().
                filter(c -> c.getGeniMethodName().startsWith("Delete")).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().equals("Allocate") || c.getGeniMethodName().equals("CreateSliver")).collect(Collectors.toList());

        assertThat(allocateCalls, hasSize(greaterThanOrEqualTo(3)));
    }


    @Test(dataProvider = "all32anyCombinations")
//    @Test(dataProvider = "33anyany-dep") //test with single one
    public void testParallelStitcherWithOneAdvertisementUpToDate(StitchingTestSetupConfig stitchingTestSetupConfigBase) throws InterruptedException {
        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder(stitchingTestSetupConfigBase);
        confBuilder.setAm2DependsOnAm1(true);
        confBuilder.setAdvertisementUpToDate(1, false);
        confBuilder.setAdvertisementUpToDate(2, true);
        StitchingTestSetupConfig stitchingTestSetupConfig = confBuilder.build();
        assert stitchingTestSetupConfig.hasSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1);

        StitchingTestSetup setup = new StitchingTestSetup(stitchingTestSetupConfig);
        setup.testStitching();
        setup.outputToFile();
        setup.assertSuccess(true/*allowDelete*/);

        List<ApiCallDetailsRef> allocateStateCallsRef = setup.getStitchingReport().getAllStateCalls("ALLOCATE");

        List<SerializableApiCallDetails> allocateStateCalls = allocateStateCallsRef.stream().map(setup.getApiCallDetailsCache()::getApiCallDetails).collect(Collectors.toList());
        List<SerializableApiCallDetails> deleteCalls = allocateStateCalls.stream().
                filter(c -> c.getGeniMethodName().startsWith("Delete")).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().equals("Allocate") || c.getGeniMethodName().equals("CreateSliver")).collect(Collectors.toList());
        List<SerializableApiCallDetails> listResourcesCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().equals("ListResources")).collect(Collectors.toList());

        if (!stitchingTestSetupConfig.am(1).supportsAny && stitchingTestSetupConfig.am(1).advUpToDate)
            assertThat(listResourcesCalls, hasSize(2));
        else
            assertThat(listResourcesCalls, hasSize(1));

        //if any not supported, just one call more, as smartness will still be used.
        int callsNeeded = 2;
        if (!stitchingTestSetupConfig.am(1).supportsAny && setup.testAm1Auth.hasSpecialCase(SfaAuthority.SpecialCase.STITCHING_SUPPORTS_ANY))
            callsNeeded += 1;
        if (!stitchingTestSetupConfig.am(1).supportsAny && !stitchingTestSetupConfig.am(1).advUpToDate)
            callsNeeded += 2;

        assertThat(allocateCalls, hasSize(equalTo(callsNeeded)));

        assertThat(deleteCalls, hasSize(0));
    }


    //TODO  //do an additional ListResources call on am 1 after the call that shows
    //TODO  //    it doesn't support any but does support up to date adv.
    //TODO  // Then it will know which vlan to pick
    //TODO
    //TODO  //This is low prioirty, as this scenario doesn't occur on any AM: All AMs that support up to date adv also support any.
    //TODO
    //TODO After this is implemented, re-enable this test to test it
//    @Test(dataProvider = "all32anyCombinations")
////    @Test(dataProvider = "33anyany-dep") //test with single one
//    public void testParallelStitcherWithBothAdvertisementUpToDate(StitchingTestSetupConfig stitchingTestSetupConfigBase) throws InterruptedException {
//        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder(stitchingTestSetupConfigBase);
//        confBuilder.setAm2DependsOnAm1(true);
//        confBuilder.setAdvertisementUpToDate(1, true);
//        confBuilder.setAdvertisementUpToDate(2, true);
//        StitchingTestSetupConfig stitchingTestSetupConfig = confBuilder.build();
//        assert stitchingTestSetupConfig.hasSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1);
//
//        StitchingTestSetup setup = new StitchingTestSetup(stitchingTestSetupConfig);
//        setup.testStitching();
//        setup.outputToFile();
//        setup.assertSuccess(true/*allowDelete*/);
//
//        List<SerializableApiCallDetails> allocateStateCalls = setup.getStitchingReport().getAllStateCalls("ALLOCATE");
//        List<SerializableApiCallDetails> deleteCalls = allocateStateCalls.stream().
//                filter(c -> c.getGeniMethodName().startsWith("Delete")).collect(Collectors.toList());
//        List<SerializableApiCallDetails> allocateCalls = allocateStateCalls.stream().filter(
//                c -> c.getGeniMethodName().equals("Allocate") || c.getGeniMethodName().equals("CreateSliver")).collect(Collectors.toList());
//        List<SerializableApiCallDetails> listResourcesCalls = allocateStateCalls.stream().filter(
//                c -> c.getGeniMethodName().equals("ListResources")).collect(Collectors.toList());
//
//        if (!stitchingTestSetupConfig.am(1).supportsAny && stitchingTestSetupConfig.am(1).advUpToDate)
//            assertThat(listResourcesCalls, hasSize(2));
//        else
//            assertThat(listResourcesCalls, hasSize(1));
//
//        //if any not supported, just one call more, as smartness will still be used.
//        int callsNeeded = 2;
//        if (!stitchingTestSetupConfig.am(1).supportsAny)
//            callsNeeded += 1;
//        if (!stitchingTestSetupConfig.am(1).supportsAny && !stitchingTestSetupConfig.am(1).advUpToDate)
//            callsNeeded += 2;
//
//        assertThat(allocateCalls, hasSize(equalTo(callsNeeded)));
//
//        assertThat(deleteCalls, hasSize(0));
//    }

    @Test(dataProvider = "all32anyCombinations")
    public void testParallelStitcherWithOneTimeoutError(StitchingTestSetupConfig stitchingTestSetupConfigBase) throws InterruptedException {
        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder(stitchingTestSetupConfigBase);
        confBuilder.setAm2DependsOnAm1(true);
        confBuilder.addSpecialReply(2, StitchingTestSetupConfig.SpecialReply.ERROR_ALLOCATE_ONE_TIMEOUT);
        StitchingTestSetupConfig stitchingTestSetupConfig = confBuilder.build();
        assert stitchingTestSetupConfig.hasSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1);
        assert stitchingTestSetupConfig.hasSpecialReply(2, StitchingTestSetupConfig.SpecialReply.ERROR_ALLOCATE_ONE_TIMEOUT);

        StitchingTestSetup setup = new StitchingTestSetup(stitchingTestSetupConfig);
        setup.testStitching();
        setup.outputToFile();
        setup.assertSuccess(true/*allowDelete*/);

        //there should be at least 3 allocate calls. (1 failed and 2 successfull)
        List<ApiCallDetailsRef> allocateStateCallsRef = setup.getStitchingReport().getAllStateCalls("ALLOCATE");

        List<SerializableApiCallDetails> allocateStateCalls = allocateStateCallsRef.stream().map(setup.getApiCallDetailsCache()::getApiCallDetails).collect(Collectors.toList());
        List<SerializableApiCallDetails> deleteCalls = allocateStateCalls.stream().
                filter(c -> c.getGeniMethodName().startsWith("Delete")).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().equals("Allocate") || c.getGeniMethodName().equals("CreateSliver")).
                filter(c -> c.getGeniResponseCode() == null || !c.getGeniResponseCode().equals(GeniAMResponseCode.INTERNAL_NONGENI_ERROR.getCode())).
                collect(Collectors.toList());

        assertThat(allocateCalls, hasSize(greaterThanOrEqualTo(3)));
    }

    @Test(dataProvider = "all32anyCombinations")
    public void testParallelStitcherWithOneAndTwoTimeoutError(StitchingTestSetupConfig stitchingTestSetupConfigBase) throws InterruptedException {
        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder(stitchingTestSetupConfigBase);
        confBuilder.setAm2DependsOnAm1(true);
        confBuilder.addSpecialReply(2, StitchingTestSetupConfig.SpecialReply.ERROR_ALLOCATE_ONE_TIMEOUT);
        confBuilder.addSpecialReply(2, StitchingTestSetupConfig.SpecialReply.ERROR_ALLOCATE_TWO_TIMEOUT);
        StitchingTestSetupConfig stitchingTestSetupConfig = confBuilder.build();
        assert stitchingTestSetupConfig.hasSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1);
        assert stitchingTestSetupConfig.hasSpecialReply(2, StitchingTestSetupConfig.SpecialReply.ERROR_ALLOCATE_ONE_TIMEOUT);
        assert stitchingTestSetupConfig.hasSpecialReply(2, StitchingTestSetupConfig.SpecialReply.ERROR_ALLOCATE_TWO_TIMEOUT);

        StitchingTestSetup setup = new StitchingTestSetup(stitchingTestSetupConfig);

        setup.testStitching();
        setup.outputToFile();
        setup.assertFailure();

        //there should be at most 2 allocate calls. (1 for AM1 and 1 failed for AM2)
        //unless AM1 does not have any, in which case there are 2 or 3 more calls
        List<ApiCallDetailsRef> allocateStateCallsRef = setup.getStitchingReport().getAllStateCalls("ALLOCATE");

        List<SerializableApiCallDetails> allocateStateCalls = allocateStateCallsRef.stream().map(setup.getApiCallDetailsCache()::getApiCallDetails).collect(Collectors.toList());
        List<SerializableApiCallDetails> deleteCalls = allocateStateCalls.stream().
                filter(c -> c.getGeniMethodName().startsWith("Delete")).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().equals("Allocate") || c.getGeniMethodName().equals("CreateSliver")).
                filter(c -> c.getGeniResponseCode() == null || !c.getGeniResponseCode().equals(GeniAMResponseCode.INTERNAL_NONGENI_ERROR.getCode())).
                collect(Collectors.toList());
        List<SerializableApiCallDetails> nonTimedOutAllocateCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().equals("Allocate") || c.getGeniMethodName().equals("CreateSliver")).
                filter(c -> c.getHttpResponse() != null && !c.getHttpResponse().isEmpty()).
                collect(Collectors.toList());
        List<SerializableApiCallDetails> timedOutAllocateCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().equals("Allocate") || c.getGeniMethodName().equals("CreateSliver")).
                filter(c -> c.getHttpResponse() == null || c.getHttpResponse().isEmpty()).
                collect(Collectors.toList());


        if (stitchingTestSetupConfig.am(1).supportsAny)
            assertThat(nonTimedOutAllocateCalls, hasSize(equalTo(1)));
        else {
            int callsNeeded = 2; //failed AM1 calls
            callsNeeded += 1; //succesfull AM 1 call
//            callsNeeded += 2; //2 timed out AM 2 call

            if (setup.testAm1Auth.hasSpecialCase(SfaAuthority.SpecialCase.STITCHING_SUPPORTS_ANY))
                callsNeeded += 1; //failed any call to am1

            assertThat(nonTimedOutAllocateCalls, hasSize(equalTo(callsNeeded)));
        }

        assertThat(timedOutAllocateCalls, hasSize(allOf(greaterThanOrEqualTo(1), lessThanOrEqualTo(3))));
    }

    @Test
    public void testParallelStitcherWithAlwaysTimeoutError() throws InterruptedException {
        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder();
        confBuilder.setAm2DependsOnAm1(true);
        confBuilder.setAmVersion(1, StitchingTestSetupConfigBuilder.AmVersion.V3);
        confBuilder.setAmVersion(2, StitchingTestSetupConfigBuilder.AmVersion.V3);
        confBuilder.setSupportsAny(1, true);
        confBuilder.setSupportsAny(2, true);
        confBuilder.addSpecialReply(2, StitchingTestSetupConfig.SpecialReply.ERROR_ALWAYS_TIMEOUT);
        StitchingTestSetupConfig conf = confBuilder.build();

        StitchingTestSetup setup = new StitchingTestSetup(conf);
        setup.testStitching();
        setup.outputToFile();
        setup.assertFailure();

        //there should be at most 3 allocate calls. (1 for AM1 and 2 failed for AM2)
        List<ApiCallDetailsRef> allocateStateCallsRef = setup.getStitchingReport().getAllStateCalls("ALLOCATE");

        List<SerializableApiCallDetails> allocateStateCalls = allocateStateCallsRef.stream().map(setup.getApiCallDetailsCache()::getApiCallDetails).collect(Collectors.toList());
        List<SerializableApiCallDetails> deleteCalls = allocateStateCalls.stream().
                filter(c -> c.getGeniMethodName().startsWith("Delete")).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().equals("Allocate") || c.getGeniMethodName().equals("CreateSliver")).
                filter(c -> c.getGeniResponseCode() == null || !c.getGeniResponseCode().equals(GeniAMResponseCode.INTERNAL_NONGENI_ERROR.getCode())).
                collect(Collectors.toList());

        assertThat(allocateCalls, hasSize(lessThanOrEqualTo(3)));
    }

    @Test(dataProvider = "allNoVlanStitchingSetupConfigs")
    public void testParallelStitcherWithNoVlans(StitchingTestSetupConfig stitchingTestSetupConfig) throws InterruptedException {
        StitchingTestSetup setup = new StitchingTestSetup(stitchingTestSetupConfig);
        setup.testStitching();
        setup.outputToFile();
        setup.assertFailure();

        Boolean noVlanAmSupportsAny = null;
        for (int amIndex : new int[]{1,2} ) {
            if (stitchingTestSetupConfig.am(amIndex).hasSpecialReply(StitchingTestSetupConfig.SpecialReply.ERROR_NO_VLANS))
                noVlanAmSupportsAny = stitchingTestSetupConfig.am(amIndex).supportsAny;
        }
        assert noVlanAmSupportsAny != null;

        List<ApiCallDetailsRef> allocateStateCallsRef = setup.getStitchingReport().getAllStateCalls("ALLOCATE");

        List<SerializableApiCallDetails> allocateStateCalls = allocateStateCallsRef.stream().map(setup.getApiCallDetailsCache()::getApiCallDetails).collect(Collectors.toList());
        List<SerializableApiCallDetails> deleteCalls = allocateStateCalls.stream().
                filter(c -> c.getGeniMethodName().startsWith("Delete")).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().equals("Allocate") || c.getGeniMethodName().equals("CreateSliver")).collect(Collectors.toList());

        if (noVlanAmSupportsAny) {
            //Since this is a permanent, unrecoverable failure, the ALLOCATE phase should give up immediately.
            //If there are more then 4 calls, that is a sure sign that the ALLOCATE call retried up to 3 times.

            //but this assumption is only valid for am1, as am2 will get direct VLAN offers due to its dep on am1

            if(stitchingTestSetupConfig.am(1).hasSpecialReply(StitchingTestSetupConfig.SpecialReply.ERROR_NO_VLANS))
                assertThat(allocateCalls, hasSize(both(greaterThan(0)).and(lessThanOrEqualTo(2))));
        } else {
            //since any is not suppoorted, each tried vlan will be rejected. So we should see many calls.
            assertThat(allocateCalls, hasSize(greaterThan(4)));

        }

        System.out.flush();
        System.err.flush();
        Thread.yield();
        System.out.flush();
        System.err.flush();
        Thread.yield();
        Thread.yield();
    }

    @Test(dataProvider = "all32anyCombinations")
    public void testParallelStitcherWithMNoMatchVlans(StitchingTestSetupConfig stitchingTestSetupConfigBase) throws InterruptedException {
        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder(stitchingTestSetupConfigBase);
        confBuilder.setSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1);
        confBuilder.setSpecialCase(StitchingTestSetupConfig.SpecialCase.am1NoMatchingVlansAm2);
        StitchingTestSetupConfig stitchingTestSetupConfig = confBuilder.build();
        assert stitchingTestSetupConfig.hasSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1);
        assert stitchingTestSetupConfig.hasSpecialCase(StitchingTestSetupConfig.SpecialCase.am1NoMatchingVlansAm2);

        StitchingTestSetup setup = new StitchingTestSetup(stitchingTestSetupConfig);
        setup.testStitching();
        setup.outputToFile();
        setup.assertFailure();

        List<ApiCallDetailsRef> allocateStateCallsRef = setup.getStitchingReport().getAllStateCalls("ALLOCATE");

        List<SerializableApiCallDetails> allocateStateCalls = allocateStateCallsRef.stream().map(setup.getApiCallDetailsCache()::getApiCallDetails).collect(Collectors.toList());
        List<SerializableApiCallDetails> deleteCalls = allocateStateCalls.stream().
                filter(c -> c.getGeniMethodName().startsWith("Delete")).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().equals("Allocate") || c.getGeniMethodName().equals("CreateSliver")).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateStateStatusCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().contains("Status")).collect(Collectors.toList());

        int am1vlanCount = new VlanRange(setup.realAvailableAm1).size();
        //the stitcher aslo has a maximum number of VLANs it will try before giving up
        if (am1vlanCount > StitchingDirector.MAX_UNAVAILABLE_VLANS_BEFORE_GIVEUP)
            am1vlanCount = StitchingDirector.MAX_UNAVAILABLE_VLANS_BEFORE_GIVEUP;

        boolean hasFailedAnyCallForAm1 = !stitchingTestSetupConfig.am(1).supportsAny && setup.testAm1Auth.hasSpecialCase(SfaAuthority.SpecialCase.STITCHING_SUPPORTS_ANY);

        assertThat("There should be at least 2 calls for every vlans on AM 1 + 1 if any is not supported",
                allocateCalls, hasSize(greaterThanOrEqualTo(am1vlanCount*2 + (hasFailedAnyCallForAm1?1:0))));

        System.out.flush();
        System.err.flush();
        Thread.yield();
        System.out.flush();
        System.err.flush();
        Thread.yield();
        Thread.yield();
    }

    //TODO: reenable this test if the race condition is fixed
    @Test(dataProvider = "all32Combinations", enabled = false)
    public void testParallelStitcherWithSingleStatusFailedUnknownAm1(StitchingTestSetupConfig stitchingTestSetupConfigBase) throws InterruptedException {
        //TODO: this test is bad: it doesn't fail for some race conditions: if the sliver that fails was being deleted when it failed (because of AM dep).
        //TODO: make AM 2 fail to prevent this scenario
        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder(stitchingTestSetupConfigBase);
        confBuilder.setSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1);
        confBuilder.addSpecialReply(1, StitchingTestSetupConfig.SpecialReply.FAIL_ONCE_SECOND_STATUS_GENERICCAUSE);
        StitchingTestSetupConfig stitchingTestSetupConfig = confBuilder.build();
        assert stitchingTestSetupConfig.hasSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1);
        assert stitchingTestSetupConfig.hasSpecialReply(1, StitchingTestSetupConfig.SpecialReply.FAIL_ONCE_SECOND_STATUS_GENERICCAUSE);

        StitchingTestSetup setup = new StitchingTestSetup(stitchingTestSetupConfig);
        setup.testStitching();
        setup.outputToFile();
        setup.assertFailure();

        List<ApiCallDetailsRef> allocateStateCallsRef = setup.getStitchingReport().getAllStateCalls("ALLOCATE");

        List<SerializableApiCallDetails> allocateStateCalls = allocateStateCallsRef.stream().map(setup.getApiCallDetailsCache()::getApiCallDetails).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().equals("Allocate") || c.getGeniMethodName().equals("CreateSliver")).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateStateStatusCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().contains("Status")).collect(Collectors.toList());
        long allocateCallsAm1Count = allocateCalls.stream().filter(c -> c.getAuthorityUrn().equals(TestCertHelper.am1Urn)).count();
        long allocateCallsAm2Count = allocateCalls.stream().filter(c -> c.getAuthorityUrn().equals(TestCertHelper.am2Urn)).count();

        //TODO count all delete calls
//        assertThat(deleteCallsAm1Count, is(1l));
//        assertThat(deleteCallsAm2Count, is(1l));
        //cannot be assumed in all cases:
//        assertThat(allocateCallsAm1Count, is(stitchingTestSetupConfig.am(1).supportsAny ? 1l : 2l));
//        assertThat(allocateCallsAm2Count, is(stitchingTestSetupConfig.am(2).supportsAny ? 1l : 2l));

        System.out.flush();
        System.err.flush();
        Thread.yield();
        System.out.flush();
        System.err.flush();
        Thread.yield();
        Thread.yield();
    }

    @Test(dataProvider = "all32Combinations")
    public void testParallelStitcherWithSingleStatusFailedVlanUnavailableAm1(StitchingTestSetupConfig stitchingTestSetupConfigBase) throws InterruptedException {
        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder(stitchingTestSetupConfigBase);
        confBuilder.setSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1);
        confBuilder.addSpecialReply(1, StitchingTestSetupConfig.SpecialReply.FAIL_ONCE_SECOND_STATUS_VLAN_UNAVAILABLE);
        StitchingTestSetupConfig stitchingTestSetupConfig = confBuilder.build();
        assert stitchingTestSetupConfig.hasSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1);
        assert stitchingTestSetupConfig.hasSpecialReply(1, StitchingTestSetupConfig.SpecialReply.FAIL_ONCE_SECOND_STATUS_VLAN_UNAVAILABLE);

        StitchingTestSetup setup = new StitchingTestSetup(stitchingTestSetupConfig);
        setup.testStitching();
        setup.outputToFile();
        setup.assertSuccess(true/*allowDelete*/);

        List<ApiCallDetailsRef> allocateStateCallsRef = setup.getStitchingReport().getAllStateCalls("ALLOCATE");

        List<SerializableApiCallDetails> allocateStateCalls = allocateStateCallsRef.stream().map(setup.getApiCallDetailsCache()::getApiCallDetails).collect(Collectors.toList());
        List<SerializableApiCallDetails> deleteCalls = allocateStateCalls.stream().
                filter(c -> c.getGeniMethodName().startsWith("Delete")).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().equals("Allocate") || c.getGeniMethodName().equals("CreateSliver")).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateStateStatusCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().contains("Status")).collect(Collectors.toList());
        long deleteCallsAm1Count = deleteCalls.stream().filter(c -> c.getAuthorityUrn().equals(TestCertHelper.am1Urn)).count();
        long deleteCallsAm2Count = deleteCalls.stream().filter(c -> c.getAuthorityUrn().equals(TestCertHelper.am2Urn)).count();
        long allocateCallsAm1Count = allocateCalls.stream().filter(c -> c.getAuthorityUrn().equals(TestCertHelper.am1Urn)).count();
        long allocateCallsAm2Count = allocateCalls.stream().filter(c -> c.getAuthorityUrn().equals(TestCertHelper.am2Urn)).count();

        //the order of trying vlans is not guaranteed, so we can't test this.
//        long vlanRetries = 3; //300,301 and 302 tested before 305 which succeeds
//        assertThat(deleteCallsAm1Count, is(greaterThanOrEqualTo(vlanRetries))); //might be additional one due to status failure
        assertThat(deleteCallsAm2Count, isOneOf(0l, 1l)); //might be deleted to to Status failure, but due to timing might also not
//        assertThat(allocateCallsAm1Count, isOneOf(vlanRetries + 1l, vlanRetries + 2l)); //might be additional one due to status failure
//        assertThat(allocateCallsAm2Count, isOneOf(vlanRetries + 1l, vlanRetries + 2l)); //might be additional one due to status failure

        System.out.flush();
        System.err.flush();
        Thread.yield();
        System.out.flush();
        System.err.flush();
        Thread.yield();
        Thread.yield();
    }

    @Test(dataProvider = "all32Combinations")
    public void testParallelStitcherWithSingleStatusFailedBandwidthAm2(StitchingTestSetupConfig stitchingTestSetupConfigBase) throws InterruptedException {
        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder(stitchingTestSetupConfigBase);
        confBuilder.setSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1);
        confBuilder.addSpecialReply(2, StitchingTestSetupConfig.SpecialReply.FAIL_ONCE_SECOND_STATUS_BANDWIDTH);
        StitchingTestSetupConfig stitchingTestSetupConfig = confBuilder.build();
        assert stitchingTestSetupConfig.hasSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1);
        assert stitchingTestSetupConfig.hasSpecialReply(2, StitchingTestSetupConfig.SpecialReply.FAIL_ONCE_SECOND_STATUS_BANDWIDTH);

        StitchingTestSetup setup = new StitchingTestSetup(stitchingTestSetupConfig);
        setup.testStitching();
        setup.outputToFile();
        setup.assertFailure();

        List<ApiCallDetailsRef> allocateStateCallsRef = setup.getStitchingReport().getAllStateCalls("ALLOCATE");

        List<SerializableApiCallDetails> allocateStateCalls = allocateStateCallsRef.stream().map(setup.getApiCallDetailsCache()::getApiCallDetails).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().equals("Allocate") || c.getGeniMethodName().equals("CreateSliver")).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateStateStatusCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().contains("Status")).collect(Collectors.toList());
        long allocateCallsAm1Count = allocateCalls.stream().filter(c -> c.getAuthorityUrn().equals(TestCertHelper.am1Urn)).count();
        long allocateCallsAm2Count = allocateCalls.stream().filter(c -> c.getAuthorityUrn().equals(TestCertHelper.am2Urn)).count();

        //TODO count all delete calls
//        assertThat(deleteCallsAm1Count, is(1l));
//        assertThat(deleteCallsAm2Count, is(1l));
        //cannot be assumed in all cases:
//        assertThat(allocateCallsAm1Count, is(stitchingTestSetupConfig.am(1).supportsAny ? 1l : 2l));
//        assertThat(allocateCallsAm2Count, is(stitchingTestSetupConfig.am(2).supportsAny ? 1l : 2l));

        System.out.flush();
        System.err.flush();
        Thread.yield();
        System.out.flush();
        System.err.flush();
        Thread.yield();
        Thread.yield();
    }
    @Test(dataProvider = "all32Combinations")
    public void testParallelStitcherWithAlwaysStatusFailedAm1(StitchingTestSetupConfig stitchingTestSetupConfigBase) throws InterruptedException {
        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder(stitchingTestSetupConfigBase);
        confBuilder.setSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1);
        confBuilder.addSpecialReply(1, StitchingTestSetupConfig.SpecialReply.FAIL_ALLWAYS_SECOND_STATUS_VLAN_UNAVAILABLE);
        StitchingTestSetupConfig stitchingTestSetupConfig = confBuilder.build();
        assert stitchingTestSetupConfig.hasSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1);
        assert stitchingTestSetupConfig.hasSpecialReply(1, StitchingTestSetupConfig.SpecialReply.FAIL_ALLWAYS_SECOND_STATUS_VLAN_UNAVAILABLE);

        StitchingTestSetup setup = new StitchingTestSetup(stitchingTestSetupConfig);
        setup.testStitching();
        setup.outputToFile();
        setup.assertFailure();

        System.out.flush();
        System.err.flush();
        Thread.yield();
        System.out.flush();
        System.err.flush();
        Thread.yield();
        Thread.yield();
    }

    @Test(dataProvider = "all32Combinations")
    public void testParallelStitcherWithAlwaysStatusFailedAm2(StitchingTestSetupConfig stitchingTestSetupConfigBase) throws InterruptedException {
        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder(stitchingTestSetupConfigBase);
        confBuilder.setSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1);
        confBuilder.addSpecialReply(2, StitchingTestSetupConfig.SpecialReply.FAIL_ALLWAYS_SECOND_STATUS_VLAN_UNAVAILABLE);
        StitchingTestSetupConfig stitchingTestSetupConfig = confBuilder.build();
        assert stitchingTestSetupConfig.hasSpecialCase(StitchingTestSetupConfig.SpecialCase.am2DependsOnAm1);
        assert stitchingTestSetupConfig.hasSpecialReply(2, StitchingTestSetupConfig.SpecialReply.FAIL_ALLWAYS_SECOND_STATUS_VLAN_UNAVAILABLE);

        StitchingTestSetup setup = new StitchingTestSetup(stitchingTestSetupConfig);
        setup.testStitching();
        setup.outputToFile();
        System.err.flush();
        System.out.println("Tested setup is: " + stitchingTestSetupConfigBase);
        System.out.println("Final phase: " + setup.getParallelStitcher().getGlobalPhase());
        System.out.println("Gave up: " + setup.isGaveUpOnStitching());
        System.out.flush();
        System.err.flush();
        setup.assertFailure();

        List<ApiCallDetailsRef> allocateStateCallsRef = setup.getStitchingReport().getAllStateCalls("ALLOCATE");

        List<SerializableApiCallDetails> allocateStateCalls = allocateStateCallsRef.stream().map(setup.getApiCallDetailsCache()::getApiCallDetails).collect(Collectors.toList());
        List<SerializableApiCallDetails> deleteCalls = allocateStateCalls.stream().
                filter(c -> c.getGeniMethodName().startsWith("Delete")).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().equals("Allocate") || c.getGeniMethodName().equals("CreateSliver")).collect(Collectors.toList());
        List<SerializableApiCallDetails> allocateStateStatusCalls = allocateStateCalls.stream().filter(
                c -> c.getGeniMethodName().contains("Status")).collect(Collectors.toList());

        System.out.flush();
        System.err.flush();
        Thread.yield();
        System.out.flush();
        System.err.flush();
        Thread.yield();
        Thread.yield();
    }

    //repeats for easy re-testing
//    @Test
//    public void testParallelStitcherAm3Am3AnyAnyAm2DependsOnAm1() throws InterruptedException {
//        StitchingTestSetupConfigBuilder confBuilder = new StitchingTestSetupConfigBuilder();
//        confBuilder.setAm2DependsOnAm1(true);
//        confBuilder.setAmVersion(1, false);
//        confBuilder.setAmVersion(2, false);
//        confBuilder.setSupportsAny(1, true);
//        confBuilder.setSupportsAny(2, true);
//        StitchingTestSetupConfig conf = confBuilder.build();
//
//        StitchingTestSetup setup = new StitchingTestSetup(conf);
//        setup.testStitching();
//        setup.outputToFile();
//        setup.assertSuccess(true/*allowDelete*/);
//    }
//    @Test
//    public void testParallelStitcherAm3Am3NoAnyNoAny() throws InterruptedException {
//        StitchingTestSetup setup = new StitchingTestSetup(
//                false/*am1HasV2*/, true/*am1HasV3*/,
//                false/*am2HasV2*/, true/*am2HasV3*/,
//                false/*am1SupportsAny*/, false/*am2SupportsAny*/);
//        setup.testStitching();
//        setup.outputToFile();
//        setup.assertSuccess(false/*allowDelete*/);
//    }
//    @Test
//    public void testParallelStitcherAm3Am3AnyNoAny() throws InterruptedException {
//        StitchingTestSetup setup = new StitchingTestSetup(
//                false/*am1HasV2*/, true/*am1HasV3*/,
//                false/*am2HasV2*/, true/*am2HasV3*/,
//                true/*am1SupportsAny*/, false/*am2SupportsAny*/);
//        setup.testStitching();
//        setup.outputToFile();
//        setup.assertSuccess(false/*allowDelete*/);
//    }
}
