package be.iminds.ilabt.jfed.lowlevel;

import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Date;
import java.util.Hashtable;
import java.util.List;

/**
 * <p>AnyCredential is a wrapper for any XML Credential.
 * <p/>
 * <p>It does not parse anything as the format is unknown <p>A name is stored alongside the actual credential, which is useful for listing the credential in GUI lists and comboboxes.
 */
public class AnyCredential {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(AnyCredential.class);

    protected String name;
    protected String credentialXml;

    //see Geni Aggregate Manager API v3: http://groups.geni.net/geni/wiki/GAPI_AM_API_V3/CommonConcepts#credentials
    protected String type;
    protected String version;

    protected AnyCredential(String name, String credentialXml, String type, String version) {
        this.name = name;
        this.credentialXml = credentialXml;
        this.type = type;
        this.version = version;

        if (credentialXml == null) throw new RuntimeException("AnyCredential credentialXml may not be null");
    }

    /**
     * Pre Geni API v3 credentials: type=geni_sfa version=2
     */
    public static AnyCredential create(String name, String credentialXml) throws CredentialException {
        return new SfaCredential(name, credentialXml, "geni_sfa", "2");
    }

    public static AnyCredential create(String name, String credentialXml, String type, String version) throws CredentialException {
        if (type.equalsIgnoreCase("sfa") || type.equalsIgnoreCase("geni_sfa"))
            return new SfaCredential(name, credentialXml, type, version);

        if (type.equalsIgnoreCase("abac") || type.equalsIgnoreCase("geni_abac"))
            return new AbacCredential(name, credentialXml, type, version);

        return new AnyCredential(name, credentialXml, type, version);
    }

    public static AnyCredential createFromV3Hashtable(String credName, Hashtable ht) throws BadReplyGeniException, CredentialException {
        Object geniTypeO = ht.get("geni_type");
        Object geniVersionO = ht.get("geni_version");
        Object genivalueO = ht.get("geni_value");

        if (geniTypeO == null || geniVersionO == null || genivalueO == null)
            throw new BadReplyGeniException("The API specified a Vector of Hashtables sepcifying credentials. These must contain the keys geni_type, geni_version and geni_value." +
                    " At least one of these is missing. Hashtable content: " + ht);

        if (!(geniTypeO instanceof String))
            throw new BadReplyGeniException("The API specified a credential, but the geni_type field of the Hashtable is of type " + geniTypeO.getClass().getName() + " instead of String. value=" + geniTypeO);

        if (!(geniVersionO instanceof Integer) && !(geniVersionO instanceof String))
            throw new BadReplyGeniException("The API specified a credential, but the geni_version field of the Hashtable is of type " + geniVersionO.getClass().getName() + " instead of String with Integer. value=" + geniVersionO);

        if (!(genivalueO instanceof String))
            throw new BadReplyGeniException("The API specified a credential, but the geni_value field of the Hashtable is of type " + genivalueO.getClass().getName() + " instead of String. value=" + genivalueO);

        return create(credName, (String) genivalueO, (String) geniTypeO, geniVersionO.toString());
    }

    public static AnyCredential toSingleCredential(List<AnyCredential> credentialList) {
        if (credentialList == null) {
            LOG.error("requested toSingleCredential, but credentialList is null");
            return null;
        }
        if (credentialList.size() <= 0) {
            LOG.error("requested toSingleCredential, but " + credentialList.size() + " credentials in list");
            return null;
        }
        if (credentialList.size() > 1)
            LOG.warn("requested toSingleCredential, but " + credentialList.size() + " credentials in list");
        return credentialList.get(0);
    }

    /**
     * work for null credential too
     */
    public static List<AnyCredential> toCredentialList(AnyCredential c) {
        List<AnyCredential> res = new ArrayList<AnyCredential>();
        res.add(c);
        return res;
    }

    public static Date getExpireTime(List<AnyCredential> credentials) {
        for (AnyCredential cred : credentials)
            if (cred instanceof SfaCredential) {
                SfaCredential sfaCred = (SfaCredential) cred;
                if (sfaCred.getExpiresDate() != null)
                    return sfaCred.getExpiresDate();
            }

        return null;

        //TODO handle case of multiple different dates (perfer one, and warn?)
    }

    public String getCredentialXml() {
        assert credentialXml != null;
        return credentialXml;
    }

    public String getType() {
        return type;
    }

    public String getVersion() {
        return version;
    }

    public Hashtable getGeniV3Hashtable() {
        Hashtable<String, Object> res = new Hashtable<String, Object>();
        res.put("geni_type", type);
        res.put("geni_version", version);
        assert credentialXml != null;
        res.put("geni_value", credentialXml);
        return res;
    }

    public String getName() {
        return name;
    }

    public void setName(String newName) {
        this.name = newName;
    }

    public String getExpires() {
        return null;
    }

    public Date getExpiresDate() {
        return null;
    }

    /**
     * @return null if unknown
     */
    public Boolean isTargetSubAuthority() {
        return null;
    }

    @Override
    public String toString() {
        return credentialXml;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) return true;
        if (o == null || getClass() != o.getClass()) return false;

        AnyCredential that = (AnyCredential) o;

        if (!credentialXml.equals(that.credentialXml)) return false;
        if (!name.equals(that.name)) return false;

        return true;
    }

    @Override
    public int hashCode() {
        int result = name.hashCode();
        result = 31 * result + credentialXml.hashCode();
        return result;
    }

    public List<AnyCredential> toCredentialList() {
        List<AnyCredential> res = new ArrayList<AnyCredential>();
        res.add(this);
        return res;
    }
}
