package be.iminds.ilabt.jfed.lowlevel.api_wrapper.impl;

import be.iminds.ilabt.jfed.log.Logger;
import be.iminds.ilabt.jfed.lowlevel.AnyCredential;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.ServerType;
import be.iminds.ilabt.jfed.lowlevel.api.*;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedConnectionProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.SfaConnection;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.RFC3339Util;
import be.iminds.ilabt.jfed.util.TextUtil;
import org.slf4j.LoggerFactory;

import java.util.*;

/**
 * ProtoGeniSAUserAndSliceApiWrapper
 */
public class UniformFederationApi1UserAndSliceApiWrapper extends UserAndSliceApiWrapper {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(UniformFederationApi1UserAndSliceApiWrapper.class);
    List<AnyCredential> userCredentials;

    public UniformFederationApi1UserAndSliceApiWrapper(Logger logger, GeniUserProvider geniUserProvider, JFedConnectionProvider connectionProvider, JFedPreferences jFedPreferences) {
        super(logger, geniUserProvider, connectionProvider, jFedPreferences);
    }

    private FederationMemberAuthorityApi1 ma(Logger logger) {
        return new FederationMemberAuthorityApi1(logger, jFedPreferences);
    }

    private FederationSliceAuthorityApi1 sa(Logger logger) {
        return new FederationSliceAuthorityApi1(logger, jFedPreferences);
    }

    private SfaConnection getSaConnection() throws JFedException {
        return getConnection(new ServerType(ServerType.GeniServerRole.GENI_CH_SA, 1));
    }

    private SfaConnection getMaConnection() throws JFedException {
        return getConnection(new ServerType(ServerType.GeniServerRole.GENI_CH_MA, 1));
    }

    public FederationSliceAuthorityApi1.GetVersionSAResult getGetVersionSAResult(Logger logger) throws JFedException {
        //TODO: buffer

        AbstractFederationApi1.FederationApiReply<FederationSliceAuthorityApi1.GetVersionSAResult> reply =
                sa(logger).getVersion(getSaConnection());

        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("GetVersion call to SA not successful: code=" +
                    reply.getGeniResponseCode() + " output=" + reply.getOutput(),
                    reply
            );

        return reply.getValue();
    }

    @Override
    public List<AnyCredential> getUserCredentials(Logger logger, GeniUrn user) throws JFedException {
        userCredentials = null;

        AbstractFederationApi1.FederationApiReply<List<AnyCredential>> reply =
                ma(logger).getCredentials(
                        getMaConnection(),
                        new ArrayList<AnyCredential>(),
                        user,
                        null);

        assert reply != null;

        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Could not retrieve user credential from server: code=" +
                    reply.getGeniResponseCode() + " output=" + reply.getOutput(),
                    reply
            );

        assert reply.getValue() != null;
        assert !reply.getValue().isEmpty();

        userCredentials = reply.getValue();
        return userCredentials;
    }

    @Override
    public boolean hasUserCredentials() {
        return hasUserCredentials(userCredentials);
    }

    public List<AnyCredential> getCachedUserCredentialsForAM() {
        return userCredentials;
    }

    @Override
    public List<AnyCredential> getSliceCredentials(Logger logger, GeniUrn sliceUrn) throws JFedException {
        AbstractFederationApi1.FederationApiReply<List<AnyCredential>> reply =
                sa(logger).getSliceCredentials(getSaConnection(), userCredentials, sliceUrn, null);

        if (!reply.getGeniResponseCode().isSuccess())
            throw new JFedException("Could not retrieve slice credential from server: code=" +
                    reply.getGeniResponseCode() + " output=" + reply.getOutput(),
                    reply
            );

        //TODO support multiple credentials
        return reply.getValue();
    }

    @Override
    public List<GeniUrn> getSlicesForUser(Logger logger, GeniUrn user) throws JFedException {
        List<GeniUrn> res = new ArrayList<GeniUrn>();
        AbstractFederationApi1.FederationApiReply<List<FederationSliceAuthorityApi1.UrnRoleTuple>> reply =
                sa(logger).lookupSlicesForMember(getSaConnection(), user, userCredentials, null);
        if (reply == null || reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
            return res;
        List<FederationSliceAuthorityApi1.UrnRoleTuple> urnRoleTuples = reply.getValue();
        for (FederationSliceAuthorityApi1.UrnRoleTuple urnRoleTuple : urnRoleTuples)
            res.add(urnRoleTuple.getUrn());

        return res;
        //get other info:
//        Map<String, String> match = new HashMap<String, String>();
//        match.put("MEMBER_URN", user.toString());
//        ma.lookupPublicMemberInfo(con, match, null, null);
//        ma.lookupPrivateMemberInfo(con, userCredentials, match, null, null);
    }

    public List<GeniUrn> getUserProjects(Logger logger, GeniUrn user) throws JFedException {
        List<GeniUrn> res = new ArrayList<GeniUrn>();

        AbstractFederationApi1.FederationApiReply<List<FederationSliceAuthorityApi1.UrnRoleTuple>> reply =
                sa(logger).lookupProjectsForMember(getSaConnection(),
                        user,
                        userCredentials,
                        null);
        if (reply == null || reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
            return res;
        List<FederationSliceAuthorityApi1.UrnRoleTuple> urnRoleTuples = reply.getValue();
        LOG.debug("UniformFederationApi1UserAndSliceApiWrapper#getUserProjects urnRoleTuples.size()=" + urnRoleTuples.size());
        boolean hasExpired = false;
        for (FederationSliceAuthorityApi1.UrnRoleTuple urnRoleTuple : urnRoleTuples) {

            //first a special exception: server might include EXPIRED field in tuple. We can use that info
            if (urnRoleTuple.dictRaw != null && urnRoleTuple.dictRaw.containsKey("EXPIRED")) {
                hasExpired = true;
                if (TextUtil.objectToBoolean(urnRoleTuple.dictRaw.get("EXPIRED"))) {
                    LOG.debug("skipping expired project: " + urnRoleTuple.getUrn());
                    continue;
                }
            }

            res.add(urnRoleTuple.getUrn());
        }

        if (!hasExpired) {
            LOG.debug("doing second call to check for expired projects");

            Map<String, Object> match = new HashMap<String, Object>();
            match.put("EXPIRED", Boolean.FALSE);
            List<String> filter = new ArrayList<String>();
            filter.add("PROJECT_URN");
            AbstractFederationApi1.FederationApiReply<AbstractFederationApi1.LookupResult> nonExpiredProj = null;
            try {
                nonExpiredProj =
                        sa(logger).lookupProjects(getSaConnection(),
                                userCredentials,
                                match,
                                filter,
                                null);
            } catch (Throwable e) {
                LOG.warn("lookup_projects call failed. Could not check for expired projects. This error will be ignored (possibly leaving expired projects in the list).", e);
                nonExpiredProj = null;
                return res;
            }
            if (nonExpiredProj == null || nonExpiredProj.getGeniResponseCode() == null || !nonExpiredProj.getGeniResponseCode().isSuccess() || nonExpiredProj.getValue() == null) {
                LOG.warn("lookup_projects call failed. Could not check for expired projects. This error will be ignored (possibly leaving expired projects in the list).");
            } else {
                LOG.debug("Found " + nonExpiredProj.getValue().size() + " non expired projects on server.");
                List<GeniUrn> expiredUserProjects = new ArrayList<GeniUrn>();
                for (GeniUrn resUrn : res) {
                    if (nonExpiredProj.getValue().get(resUrn) == null)
                        expiredUserProjects.add(resUrn);
                }
                LOG.debug("found " + expiredUserProjects.size() + " expired projects in " + res.size() + " user projects: will remove them");
                res.removeAll(expiredUserProjects);
            }
        } else {
            LOG.debug("no second call needed to check for expired projects");
        }

        LOG.debug("UniformFederationApi1UserAndSliceApiWrapper#getUserProjects res.size()=" + res.size());
        return res;
    }

    public AbstractFederationApi1.LookupResult getUserIdentifyingInfo(Logger logger, GeniUrn user) throws JFedException {
        Hashtable<String, String> match = new Hashtable<String, String>();
        match.put("MEMBER_URN", getUser().getUserUrnString());

//        Vector<String> filter = new Vector<String>();

        AbstractFederationApi1.FederationApiReply<AbstractFederationApi1.LookupResult> reply =
                ma(logger).lookupIdentifyingMemberInfo(getMaConnection(),
                        userCredentials,
                        match,
                        null/*filter*/,
                        null);


        if (reply == null)
            throw new JFedException("Problem looking up user info for " + getUser().getUserUrnString());

        if (reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
            throw new JFedException("Problem looking up user info for " + getUser().getUserUrnString(), reply);

        return reply.getValue();
    }

    @Override
    public List<String> getAggregatesForSlice(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn) throws JFedException {
        AbstractFederationApi.FederationApiReply<AbstractFederationApi.LookupResult> reply =
                sa(logger).lookupSliverInfo(getSaConnection(),
                        sliceCredentials, sliceUrn, null, null, null);

        if (!reply.getGeniResponseCode().isSuccess()) return null;
        AbstractFederationApi.LookupResult lookupRes = reply.getValue();
        if (lookupRes == null) return null;

        List<String> res = new ArrayList<String>();
        List<String> keys = lookupRes.getKeys();
        for (String key : keys) {
            String aggManUrn = (String) lookupRes.get(key).get("SLIVER_INFO_AGGREGATE_URN");
            if (aggManUrn != null)
                res.add(aggManUrn);
        }

        return res;
    }

    @Override
    public void shareSlice(Logger logger, GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, GeniUrn userUrn) throws JFedException {
        throw new RuntimeException("Not implemented");
    }

    @Override
    public List<GeniUrn> getUsersForSubAuthority(Logger logger, String projectName) throws JFedException {
        GeniUrn projectUrn = GeniUrn.createGeniUrnFromEncodedParts(getUserAuthority().getNameForUrn(), "project", projectName);
        AbstractFederationApi.FederationApiReply<List<FederationSliceAuthorityApi1.UrnRoleTuple>> reply =
                sa(logger).lookupProjectMembers(getSaConnection(), projectUrn, userCredentials, null/*extraOptions*/);

        if (reply == null || reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
            if (reply == null)
                throw new JFedException("Error getting users in project "+projectUrn);
            else {
                throw new JFedException("Error getting users in project "+projectUrn, reply);
            }

        List<GeniUrn> res = new ArrayList<GeniUrn>();
        for (FederationSliceAuthorityApi1.UrnRoleTuple tuple : reply.getValue()) {
            res.add(tuple.getUrn());
        }

        return res;
    }

    @Override
    public List<GeniUrn> getUsersForSlice(Logger logger, GeniUrn sliceUrn) throws JFedException {
        AbstractFederationApi.FederationApiReply<List<FederationSliceAuthorityApi1.UrnRoleTuple>> reply =
                sa(logger).lookupSliceMembers(getSaConnection(), sliceUrn, userCredentials, null/*extraOptions*/);

        if (reply == null || reply.getGeniResponseCode() == null || !reply.getGeniResponseCode().isSuccess() || reply.getValue() == null)
            if (reply == null)
                throw new JFedException("Error getting users in "+sliceUrn);
            else {
                throw new JFedException("Error getting users in "+sliceUrn, reply);
            }

        List<GeniUrn> res = new ArrayList<GeniUrn>();
        for (FederationSliceAuthorityApi1.UrnRoleTuple tuple : reply.getValue()) {
            res.add(tuple.getUrn());
        }

        return res;
    }

    @Override
    public void addUsersToSlice(Logger logger, GeniUrn sliceUrn, List<AnyCredential> sliceCredentials, List<GeniUrn> userUrns, String role) throws JFedException {
        throw new RuntimeException("Not implemented");
    }

    @Override
    public SliceInfo createSlice(Logger logger, String sliceName, Date expirationDate, String subAuthName) throws JFedException {
        //2 steps: create slice, and then get its credential(s)

        SfaConnection con = getSaConnection();

        FederationSliceAuthorityApi1.GetVersionSAResult gv = getGetVersionSAResult(logger);
        assert gv != null;

        Map<String, String> fields = new HashMap<String, String>();
        fields.put("SLICE_DESCRIPTION", "jFed Experimenter GUI slice '" + sliceName + "' for user " + getUser().getUserUrnString());

        //old version: automatically choose
//        if (gv.getServices().contains("PROJECT")) {
//            List<GeniUrn> userProjects = getUserProjects(userCredentials, getUser().getUserUrn());
//            if (userProjects.isEmpty())
//                throw new JFedException("Slice creation requires a project, but user is not a member of any project.");
//
//            if (userProjects.size() <= 0) {
//                LOG.error("User " + getUser().getUserUrnString() + " is member of no projects! Will try to create slice without project, but that will most likely fail.");
//                return null;
//            } else {
//                GeniUrn chosenProject = null;
//                if (userProjects.size() > 1) {
//                    if (ExtraInfoCallback.getFederationApiProjectSelectionCallback() != null) {
//                        LOG.debug("User is member of "+userProjects.size()+" projects. Using callback to let user choose one.");
//                        chosenProject = ExtraInfoCallback.getFederationApiProjectSelection(userProjects);
//                    } else {
//                        LOG.warn("User "+getUser().getUserUrnString()+" is member of multiple projects: "+userProjects+". " +
//                                "No callback available to let user choose project: Choosing first project as project for this slice.");
//                        chosenProject = userProjects.get(0);
//                    }
//                } else
//                    chosenProject = userProjects.get(0);
//                fields.put("SLICE_PROJECT_URN", chosenProject.getValue());
//            }
//        }

        GeniUrn chosenProject = null;
        boolean projectSupported = true;//gv.getServices().contains("PROJECT");
        if (projectSupported) {
            if (subAuthName == null)
                throw new JFedException("Slice creation requires a project, but user did not specify any project.");

            chosenProject = GeniUrn.createGeniUrnFromEncodedParts(getUser().getUserUrn().getEncodedTopLevelAuthority(), "project", subAuthName);
        }
        if (!projectSupported && subAuthName != null) {
            LOG.warn("The server does not support PROJECTS, but the user specified a project name. Will ignore project");
        }
        if (chosenProject != null)
            fields.put("SLICE_PROJECT_URN", chosenProject.getValue());

        if (gv.getFieldsForObject("PROJECT").containsKey("_GENI_SLICE_EMAIL")) {
            AbstractFederationApi1.LookupResult userIdentifyingInfo =
                    getUserIdentifyingInfo(logger, getUser().getUserUrn());

            if (userIdentifyingInfo != null && userIdentifyingInfo.get(getUser().getUserUrn()).containsKey("MEMBER_EMAIL")) {
                String userEmail = (String) userIdentifyingInfo.get(getUser().getUserUrn()).get("MEMBER_EMAIL");
                fields.put("_GENI_SLICE_EMAIL", userEmail);
            } else {
                throw new JFedException("Slice creation requires a user email, but the user email is not known.");
            }
        }

//      TODO  addRequiredSupplementaryFieldsToCreate(fields, "SLICE");

        if (expirationDate != null) {
            //note: SLICE_EXPIRATION is allowed on Creation, so this will always work
            fields.put("SLICE_EXPIRATION", RFC3339Util.dateToRFC3339String(expirationDate, true, true, true));
        }

        AbstractFederationApi1.FederationApiReply<Hashtable<String, Object>> reply =
                sa(logger).createSlice(
                        con,
                        userCredentials,
                        sliceName/* automatically added to "SLICE_NAME" field*/,
                        fields,
                        null/*extraOptions*/);


        if (reply != null && reply.getGeniResponseCode() != null && reply.getGeniResponseCode().isSuccess() && reply.getValue() != null) {
            Hashtable<String, Object> sliceInfo = reply.getValue();

//            if (sliceInfo.containsKey("SLICE_CREDENTIAL"?)) {
            //        //TODO slice credential is normally in returned fields? Otherwise, we can retrieve it with a call.
//            }

            if (sliceInfo.containsKey("SLICE_URN")) {
                String sliceUrnStr = sliceInfo.get("SLICE_URN").toString();
                GeniUrn sliceUrn = GeniUrn.parse(sliceUrnStr);
                if (sliceUrn != null) {
                    AbstractFederationApi1.FederationApiReply<List<AnyCredential>> credReply =
                            sa(logger).getSliceCredentials(con, userCredentials, sliceUrn, null);

                    //TODO support multiple credentials
                    return new SliceInfo(sliceName, sliceUrn, credReply.getValue());
                }
            }
            throw new JFedException("Error creating slice", reply);
        } else
            throw new JFedException("Error creating slice: reply == null");
    }

    @Override
    public List<AnyCredential> renewSlice(Logger logger, List<AnyCredential> sliceCredentials, Date newExpirationDate) throws JFedException {
        throw new UnsupportedOperationException("Currently not supported");
    }

    @Override
    public List<String> getSshKeysForUser(Logger logger, GeniUrn userUrn) throws JFedException {
//        if (!getGetVersionSAResult(logger).getServices().contains("KEY")) {
//            LOG.warn("Server does not support fetching SSH keys: not doing call");
//            return null;
//        }

        Map<String, String> match = new HashMap<String, String>();
        match.put("KEY_MEMBER", userUrn.toString());
        AbstractFederationApi.FederationApiReply<Map<GeniUrn, List<Map<String, Object>>>> reply =
                ma(logger).lookupKeys(getMaConnection(), userCredentials, match, null/*filter*/, null);

        if (reply.getGeniResponseCode().isSuccess()) {
            List<String> res = new ArrayList<String>();
            List<Map<String, Object>> keylist = reply.getValue().get(userUrn);
            for (Map<String, Object> keymap : keylist) {
                String key = (String) keymap.get("KEY_PUBLIC");
                if (key != null) res.add(key);
            }
            return res;
        } else
            throw new JFedException("Failed to get SSH keys", reply);
    }


    /* sub authority support */
    @Override
    public SubAuthoritySupport getSubAuthoritySupport(Logger logger) throws JFedException {
        FederationSliceAuthorityApi1.GetVersionSAResult getSAVersionRes = getGetVersionSAResult(logger);
        if (!getSAVersionRes.getServices().contains("PROJECT"))
            return SubAuthoritySupport.SUB_AUTHORITY_FORBIDDEN;
        else
            return SubAuthoritySupport.SUB_AUTHORITY_MANDATORY;
    }

    @Override
    public List<String> getSubAuthorityNames(Logger logger, GeniUrn user) throws JFedException {
        List<String> res = new ArrayList<String>();

        List<GeniUrn> projectUrns = getUserProjects(logger, user);
        for (GeniUrn projectUrn : projectUrns) {
            res.add(projectUrn.getEncodedResourceName());
        }

        LOG.debug("returning subauthorities: " + res);

        return res;
    }
    @Override
    public boolean isRegisterAggregatesForSliceSupported() {
        return false;
    }
    @Override
    public void registerAggregatesForSlice(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn, GeniUrn aggregateUrn, Collection<GeniUrn> sliverUrns, Date sliverExpirationDate) throws JFedException {
        throw new RuntimeException("Not yet implemented");
    }
    @Override
    public void updateSliverExpirationDateInfoForSlice(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn, Collection<GeniUrn> sliverUrns, Date sliverExpirationDate) throws JFedException {
        throw new RuntimeException("Not yet implemented");
    }
    @Override
    public void unregisterAggregatesForSlice(Logger logger, List<AnyCredential> sliceCredentials, GeniUrn sliceUrn, GeniUrn aggregateUrn, Collection<GeniUrn> sliverUrns) throws JFedException {
        throw new RuntimeException("Not yet implemented");
    }
}
