package be.iminds.ilabt.jfed.lowlevel.authority;

import be.iminds.ilabt.jfed.preferences.CorePreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedPreferences;
import be.iminds.ilabt.jfed.util.JFedUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Named;
import javax.inject.Provider;
import java.io.File;

/**
 * jFedCombinedAuthorityList
 * <p/>
 * This combines authority info from various sources, to create the list that is used in all jFed tools.
 */
public class JFedAuthorityList implements Provider<AuthorityListModel> {
    private static final Logger LOG = LoggerFactory.getLogger(JFedAuthorityList.class);
    private static File defaultAuthoritiesFile = getDefaultAuthoritiesFile();

    /**
     * "AUTO" (= use testbeds.xml or fallback), "file:*" or "http*" or "BUILTIN"   see also PreferencesUtil
     */
    private static String userPreference = null;
    private File authoritiesFile;

    @Inject
    public JFedAuthorityList(@Named("authoritiesFile") @Nullable File authoritiesFile,
                             JFedPreferences jFedPreferences) {
        this.authoritiesFile = authoritiesFile;
        userPreference = jFedPreferences.getString(CorePreferenceKey.PREF_TESTBEDS_XML_SOURCE);
    }

    public static File getDefaultAuthoritiesFile() {
        String userDataDir = JFedUtils.getUserDataDirectory();
        if (userDataDir == null)
            return null;
        return new File(userDataDir, "authorities.xml");
    }


    @Override
    public AuthorityListModel get() {
        AuthorityListModel authorityListModel = new AuthorityListModel();

        String altUrl = null;
        boolean builtinOnly = false;
        if (userPreference != null) {
            String testbedsSourcePref = userPreference;
            if (testbedsSourcePref.equalsIgnoreCase("AUTO") || testbedsSourcePref.equalsIgnoreCase("DEFAULT")) {
                //nothing to do, we just go through the normal procedure below
            } else {
                LOG.debug("Skipping default authority list method, and using user specified method: \"" + userPreference + "\"");
                if (testbedsSourcePref.equalsIgnoreCase("BUILTIN")) {
                    builtinOnly = true;
                } else {
                    if (testbedsSourcePref.startsWith("file:")) {
                        String filename = testbedsSourcePref.substring(5);
                        File userFile = new File(filename);
                        try {
                            StoredAuthorityList.LoadResult loadResult = StoredAuthorityList.load(userFile, authorityListModel);
                            if (!loadResult.getExtraFiles().isEmpty()) {
                                LOG.warn("extra files are not supported in local file "+userFile);
                            }
                            if (!loadResult.isSuccess())
                                LOG.error("Something went wrong while trying to read AuthorityList at " + userFile);
                            else
                                return authorityListModel; //stop here, we are successfully done
                        } catch (Throwable e) {
                            LOG.error("Something went wrong while trying to read AuthorityList at " + userFile, e);
                        }
                        //fall through to normal procedure
                        //TODO callback to inform user it's preference could not be used
                        LOG.debug("Could not use user specified file as authorities source");
                    } else {
                        altUrl = testbedsSourcePref;
                    }
                }
            }
        }

        //first read from the Fed4FireAuthorityList
        if (!builtinOnly) {
            boolean fed4fireSuccess = false;
            try {
                if (altUrl != null)
                    fed4fireSuccess = Fed4FireAuthorityList.load(authorityListModel, altUrl);
                else
                    fed4fireSuccess = Fed4FireAuthorityList.load(authorityListModel);
            } catch (Throwable e) {
                LOG.error("Something went wrong while trying to fetch Fed4Fire AuthorityList: ", e);
                fed4fireSuccess = false;
            }
            try {
                if (!fed4fireSuccess) {
                    if (altUrl == null)
                        LOG.debug("Something went wrong trying to load the default testbeds.xml URL");
                    else
                        LOG.debug("Something went wrong trying to load testbeds.xml from " + altUrl);

                    LOG.debug("Using builtin authority list as backup");
                    BuiltinAuthorityList.load(authorityListModel);
                }
            } catch (Throwable e) {
                LOG.error("Something went wrong while using BuiltinAuthorityList: ", e);
            }
        } else {
            try {
                BuiltinAuthorityList.load(authorityListModel);
                return authorityListModel;
            } catch (Throwable e) {
                LOG.error("Something went wrong while using BuiltinAuthorityList: ", e);
            }
        }

        //then overwrite from the default StoredAuthorityList if it exists.

        if (!builtinOnly) {
            if (authoritiesFile != null && authoritiesFile.exists()) {
                try {
                    StoredAuthorityList.LoadResult loadResult = StoredAuthorityList.load(authoritiesFile, authorityListModel);
                    if (!loadResult.getExtraFiles().isEmpty()) {
                        LOG.warn("extra files are not supported in "+authoritiesFile);
                    }
                    if (!loadResult.isSuccess())
                        LOG.error("Something went wrong while trying to read AuthorityList at " + authoritiesFile);
                    else
                        LOG.debug("Authority List from file \"" + authoritiesFile + "\" has been processed " +
                                "(for each authority in that file, previously loaded data for that authority is overwritten. " +
                                "If the previous loaded data contained authorities not in the file, these are kept unchanged.)");
                } catch (Throwable e) {
                    LOG.error("Something went wrong while trying to read AuthorityList at " + authoritiesFile, e);
                }
            } else {
                if (authoritiesFile != null)
                    LOG.info("No local AuthorityList found at " + authoritiesFile);
            }
        }

        return authorityListModel;
    }

    public void save(AuthorityListModel authorityListModel) {
        if (authoritiesFile == null) {
            LOG.warn("Could not save jFed authority list because there is no location to save it to.");
            return;
        }
        LOG.info("Saving jFed authority list to file \"" + authoritiesFile + "\"");
        StoredAuthorityList.save(authoritiesFile, authorityListModel);
    }

}
