package be.iminds.ilabt.jfed.lowlevel.connection;

import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import org.apache.http.Header;
import org.apache.http.ProtocolVersion;
import org.apache.http.StatusLine;
import org.apache.xmlrpc.XmlRpcClient;
import org.apache.xmlrpc.XmlRpcClientException;
import org.apache.xmlrpc.XmlRpcTransport;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;
import java.util.Vector;

/**
 * GeniConnection: XmlRpc over HTTP without SSL (= unencrypted, plain HTTP)
 */
public class MockSfaConnection extends SfaConnection {
    private final static Logger LOG = LoggerFactory.getLogger(MockSfaConnection.class);

    private XmlRpcClient xmlRpcClient;
    private MockXmlRpcTransportFactory mockXmlRpcTransportFactory;

    private List<MockXmlRpcReply> mockReplies = new ArrayList<MockXmlRpcReply>();
    private List<OnMockXmlRpcListener> noMockListeners = new ArrayList<OnMockXmlRpcListener>();
    private List<MockXmlRpcHistoryListener> historyListeners = new ArrayList<MockXmlRpcHistoryListener>();

    /**
     * Create a connection with debugMode disabled and with the specified auth added as debugInfo
     *
     * @param serverUrl url to connect to (may not be null)
     * @param auth the authority to add for debug information purposes (may be null)
     * @param proxyInfo the proxy to use (may be null if no proxy)
     * */
    public MockSfaConnection(String serverUrl, SfaAuthority auth, ProxyInfo proxyInfo) throws JFedException {
        this(serverUrl, proxyInfo, new DebugInfo(false, auth, null));
    }

    /**
     * Create a connection by specifying all low level details needed.
     *
     * @param serverUrl url to connect to (may not be null, unless this is a fake connection used for debugging)
     * @param proxyInfo the proxy to use (may be null if no proxy)
     * @param debugInfo debug information, such as authority and user, and whether debug mode is enabled. (may be null)
     * @throws be.iminds.ilabt.jfed.lowlevel.JFedException
     */
    public MockSfaConnection(String serverUrl, ProxyInfo proxyInfo, DebugInfo debugInfo) throws JFedException {
        super(serverUrl, null, proxyInfo, debugInfo);

        mockXmlRpcTransportFactory = new MockXmlRpcTransportFactory();
        try {
            xmlRpcClient = new XmlRpcClient(new URL(serverUrl), mockXmlRpcTransportFactory);
            error = false;
        } catch (Exception e) {
            throw new JFedException("Error creating XmlRpcClient in MockSfaConnection constructor: " + e.getMessage(), e);
        }
    }

    @Override
    public MockXmlRpcTransportFactory getXmlRpcTransportFactory() {
        return mockXmlRpcTransportFactory;
    }

    @Override
    public XmlRpcClient getXmlRpcClient() {
        assert !debugInfo.isFakeForDebugging() :
                "Error: This connection is a fake debugging connection, yet the XmlRpcClient was requested";
        return xmlRpcClient;
    }

    public void clearLastCallData() {
        mockXmlRpcTransportFactory.clearLastCallData();
    }

    public void close() {
        error = true;
        if (mockXmlRpcTransportFactory != null) {
            mockXmlRpcTransportFactory.clearLastCallData();
            mockXmlRpcTransportFactory.close();
            mockXmlRpcTransportFactory = null;
        }
        xmlRpcClient = null;
    }

    private class MockXmlRpcTransportFactory implements HttpInfoXmlRpcTransportFactory {
        private MockXmlRpcTransport transport = null;

        /////////////////////////////////////
        // XmlRpcTransportFactory methods
        /////////////////////////////////////
        @Override
        public XmlRpcTransport createTransport() throws XmlRpcClientException {
            if (transport == null)
                transport = new MockXmlRpcTransport();
            return transport;
        }

        @Override
        public void setProperty(String s, Object o) {

        }

        /////////////////////////////////////
        // HttpInfoXmlRpcTransportFactory methods
        /////////////////////////////////////
        private boolean debug = false;
        public String httpRecvHist = null;
        public String httpSentHist = null;
        public StatusLine httpStatusLine = null;
        public String httpRequestLine = null;
        public String httpRequestHeaders = null;
        public String httpResponseHeaders = null;
        @Override
        public boolean isDebugMode() {
            return debug;
        }

        @Override
        public void setDebugMode(boolean debug) {
            this.debug = debug;
        }

        @Override
        public String getHttpReceivedHistory() {
            return httpRecvHist;
        }

        @Override
        public String getHttpSentHistory() {
            return httpSentHist;
        }

        @Override
        public StatusLine getHttpStatusLine() {
            return httpStatusLine;
        }

        @Override
        public String getHttpRequestLine() {
            return httpRequestLine;
        }

        @Override
        public String getHttpRequestHeaders() {
            return httpRequestHeaders;
        }

        @Override
        public String getHttpResponseHeaders() {
            return httpResponseHeaders;
        }

        @Override
        public List<Header> getRawHttpResponseHeaders() {
            return null;
        }

        @Override
        public void clearLastCallData() {
            httpRecvHist = "";
            httpRequestHeaders = null;
            httpStatusLine = null;
            httpRequestLine = null;
            httpResponseHeaders = null;
            httpSentHist = "";
        }

        @Override
        public void close() {
            if (transport != null)
                try {
                    transport.endClientRequest();
                } catch (XmlRpcClientException e) {
                    //ignore
                }
            transport = null;
            clearLastCallData();
        }

        private class MockXmlRpcTransport implements XmlRpcTransport {
            private ByteArrayInputStream is;

            public InputStream sendXmlRpc(byte[] bytes) throws IOException, org.apache.xmlrpc.XmlRpcClientException {
                String request = new String(bytes, "UTF-8");
                httpSentHist += request;
                MockXmlRpcAnswer reply = null;
                Throwable seenException = null;
                MockXmlRpcRequestInfo requestInfo = new MockXmlRpcRequestInfo(serverUrl, lastCommand, lastArgv, request);
                for (MockXmlRpcReply mockXmlRpcReply : mockReplies) {
                    try {
                        reply = mockXmlRpcReply.matches(requestInfo);
                    } catch (Throwable e) {
                        LOG.error("Throwable while matching for mock reply", e);
                        seenException = e;
                    }

                    if (reply != null) {
                        mockXmlRpcReply.registerUsage();
                        MockXmlRpcHistoryItem historyItem = new MockXmlRpcHistoryItem(requestInfo, reply);
                        for (MockXmlRpcHistoryListener histListener : historyListeners)
                            histListener.onMockXmlRpcHistory(historyItem);
                        break;
                    }
                }
                if (reply == null && seenException == null) {
                    for (OnMockXmlRpcListener l : noMockListeners)
                        l.onNoMockFor(requestInfo);
                    throw new XmlRpcClientException("No mock answer for this request: " + request, null);
                }

                if (reply == null && seenException != null) {
                    throw new XmlRpcClientException("Got exception while looking for mock answer for " + request, seenException);
                }

                if (reply.isException()) {
                    LOG.debug("Mocked exception", reply.getEx());
                    throw new XmlRpcClientException("Mocked exception", reply.getEx());
                } else {
                    httpRecvHist += reply.getString();
                    httpStatusLine = new StatusLine() {
                        @Override
                        public ProtocolVersion getProtocolVersion() {
                            return new ProtocolVersion("http", 1, 1);
                        }

                        @Override
                        public int getStatusCode() {
                            return 200;
                        }

                        @Override
                        public String getReasonPhrase() {
                            return "OK";
                        }
                    };
                    is = new ByteArrayInputStream(reply.getString().getBytes(StandardCharsets.UTF_8));
                    return is;
                }
            }

            public void endClientRequest() throws org.apache.xmlrpc.XmlRpcClientException {
                try {
                    if (is != null)
                        is.close();
                } catch (IOException e) {
                    LOG.warn("Error closing ByteArrayInputStream MockXmlRpcTransport.is", e);
                }
                is = null;
            }
        }
    }

    //this is dirty, but it works  (it enables adding to these lists using TestCoreModule after injection)
    public void setMockReplies(List<MockXmlRpcReply> mockReplies) {
        this.mockReplies = mockReplies;
    }

    public void setNoMockListeners(List<OnMockXmlRpcListener> noMockListeners) {
        this.noMockListeners = noMockListeners;
    }

    public void setHistoryListeners(List<MockXmlRpcHistoryListener> historyListeners) {
        this.historyListeners = historyListeners;
    }

//    public void addMockReply(MockXmlRpcReply mockReply) {
//        mockReplies.add(mockReply);
//    }
//
//    public List<MockXmlRpcHistoryListener> getHistoryListeners() {
//        return historyListeners;
//    }
//
//    public List<OnMockXmlRpcListener> getNoMockListeners() {
//        return noMockListeners;
//    }
//
//    public List<MockXmlRpcReply> getMockReplies() {
//        return mockReplies;
//    }
    public static class MockXmlRpcRequestInfo {
        public final String serverUrl;
        public final String method;
        public final Vector params;
        public final String xmlRpcRawRequest;

        public MockXmlRpcRequestInfo(String serverUrl, String method, Vector params, String xmlRpcRawRequest) {
            this.serverUrl = serverUrl;
            this.method = method;
            this.params = new Vector(params);
            this.xmlRpcRawRequest = xmlRpcRawRequest;
        }

        @Override
        public String toString() {
            return "MockXmlRpcRequestInfo{" +method+ " @ " + serverUrl + " }";
    }
}
    public static class MockXmlRpcHistoryItem {
        public final MockXmlRpcRequestInfo requestInfo;
        public final MockXmlRpcAnswer reply;

        public MockXmlRpcHistoryItem(MockXmlRpcRequestInfo requestInfo, MockXmlRpcAnswer reply) {
            this.requestInfo = requestInfo;
            this.reply = reply;
        }
    }
    public static interface MockXmlRpcHistoryListener {
        public void onMockXmlRpcHistory(MockXmlRpcHistoryItem item);
    }


    public static class MockXmlRpcAnswer {
        private final String reply;
        private final Throwable ex;

        public MockXmlRpcAnswer(String reply) {
            this.reply = reply;
            this.ex = null;
        }

        public MockXmlRpcAnswer(Throwable ex) {
            this.reply = null;
            this.ex = ex;
        }

        public boolean isException() {
            return ex != null;
        }

        public String getString() {
            return reply;
        }

        public Throwable getEx() {
            return ex;
        }
    }

    public static interface MockXmlRpcReply {
        public MockXmlRpcAnswer matches(MockXmlRpcRequestInfo info);
        public void registerUsage();
        public int usageCount();
    }
    public static interface OnMockXmlRpcListener {
        public void onNoMockFor(MockXmlRpcRequestInfo mockXmlRpcRequestInfo);
    }

    public static abstract class AbstractMockXmlRpcReply implements MockXmlRpcReply {
        private int usages = 0;
        public void registerUsage() {
            usages++;
        }

        public int usageCount() {
            return usages;
        }
    }
}
