package be.iminds.ilabt.jfed.rspec.model.impl;

import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.lowlevel.authority.SfaAuthority;
import be.iminds.ilabt.jfed.rspec.model.*;
import be.iminds.ilabt.jfed.rspec.parser.ExtraXml;
import be.iminds.ilabt.jfed.rspec.parser.RspecCompare;
import be.iminds.ilabt.jfed.rspec.parser.RspecParseException;
import be.iminds.ilabt.jfed.rspec.parser.StaxRspecWriter;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.RFC3339Util;
import org.slf4j.LoggerFactory;

import java.text.ParseException;
import java.util.*;

/**
 * BasicModelRspec
 */
public class BasicModelRspec implements ModelRspec {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(BasicModelRspec.class);

    private final List<ExtraXml> extraXml = new ArrayList<ExtraXml>();
    private final Map<String, String> origNameSpaceInfo = new HashMap<String, String>();

    Boolean losingData;
    List<RspecCompare.Difference> losingDataDifferences;

    private final List<BasicRspecNode> nodes = new ArrayList<BasicRspecNode>();
    private final List<BasicRspecLink> links  = new ArrayList<BasicRspecLink>();
    private final List<BasicLease> leases = new ArrayList<BasicLease>();
    private final List<BasicChannel> channels = new ArrayList<BasicChannel>();
    private final List<BasicOpenflowDataPath> openflowDataPaths  = new ArrayList<BasicOpenflowDataPath>();
    private boolean origIsParsed = false;
    private String origDefaultNamespace = null;
    private String origSchemaLocation = null;
    private String rspecType;
    private final List<UserSpec> sshKeys = new ArrayList<UserSpec>();
    private String expiresString;

    private final RspecFactory factory;

    /**
     * Create an empty Rspec
     */
    public BasicModelRspec(String rspecType) {
        this.rspecType = rspecType;
        this.losingData = false;
        this.factory = RspecFactoryFactory.getRspecFactoryInstance(getModelRspecType());
    }

    @Override
    public ModelRspecType getModelRspecType() {
        return  ModelRspecType.BASIC;
    }

    @Override
    public void setParseData(String schemaLocation, String defaultNamespace, Map<String, String> namespaces) {
        origIsParsed = true;
        this.origSchemaLocation= schemaLocation;
        this.origDefaultNamespace = defaultNamespace;
        this.origNameSpaceInfo.putAll(namespaces);
    }

    @Override
    public String getSchemaLocation() {
        return origSchemaLocation;
    }
    @Override
    public String getOrigDefaultNamespace() {
        return origDefaultNamespace;
    }
    @Override
    public Map<String, String> getNameSpaceInfo() {
        return origNameSpaceInfo;
    }

    @Override
    public Boolean isLosingData() {
        return losingData;
    }
    @Override
    public List<RspecCompare.Difference> getLosingDataDifferences() {
        return losingDataDifferences;
    }
    @Override
    public void setLosingData(Boolean losingData, List<RspecCompare.Difference> differences) {
        this.losingData = losingData;
        this.losingDataDifferences = differences;
    }

    public void doubleCheckLosingData(String origXml) {
        String currentRspec = toGeni3Rspec();
        boolean checkLosingData = !StringRspec.isSimilarXml(origXml, currentRspec); //can be very slow method on large XMLs
        if (checkLosingData != losingData)
            throw new RuntimeException("JFed bug: incorrect losingData detection. losingData="+losingData+" checkLosingData="+checkLosingData);
        LOG.debug("doubleCheckLosingData success: losingData="+losingData+" checkLosingData="+checkLosingData);
    }

    @Override
    public void addNode(RspecNode node) {
        assert node instanceof BasicRspecNode;
        nodes.add((BasicRspecNode) node);
    }
    @Override
    public void addLink(RspecLink link) {
        assert link instanceof BasicRspecLink;
        links.add((BasicRspecLink) link);
    }

    @Override
    public List<BasicRspecNode> getNodes() {
        return nodes;
    }

    @Override
    public List<BasicRspecLink> getLinks() {
        return links;
    }

    @Override
    public void addOpenflowDataPath(OpenflowDataPath dp) {
        assert dp instanceof BasicOpenflowDataPath;
        openflowDataPaths.add((BasicOpenflowDataPath)dp);
    }

    @Override
    public List<BasicOpenflowDataPath> getOpenflowDataPaths() {
        return openflowDataPaths;
    }


    @Override
    public void addLease(Lease lease) {
        assert lease instanceof BasicLease;
        leases.add((BasicLease) lease);
    }

    @Override
    public List<? extends Lease> getLeases() {
        return leases;
    }

    @Override
    public void addChannel(Channel channel) {

        assert channel instanceof BasicChannel;
        channels.add((BasicChannel)channel);
    }

    @Override
    public List<BasicChannel> getChannels() {
        return channels;
    }

    @Override
    public BasicRspecNode getNodeByUniqueId(String id) {
        for (BasicRspecNode node : nodes)
            if (node.getUniqueId().equals(id))
                return node;
        return null;
    }
    @Override
    public BasicRspecNode getNodeByClientId(String id) {
        for (BasicRspecNode node : nodes)
            if (node.getClientId().equals(id))
                return node;
        return null;
    }
    @Override
    public BasicRspecNode getNodeByComponentId(String id) {
        for (BasicRspecNode node : nodes)
            if (node.getComponentId().toString().equals(id))
                return node;
        return null;
    }

    @Override
    public BasicChannel getChannelByComponentId(GeniUrn urn) {
        for(BasicChannel channel: channels)
        if(channel.getComponentId().equals(urn))
            return channel;
        return null;
    }



    @Override
    public List<BasicRspecNode> getNodesByAuthority(SfaAuthority authority) {
        List<BasicRspecNode> nodesOfAuthority = new ArrayList<BasicRspecNode>();
        for (BasicRspecNode node : nodes) {
            if (node.getComponentManagerId().equalsIgnoringSubAuth(authority.getUrn())) {
                nodesOfAuthority.add(node);
            }
        }
        return nodesOfAuthority;
    }

    @Override
    public BasicRspecLink getLinkByClientId(String id) {
        for (BasicRspecLink link : links)
            if (link.getClientId().equals(id))
                return link;
        return null;
    }
    @Override
    public BasicRspecLink getLinkByUniqueId(String id) {
        for (BasicRspecLink link : links)
            if (link.getUniqueId().equals(id))
                return link;
        return null;
    }

    @Override
    public void deleteNode(RspecNode node) {
        assert node instanceof BasicRspecNode;
        List<RspecLink> linksToDelete = new ArrayList<RspecLink>();
        List<RspecInterface> ifacesToDelete = new ArrayList<RspecInterface>();
        for (RspecInterface iface : node.getInterfaces()) {
            if (iface.isLinkBound()) {
                RspecLink link = iface.getLink();
                if (link != null && link.getInterfaces().size() <= 2)
                    linksToDelete.add(link);
            }
            ifacesToDelete.add(iface);
        }
        for (RspecInterface iface : ifacesToDelete)
            iface.delete();
        nodes.remove(node);
        for (RspecLink link : linksToDelete)
            deleteLink(link);
    }

    @Override
    public void deleteLink(RspecLink link) {
        assert link instanceof BasicRspecLink;
        AbstractList<RspecInterface> ifacesToDelete = new ArrayList<RspecInterface>(link.getInterfaces());
        for (RspecInterface iface : ifacesToDelete)
            iface.delete();
        links.remove(link);
    }

    @Override
    public boolean deleteChannel(Channel channel) {
        assert channel instanceof BasicChannel;
        return channels.remove(channel);
    }

    @Override
    public String toGeni3Rspec(RequestRspecSpecialCases specialCase) {
        StaxRspecWriter staxRspecWriter = new StaxRspecWriter(this);

        switch (specialCase) {
            case NONE: {
                //default staxRspecWriter options are fine
                break;
            }
            case PLE: {
                //already using quick hack to get ple to work in: RspecNode.writePropertiesToGeni3RequestRspec()
                List<String> allowedComponentManagers = new ArrayList<String>();
                allowedComponentManagers.add("urn:publicid:IDN+ple:ibbtple+authority+cm");
                //hack for ple: don't add nodes from other componentmanagers
                staxRspecWriter.setComponentManagerFilter(allowedComponentManagers);

                //another hack for ple: ple also doesn't like schemaLocation
                staxRspecWriter.setAddSchemaLocation(false);

                //yet another hack for PLE: add a meaningless, empty <sliver/> element to each node
                staxRspecWriter.setAddEmptySliverToNodeHack(true);
                break;
            }
            default:
                throw new RuntimeException("Special case " + specialCase + " not supported");
        }

        try {
            return staxRspecWriter.call();
        } catch (RspecParseException e) {
            LOG.error("Error converting to Rspec: ", e);
            throw new RuntimeException("Error converting to Rspec: " + e, e);
        }
    }

    @Override
    public String toGeni3Rspec() {
        return toGeni3Rspec(RequestRspecSpecialCases.NONE);
    }

    @Override
    public Date getExpires() throws ParseException {
        if (expiresString == null) return null;
        return RFC3339Util.rfc3339StringToDate(expiresString);
    }

    @Override
    public void setExpires(Date expires) {
        if (expires == null)
            this.expiresString = null;
        else
            this.expiresString = RFC3339Util.dateToRFC3339String(expires, true/*forceZulu*/); //must be Z for emulab
    }

    @Override
    public String getExpiresString() {
        return expiresString;
    }

    @Override
    public void setExpires(String expires) {
        this.expiresString = expires;
    }

    @Override
    public List<GeniUrn> getAllComponentManagerUrns() {
        Set<GeniUrn> urns = new HashSet<GeniUrn>();
        for (RspecNode node : nodes)
            if (node.getComponentManagerId() != null)
                urns.add(node.getComponentManagerId());
        for (RspecLink link : links)
            for (GeniUrn cmUrn : link.getComponentManagerUrns())
                urns.add(cmUrn);

        return new ArrayList<GeniUrn>(urns);
    }

    @Override
    public List<UserSpec> getSshKeys() {
        return sshKeys;
    }

    @Override
    public void setSshKeys(List<UserSpec> newKeys) {
        this.sshKeys.clear();
        this.sshKeys.addAll(newKeys);
    }

    @Override
    public String nextLinkName() {
        int c = links.size();
        String res = "link" + c;
        while (getLinkByClientId(res) != null)
            res = "link" + (++c);

        return res;
    }

    @Override
    public String nextNodeName() {
        int c = nodes.size();
        String res = "node" + c;
        while (getNodeByClientId(res) != null)
            res = "node" + (++c);

        return res;
    }

    @Override
    public String getType() {
        return rspecType;
    }

    @Override
    public String nextIfaceName(RspecNode node) {
        int c = node.getInterfaces().size();
        String res = node.getClientId() + ":if" + c;
        while (getInterfaceByClientId(res) != null)
            res = node.getClientId() + ":if" + (++c);

        return res;
    }

    @Override
    public BasicRspecInterface getInterfaceByClientId(String clientId) {
        for (BasicRspecNode node : nodes) {
            BasicRspecInterface iface = node.getInterfaceByClientId(clientId);

            if (iface != null)
                return iface;
        }
        return null;
    }

    @Override
    public BasicRspecInterface getInterfaceByUniqueId(String uniqueId) {
        for (BasicRspecNode node : nodes) {
            BasicRspecInterface iface = node.getInterfaceByUniqueId(uniqueId);

            if (iface != null)
                return iface;
        }
        return null;
    }

    @Override
    public List<ExtraXml> getExtraXml() {
        return extraXml;
    }

    @Override
    public void setExtraXml(List<ExtraXml> extraXml) {
        this.extraXml.clear();
        this.extraXml.addAll(extraXml);
    }

    @Override
    public ModelRspec copy() {
        return factory.copyModelRspec(this, null/*new rspec type (request/manifest/...)*/);
    }
}
